/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id$
 *-----------------------------------------------------------------------------
 */





package com.lutris.classloader;

// lutris packages
import com.lutris.logging.LogChannel;
import com.lutris.util.FatalExceptionError;

// net packages
import java.net.URL;

// io packages
import java.io.InputStream;

// io exceptions
import java.io.IOException;
import java.io.FileNotFoundException;

// zip packages
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

/**
 * <P>A <CODE>Resource</CODE> that is an entry in 
 * a specified zip file on a remote machine.  The zip file is represented by a 
 * <CODE>ClassPathEntry</CODE>, and the filename is specified by a String.
 *
 * @author Kristen Pol, Lutris Technologies
 * @version $Revision : 1.1 $
 * @see com.lutris.classloader.MultiClassLoader
 * @see com.lutris.classloader.ClassPathEntry
 * @see com.lutris.classloader.Resource
 * @see java.io.File
 */
public class RemoteZipResource extends Resource {

    // data members

    URL zipFileURL = null;

    // constructors

    // FIXME: Test and change to protected and add javadoc. (kp)
    private RemoteZipResource(String name, ClassPathEntry location,
	    LogChannel loadLogChannel) 
	    throws FileNotFoundException {
	super(name, location, loadLogChannel);

        // Get location's URL
        zipFileURL = location.getURL();
        if (zipFileURL == null) {
	    String error = "The URL for location, " + location + ", is null";
	    logChannel.write(logLevel, error);
            throw new FileNotFoundException(error);
        }
 
	// Make sure the ZipEntry exists in the remote zip file
	// FileNotFoundException will be thrown if it is not found
	ZipInputStream zipStream = null;
	try {
	    zipStream = getZipInputStream();
	    if (zipStream == null) {
		String error = "URL \"" + zipFileURL + "\" does not exist, " +
		    "can not be reached, or is not a zip file; or the " +
		    "resource \"" + name + "\" is not in the zip file";
		logChannel.write(logLevel, error);
		throw new FileNotFoundException(error);
	    }
	} finally {
	    if (zipStream != null) {
		try {
		    zipStream.close();
		} catch (IOException ioe) {
		    // This should not happen because it was already checked
		    String error = ioe.toString() + ": URL zip file \"" + 
			zipFileURL + "\" is corrupt or the connection died";
		    logChannel.write(logLevel, error);
		    throw new FatalExceptionError(new IOException(error));
		}
	    }
	}
    }
    
    // public methods
    
    /**
     * Gets the specified resource as an input stream.
     *
     * @return an <CODE>InputStream</CODE> representing the specified resource.
     * @exception an <CODE>IOException</CODE>
     */
    public InputStream getInputStream() throws IOException {
	try {
	    return getZipInputStream();
	} catch (FileNotFoundException fnfe) {
	    // This should not happen because it was already checked
            String error = "URL zip file \"" + zipFileURL + 
		"\" has disappeared or can no longer be reached; or the " + 
		"resource \"" + name + "\" is no longer in the zip file";
	    logChannel.write(logLevel, error);
            throw new FatalExceptionError(new FileNotFoundException(error));
	}
    }

    // Get the input stream for the resource to if it exists.
    private ZipInputStream getZipInputStream() throws FileNotFoundException {
        ZipInputStream zipStream;
        try {
            zipStream = new ZipInputStream(zipFileURL.openStream());
            while (zipStream.available() > 0) {
                ZipEntry entry = zipStream.getNextEntry();
                if (entry == null) {
		   logChannel.write(logLevel, "Null zip entry.");
                   throw new IOException();
                }
                if (entry.getName().equals(name)) {
		    size = entry.getSize();
		    lastModifiedTime = entry.getTime();
                    return zipStream;
                }
                zipStream.closeEntry();
            }
        } catch (IOException e) {
            String error = "URL, " + zipFileURL + ", does not exist, " +
		"can not be reached, is not a zip file, or the resource, " + 
		name + ", is not in the zip file";
	    logChannel.write(logLevel, error);
            throw new FileNotFoundException(error);
        }
	return null;
    }

    /**
     * Get current last-modification time of resource.  This is the
     * time on the disk file the resource is associated with.
     *
     * @return the last-modified time of the permanent copy of the resource
     * in milliseconds.
     */
    public long getCurrentLastModifiedTime() {
        return -1; //FIXME: implement
    }
}
