/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.common;

import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Properties;

/**
 * This class manages global variables and properties used by EJB Server. 
 * These properties are read using the rules defines in Prop.java.
 * A static design pattern is used. 
 */
public class Env {

    static private final int trace = Trace.DB_2;

    // Variables related to OS name
    static public  final int  WINDOWS_NT = 0;
    static public  final int  AIX        = 1;
    static public  final int  SOLARIS    = 2;
    static private       int  osName     = -1;

    // Variables related to Java Version
    static public  final int  JAVA_1_1_6  = 116;
    static public  final int  JAVA_1_1_7  = 117;
    static public  final int  JAVA_1_1_8  = 118;
    static public  final int  JAVA_1_2    = 120;
    static public  final int  JAVA_1_3    = 130;
    static private       int  javaVersion = -1;

    public static final String DATASOURCES = "jonas.datasources";
    public static final String JONAS_NAME = "jonas.name";

    private static final String EXTENSION_PROPERTIES = ".properties";

    /**
     * Gets OS.
     * @return	WINDOWS_NT|SOLARIS|AIX or -1 if error
     */
    static public int getOS() {

	if (osName == -1) {
	    // Sets OS name
	    String stros = System.getProperties().getProperty("os.name", "");
	    if (stros.equalsIgnoreCase("Windows NT"))
		osName = WINDOWS_NT;
	    if (stros.equalsIgnoreCase("AIX"))
		osName = AIX;
	    if (stros.equalsIgnoreCase("Solaris"))
		osName = SOLARIS;
	} 
	return osName;
    }

    /**
     * Gets Java Version.
     * @return javaVersion or -1 if error
     */
    static public int getJavaVersion() {

	if (javaVersion == -1) {
	    // Sets Java Version
	    String strjv = System.getProperties().getProperty("java.version", "");
	    if (strjv.indexOf("1.1.6") == 0)
		javaVersion = JAVA_1_1_6;
	    if (strjv.indexOf("1.1.7") == 0)
		javaVersion = JAVA_1_1_7;
	    if (strjv.indexOf("1.1.8") == 0)
		javaVersion = JAVA_1_1_8;
	    if (strjv.indexOf("1.2") == 0)
		javaVersion = JAVA_1_2;
	    if (strjv.indexOf("1.3") == 0)
		javaVersion = JAVA_1_3;
	} 
	return javaVersion;
    }

    /**
     * return true if JDK 1.2 or later
     */
    static public boolean isJAVA2() {
	return (getJavaVersion() >= JAVA_1_2);
    }

    /**
     * Return true if the current instance of JOnAS has been
     * build with RMI over Jeremie 
     */
    public static boolean isJeremie() {
	return (Orb.ORB > 0);
    }

    /**
     * Gets DataSource properties. 
     *
     * @param name if name is set "name.properties" must be a file containing
     *        properties describing a DataSource
     * @return Properties describing the DataSource or 
     */
    static private Properties getDataSource(String dsN) {

	String propFileName   = dsN+EXTENSION_PROPERTIES;
	Prop   prop           = new Prop(propFileName);
	return prop.getFilesEnv();
    }


    /**
     * Gets DataSources names vector. 
     * @return DataSources vector of properties objects describing dataSources.
     */
    static public Vector getDataSources() {

	Trace.outln(trace, "Env.getDataSources() ");

	Vector retVect     = new Vector();
	String dataSources = Prop.svr.getValue(DATASOURCES, null);
	if (dataSources != null) {
	    StringTokenizer st = new StringTokenizer(dataSources, ",");
	    while(st.hasMoreTokens()){ // Loops on dataSources
		String datasourcename = st.nextToken().trim();
		retVect.addElement(getDataSource(datasourcename));
		Trace.outln(trace, "    datasource ="+datasourcename);
	    }
	} 
	return retVect;
    }


    /** 
     * Gets JOnAS Server name.
     * @return JOnAS Server name.
     */
    static public String getJonasName() {
	String name = Prop.svr.getValue(JONAS_NAME, "");
	return(name);
    }

}

