/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.common;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.PrintWriter;

import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Date;
import java.util.StringTokenizer;


/**
 *  This class implements the tracing facilities.
 */
public class Trace {
    public static final int DB_0 = (1<<0); 	// 
    public static final int DB_1 = (1<<1);  	// tests ejb
    public static final int DB_2 = (1<<2); 	// env
    public static final int DB_3 = (1<<3); 	// naming
    public static final int DB_4 = (1<<4);   	// jtm
    public static final int DB_5 = (1<<5); 	// jtm -jta
    public static final int DB_6 = (1<<6); 	// implicit propagation of context
    public static final int DB_7 = (1<<7); 	// 
    public static final int DB_8 = (1<<8); 	// connection manager
    public static final int DB_9 = (1<<9); 	// jdbc,ldap
    public static final int DB_10 = (1<<10);   	// jdbc_xa
    public static final int DB_11 = (1<<11); 	// trace connection pool (very verbose)
    public static final int DB_12 = (1<<12);    // Context
    public static final int DB_13 = (1<<13); 	// xa resource
    public static final int DB_14 = (1<<14);   	// Container, Server
    public static final int DB_15 = (1<<15); 	// Generated classes
    public static final int DB_16 = (1<<16); 	// container
    public static final int DB_17 = (1<<17); 	// container (preinvoke/postinvoke)
    public static final int DB_18 = (1<<18); 	// tests ejb (old tests)
    public static final int DB_19 = (1<<19); 	// tools
    public static final int DB_20 = (1<<20);  	// deployment
    public static final int DB_21 = (1<<21); 	// 
    public static final int DB_22 = (1<<22); 	// 
    public static final int DB_23 = (1<<23); 	// tests ejb (old tests)
    public static final int DB_24 = (1<<24); 	// 
    public static final int DB_25 = (1<<25);  	// 
    public static final int DB_26 = (1<<26);   	// 
    public static final int DB_27 = (1<<27);   	// 
    public static final int DB_28 = (1<<28);   	// 
    public static final int DB_29 = (1<<29);   	//   
    public static final int DB_30 = (1<<30); 	// ldap buffers in and out
    public static final int DB_31 = (1<<31);  	// Warnings
    public static final int DB_NONE = 0x0000;   
    public static final int DB_ALL = 0xFFFFFFFF;    
  
    private static Trace mT = new Trace("");  // "" means current directory

    // standard output
    private static int mOutDevice; // 0: none; 1: standard output; 2: standard error; 3: file
    private static PrintWriter mOutPrintWriter = null;
	
    // tempo before error message output
    private static final int TEMPO_DEFAULT = 1000;
    private static final int TEMPO_HELENE = 10000;
    private static int mTempo = 0; // milliseconds
  
    // standard error
    private static int mErrDevice; // 0: none; 1: standard output; 2: standard error; 3: file
    private static PrintWriter mErrPrintWriter = null;
  
    private static int mLevel = 0x00; 

    private static boolean mWithHeader = true;

    private static String mPrefixString = "";
    private static String mTmpDir;


    // constructor is private to prevent client program from instanciating a 'Trace'
    private Trace(String s) { mTmpDir = s; };


    /**
     * To make a tempo after message
     */
    private static void tempo() {
	if (mTempo > 0) {
	    try {
		Thread.currentThread().sleep(mTempo);
	    } catch (InterruptedException e) {
		if (mTempo == TEMPO_DEFAULT)
		    mTempo = TEMPO_HELENE;
		else
		    mTempo = TEMPO_DEFAULT;
	    }
	}
    }


    /**
     * To initialize trace from the properties
     */
    public static void configure() {
	boolean bHeader = false;
	String out = Prop.svr.getValue("jonas.trace.out");
	String err = Prop.svr.getValue("jonas.trace.err");
	String level = Prop.svr.getValue("jonas.trace.level");
	String header = Prop.svr.getValue("jonas.trace.header");

	if (header.equalsIgnoreCase("true"))
	    bHeader = true;
	configure(out, err, levelValue(level), bHeader);
	if (Prop.svr.getValue("jonas.trace.tempo", "false").equalsIgnoreCase("true"))
	    mTempo = TEMPO_DEFAULT;
    }


    /**
     * To initialize trace from parameters 
     *  No header
     * @param out    output device  "out": standard output; "err": standard error; other: filename
     * @param err    error device  "out": standard output; "err": standard error; other: filename
     * @param level  trace level in string format
     */
    public static void configure(String out, String err, String level) {
	configure(out, err, levelValue(level));
    }


    /**
     * To initialize trace from parameters without header
     *  No header
     * @param out    output device  "out": standard output; "err": standard error; other: filename
     * @param err    error device  "out": standard output; "err": standard error; other: filename
     * @param level  trace level in int format
     */
    public static void configure(String out, String err, int level) {
	configure(out, err, level, false);
    }


    /**
     * To initialize trace from parameters
     * @param out    output device  "out": standard output; "err": standard error; other: filename
     * @param err    error device  "out": standard output; "err": standard error; other: filename
     * @param level  trace level in int format
     * @param header with header if set to true, without header if set to false.
     */
    public static void configure(String out, String err, int level, boolean withHeader) {

	mOutDevice = 0;
	if (out.equalsIgnoreCase("out"))
	    mOutDevice = 1;
	if (out.equalsIgnoreCase("err"))
	    mOutDevice = 2;
	if (!(out.equalsIgnoreCase("out")) && 
	    !(out.equalsIgnoreCase("err")) && 
	    (out != null)) {
	    if (out.equalsIgnoreCase(err)) {
		String s = mTmpDir + out + ".tracouterr";
		try {
		    mOutPrintWriter = new PrintWriter(new FileOutputStream(s));
		} catch (FileNotFoundException e) {
		    System.err.println("File not found : " + e);
		} catch (IOException e) {
		    System.err.println("IOexception :" + e);
		}
		mErrPrintWriter = mOutPrintWriter;
		mOutDevice = 3;
		mErrDevice = 3;
	    } else {
		String s = mTmpDir + out + ".tracout";
		try {
		    mOutPrintWriter = new PrintWriter(new FileOutputStream(s));
		} catch (FileNotFoundException e) {
		    System.err.println("File not found : " + e);
		} catch (IOException e) {
		    System.err.println("IOexception :" + e);
		}
		mOutDevice = 3;
	    }
	}
	    
	mErrDevice = 0;
	if (err.equalsIgnoreCase("out"))
	    mErrDevice = 1;
	if (err.equalsIgnoreCase("err"))
	    mErrDevice = 2;
	if (!(err.equalsIgnoreCase("out")) && 
	    !(err.equalsIgnoreCase("err")) && 
	    (err != null)) {
	    if (out.equalsIgnoreCase(err)) {
		// nothing to be done
	    } else {
		String s = mTmpDir + err + ".tracerr";
		try {
		    mErrPrintWriter = new PrintWriter(new FileOutputStream(s));
		} catch (FileNotFoundException e) {
		    System.err.println("File not found : " + e);
		} catch (IOException e) {
		    System.err.println("IOexception :" + e);
		}
		mErrDevice = 3;
	    }
	}	
	mLevel = level;
	mWithHeader = withHeader;
    }


    /**
     * Sets the trace prefix
     *
     */
    public static void setPrefix(String prefix) {
	mPrefixString = prefix+" ";
    }


    /**
     * Returns the integer value for a string trace level
     */
    public static int levelValue(String level) {
	int ret = 0;
	StringTokenizer tok = new StringTokenizer(level, ",.-");

	while (tok.hasMoreTokens()) {
	    String token = tok.nextToken().trim();
	    if (token.equalsIgnoreCase("all"))
		return DB_ALL;
	    if (token.equalsIgnoreCase("none"))
		return DB_NONE;
	    Integer it = new Integer(token);
	    ret = ret + (1 << it.intValue());
	}
	return ret;
    }


    /**
     * Returns true if this trace level is set
     */
    public static boolean isLevelSet(int level) {
	if ((level & mLevel) != 0) return true;
	return false;
    }


    /**
     *  Sends the trace message to the trace output device if flag is configured
     * @param f   this flag must be configured to get the message on output
     * @param s   message to send
     */
    public static void out(int f, String s) {
	if (mLevel == 0x0000) return;
	if ((f & mLevel) != 0)
	    out(s);
    }


    /**
     *  Sends the trace message to the trace output device if flag is configured
     * @param f   this flag must be configured to get the message on output
     * @param s   message to send
     */
    public static void outln(int f, String s) {
	if ((f & mLevel) != 0)
	    out(s + "\n");
    }

   /**
     *  Sends the trace message to the trace output device if flag is configured
     *  the message is build by concatenation of two parameters
     * @param f   this flag must be configured to get the message on output
     * @param s1  first part of message
     * @param s2  second part of message
     */
    public static void outln(int f, String s, Object  o) {
	if ((f & mLevel) != 0)
	    out(s +o+ "\n");
    }

    /**
     *  Sends the trace message to the trace output device if flag is configured
     *  the message is build by concatenation of two parameters
     * @param f   this flag must be configured to get the message on output
     * @param s1  first part of message
     * @param s2  second part of message
     */
    public static void outln(int f, String s, int  i) {
	if ((f & mLevel) != 0)
	    out(s +i + "\n");
    }
    /**
     *  Sends the trace message to the trace output device.
     * @param s   message to send
     */
    public static void outln(String s) {
	out(s + "\n");
    }

    /**
     *  Sends the trace message to the trace output device.
     * @param s   message to send
     */
    public static void out(String s) {
	switch (mOutDevice) {
	case 0:
	    break;
	case 1:
	    System.out.print(header() + s);
	    break;
	case 2:
	    System.err.print(header() + s);
	    break;
	case 3:
	    mOutPrintWriter.print(header() + s);
	    mOutPrintWriter.flush();
	    break;
	}
	tempo();
    }


    /**
     *  Sends the trace message to the trace error device.
     * @param s   message to send
     */
    public static void errln(String s) {
	err(s + "\n");
    }

    /**
     *  Sends the exception and the current stack to the trace error device.
     * @param e      exception to send
     * @param stack  dumps stack if set to true
     */
    public static void errln(Throwable e, boolean stack) {
	if (stack) {
	    switch (mErrDevice) {
	    case 0:
		break;
	    case 1:
		e.printStackTrace(System.out);
		break;
	    case 2:
		e.printStackTrace(System.err);
		break;
	    case 3:
		e.printStackTrace(mErrPrintWriter);
		mErrPrintWriter.flush();
		break;
	    }
	    err("\n");
	} else {
	    errln(e.toString());
	}
    }


    /**
     *  Sends the exception to the trace error device.
     * @param e      exception to send
     */
    public static void errln(Throwable e) {
	errln(e, true);
    }


    /**
     *  Sends the stream to the trace error device.
     * @param f      stream to send
     */    
    public static void errln(InputStream f) {
	try {
	    InputStreamReader isr = new InputStreamReader(f);
	    BufferedReader br = new BufferedReader(isr);
	    String line;
	    while ((line = br.readLine()) != null) {
		errln(line);
	    }
	    isr.close();
	    br.close();
	} catch (IOException e) {
	    errln("Trace.errln(InputStream) ERROR: " + e);
	}
    }


    /**
     *  Sends the message to the trace error device.
     * @param s      message to send
     */  
    public static void err(String s) {
	switch (mErrDevice) {
	case 0:
	    break;
	case 1:
	    System.out.print(header() + s);
	    break;
	case 2:
	    System.err.print(header() + s);
	    break;
	case 3:
	    mErrPrintWriter.print(header() + s);
	    mErrPrintWriter.flush();
	    break;
	}
	tempo();
    }

  
    /**
     * Closes output and error devices
     */
    public static void close() {
	if (mOutDevice == 3)
	    mOutPrintWriter.close();
	if (mErrDevice == 3)
	    mErrPrintWriter.close();
    }

    /**
     * Returns an header message for trace messages
     * @returns     header message
     */
    private static String header() {
	if (mWithHeader) {
	    return(mPrefixString +
		   Thread.currentThread().getName() + 
		   " : " + 
		   hour() );
	} else {
	    return("");
	}
    }

    /**
     *  Returns an hour message to add in header message
     * @returns     hour message
     */
    private static String hour() {

        GregorianCalendar cal = new GregorianCalendar();
	Date             date = new Date();
	cal.setTime(date);
	   
	return 
	    cal.get(Calendar.HOUR)+":"+
	    cal.get(Calendar.MINUTE)+":"+
	    cal.get(Calendar.SECOND)+":"+
	    cal.get(Calendar.MILLISECOND)+" ";
    }
}

