/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.container;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectInputStream;
import java.lang.Class;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import javax.naming.Context;
import javax.naming.LinkRef;
import javax.naming.NamingException;
import javax.naming.RefAddr;
import javax.transaction.TransactionManager;

import org.objectweb.jonas.common.ClassLoaderFactory;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.deployment.api.BeanDesc;
import org.objectweb.jonas.deployment.api.DeploymentDesc;
import org.objectweb.jonas.deployment.api.DeploymentDescException;
import org.objectweb.jonas.deployment.api.EjbRefDesc;
import org.objectweb.jonas.deployment.api.EntityJdbcCmpDesc;
import org.objectweb.jonas.deployment.api.EnvEntryDesc;
import org.objectweb.jonas.deployment.api.ResourceRefDesc;
import org.objectweb.jonas.deployment.api.SessionDesc;
import org.objectweb.jonas.jtm.TransactionService;
import org.objectweb.jonas.lib.BeanNaming;
import org.objectweb.jonas.naming.ContainerNaming;
import org.objectweb.jonas.naming.NamingManager;


/**
 * This class represents an EJB container as specified in EJB specification 8.2
 *
 * A container is where an enterprise Bean object lives. 
 * Multiple EJB classes can be installed in a single container.
 * For each EJB class installed, the container provides a home interface and
 *  makes this interface available in the JNDI name space.
 */
public class Container {

    static private final int trace = Trace.DB_14;

    static ContainerNaming naming = null;
    private EJBServiceImpl ejbserv = null;

    /**
     * constructor
     */
    public Container(EJBServiceImpl ejbs) {

	Trace.outln(trace, "Container.Container() ");

	naming = NamingManager.getInstance();
	ejbserv = ejbs;
    }

    /**
     * Adds beans in container.
     *
     * @param desc  DD. 
     *
     */
    public void addBeans(BeanDesc  desc) {
	Trace.outln(trace, "Container.addBeans() ");
	
	boolean   isSession = (desc instanceof SessionDesc);
		
	Trace.outln (trace, "Bean Deployment Descriptor of '" + desc.getEjbName() +
			     "' is read");
	Trace.outln (trace, desc.toString());
		
	// Build the JOnAS generated names for that bean.
	BeanNaming beanNames = new BeanNaming(desc);
		
	// Finds the DataSource associated to the bean
	// Useful for Entity with Container Managed Persistence.
	String dataSourceName = null;
	if (desc instanceof EntityJdbcCmpDesc) {
	    dataSourceName = ((EntityJdbcCmpDesc)desc).getDatasourceJndiName();
	}
	Object dataSource = null;

	Context context = naming.getInitialContext();
	if (dataSourceName != null) {
	    try {
		dataSource = context.lookup(dataSourceName);
	    }catch(Exception  ex){
		Trace.errln("JOnAS Server configuration error!");
		Trace.errln("     Datasource '"+dataSourceName+"' is not known by the EJB server.");
		Trace.errln("     Please check the 'jonas.datasources' property in the 'jonas.properties' file.");  
	    }
	}
	// Loads the specific EJBHome class
	String ejbHomeClassName = beanNames.getFullWrpHomeName();
	Class ejbHomeClass = null;
	try {
	    // Muly changes partly undone
	    //ejbHomeClass = ClassLoaderFactory.getClassLoader().loadClass(ejbHomeClassName);
	    ejbHomeClass = desc.getHomeClass().getClassLoader().loadClass(ejbHomeClassName);
	} catch (ClassNotFoundException e) {
	    Trace.errln("    Cannot load "+ ejbHomeClassName );
	    System.exit(1);
	}
	Trace.outln (trace, ejbHomeClassName+ " loaded");
		
		
	// Creates a unique EJBHome instance by calling the appropriate constructor
	Object ejbHomeInstance  = null;
	String beanName         = desc.getEjbName();
	try {
	    int nbParams = isSession ? 2 : 3;
	    Class paramTypes[]   = new Class[nbParams];  // constructor argument types
	    Object paramObjects[] = new Object[nbParams]; // constructor argument values
		    
	    if (isSession) {
		paramTypes[0]       = org.objectweb.jonas.deployment.api.SessionDesc.class;
	    } else {
		paramTypes[0]       = org.objectweb.jonas.deployment.api.EntityDesc.class;
	    }
	    paramObjects[0]     = (Object)desc;
		    
	    // Second parameter
	    paramTypes[1]       = java.lang.String.class;
	    paramObjects[1]     = (Object)beanName;
		    
	    // Third parameter
	    if (!isSession) {
		paramTypes[2]   = java.lang.Object.class;
		paramObjects[2] = (Object)dataSource;
	    }
		    
	    Constructor constructor = ejbHomeClass.getConstructor(paramTypes);
	    ejbHomeInstance = constructor.newInstance(paramObjects);
		    
	} catch (InvocationTargetException te) {
	    Trace.errln("    InvocationTargetException : " + te.getTargetException());
	    te.getTargetException().printStackTrace();
	} catch (Exception e) {
	    Trace.errln("    Call of " + ejbHomeClassName + " constructor : " + e);
	    e.printStackTrace();
	}

	// Keep the new EJBHome in a list
	ejbserv.addHome(ejbHomeClassName, ejbHomeInstance);
	Trace.outln (Trace.DB_31, "EJBHome:"+ejbHomeClassName+" for " + beanName + " available");

	// Bean Environment
	try {
	    // Create a JNDI context for this bean home
	    Context ctx = naming.createComponentContext(beanName);
	    ((JBeanHome)ejbHomeInstance).initJNDIContext(ctx);

	    // setup the bean environment in this context
	    setBeanEnvironment(ctx, desc);
	} catch (NamingException e) {
	    Trace.errln("addBeans: could not setup environment");
	}
    }


    /**
     * Set the bean environment
     * args ctx Context for this bean
     * args dd Bean Deployment Descriptor
     * throws NamingException if could not rebind objects
     */
    private void setBeanEnvironment(Context ctx, BeanDesc dd) throws NamingException {

	// Set bean context
	Context ctxold = naming.setComponentContext(ctx);

	// Bean Environment
	EnvEntryDesc [] envt = dd.getEnvEntryDesc();
	for (int i = 0; i < envt.length; i++) {
	    // get information in descriptor
	    String name = envt[i].getName();
	    Object obj = envt[i].getValue();
	    // register object in JNDI
	    Trace.outln(trace, "Binding object "+name+" -> "+obj);
	    ctx.rebind(name, obj);
	}

	// Resource References
	ResourceRefDesc [] resref = dd.getResourceRefDesc();
	for (int i = 0; i < resref.length; i++) {
	    // get information in descriptor
	    String name = resref[i].getName();
	    String resname = resref[i].getJndiName();
	    // build the LinkRef that will be registered:
	    // FactoryClassName = null, size = 1, refAddr = resname.
	    LinkRef lref = new LinkRef(resname);
	    // register object in JNDI
	    Trace.outln(trace, "Linking resource "+name+" -> "+resname);
	    ctx.rebind(name, lref);
	}
	
	// EJB References
	EjbRefDesc [] ejbref = dd.getEjbRefDesc();
	for (int i = 0; i < ejbref.length; i++) {
	    // get information in descriptor
	    String name = ejbref[i].getName();
	    String ejbname = ejbref[i].getJndiName();
	    LinkRef lref = new LinkRef(ejbname);
	    Trace.outln(trace, "Linking ejb "+name+" -> "+ejbname);
	    ctx.rebind(name, lref);
	}

	// Reset bean context
	naming.setComponentContext(ctxold);
    }


}
