/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.container;

import java.security.Identity;
import java.security.Principal;
import java.util.Enumeration;
import java.util.Properties;
import javax.ejb.EJBContext;
import javax.ejb.EJBHome;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import javax.transaction.UserTransaction;
import org.objectweb.jonas.common.Trace;


/**
 * This class implements javax.ejb.EJBContext class.
 */
public class EJBContextImpl implements EJBContext { 

    protected static final int trace = Trace.DB_12;

    protected JBeanHome home = null;
    protected TransactionManager tm = EJBServiceImpl.tm;

    // These 2 fields should be in RequestCtx (pb if 2 clients use the same EJBContext,
    // for example if they are outside any global transaction ???)
    protected Identity callerId = null;
    protected Principal principal = null;

    // ------------------------------------------------------------------
    // constructors
    // ------------------------------------------------------------------

    /**
     * Constructs an EJBContext 
     * 
     * @param home enterprise bean's home interface 
     */
    public EJBContextImpl(JBeanHome home) {

	Trace.outln(trace, "EJBContextImpl.EJBContextImpl(JBeanHome home)");

	this.home = home;
    }

    // ------------------------------------------------------------------
    // EJBContext implementation
    // ------------------------------------------------------------------

    /**
     * Obtains the java.security.Identity of the caller. 
     * DEPRECATED
     * @return The Identity object that identifies the caller.
     */
    public Identity getCallerIdentity() {
	return callerId;
    }

    /**
     * Obtain the java.security.Principal that identifies the caller.
     *
     * @return The Principal object that identifies the caller.
     */
    public Principal getCallerPrincipal() {
	return principal;
    }

    /**
     * Obtains the enterprise bean's home interface. 
     * @return The enterprise bean's home interface. 
     */
    public EJBHome getEJBHome() {
	return home;
    }

    /**
     * Obtains the enterprise bean's environment properties for EJB 1.0 style.
     * (Conform to EJB 1.1 specification for a container chooses to provide support
     *  for EJB 1.0 style environment properties. Cf EJB 1.1 Chapter 14.5, pages 216-217)
     *  Note: If the enterprise bean has no environment properties, this method returns an empty 
     *        java.util.Properties object. This method never returns null. 
     * @return The environment properties for the enterprise bean. 
     */
    public Properties getEnvironment() {
	Trace.outln(trace, "EJBContextImpl.getEnvironment");
	return home.getEjb10Environment();
    }
	
    /** 
     * Tests if the transaction has been marked for rollback only.
     */
    public boolean getRollbackOnly() {
	Trace.outln(trace, "EJBContextImpl.getRollbackOnly() ");

	try {
	    switch (tm.getStatus()) {
	    case Status.STATUS_MARKED_ROLLBACK:
	    case Status.STATUS_ROLLEDBACK:
	    case Status.STATUS_ROLLING_BACK:
		return true;
	    default:
		return false;
	    } 
	} catch (SystemException e) {
	    Trace.errln("Cannot get status:"+e);
	}
	return true;
    }
 	
    /**
     * Obtains the transaction demarcation interface. 
     * @return The UserTransaction interface that the enterprise bean instance 
     *         can use for transaction demarcation.  
     * @exception IllegalStateException: Thrown if the instance container does 
     *            not make the UserTransaction interface available to the 
     *            instance (only the enterprise beans with the TX_BEAN_MANAGED 
     *            transaction attribute are allowed to use the UserTransaction 
     *            interface). 
     */
    public UserTransaction getUserTransaction() throws IllegalStateException {

	Trace.outln(trace,"EJBContextImpl.getUserTransaction() ");

	if (home.isTxBeanManaged() == false) {
	    throw  new IllegalStateException("This bean is not allowed to use UserTransaction interface");
	}
	return (UserTransaction) tm;
    }

    /**
     * Deprecated. Use boolean isCallerInRole(String roleName) instead.
     * Tests if the caller has a given role.
     * @param role - The java.security.Identity of the role to be tested. 
     * @return True if the caller has the specified role. 
     */
    public boolean isCallerInRole(Identity role) {
	Trace.outln(trace,"EJBContextImpl.isCallerInRole(Identity) ");	
	// TO BE DONE
	return true;
    }

    /**
     * Test if the caller has a given role.
     *
     * @param roleName The name of the security role. 
     * The role must be one of the security roles that is defined in the 
     * deployment descriptor.
     *
     * @return True if the caller has the specified role.
     */
    public boolean isCallerInRole(java.lang.String roleName) {
	Trace.outln(trace,"EJBContextImpl.isCallerInRole(String) ");	
	// TO BE DONE
	return true;
    }

    /** 
     * Marks the current transaction for rollback.
     * The transaction will become permanently marked for rollback.
     *
     */
    public void setRollbackOnly() {

	Trace.outln(trace,"EJBContextImpl.setRollbackOnly() ");

	try {
	    tm.setRollbackOnly();
	} catch (IllegalStateException e) {
	    Trace.errln("Current thread not associated with transaction");
	} catch (SystemException e) {
	    Trace.errln("setRollbackOnly unexpected exception:"+e);
	}
    }

}
