/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999,2000 Bull S.A. All Rights Reserved.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.container;

import com.lutris.classloader.MultiClassLoader;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import javax.naming.Context;
import javax.naming.NamingException;
import javax.transaction.TransactionManager;
import org.objectweb.jonas.common.ClassLoaderFactory;
import org.objectweb.jonas.common.Env;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.deployment.api.BeanDesc;
import org.objectweb.jonas.deployment.api.DeploymentDesc;
import org.objectweb.jonas.deployment.api.DeploymentDescException;
import org.objectweb.jonas.jtm.TransactionServiceImpl;
import org.objectweb.jonas.lib.BeanNaming;

public class EJBServiceImpl implements EJBService {

    static private final int trace = Trace.DB_14;

    // For the moment we have a unique Container
    private Container cont;

    private MultiClassLoader loader;

    private Hashtable homeList = new Hashtable();

    // tm can be accessed directly from everywhere in the package.
    static TransactionManager tm = null;

    // -------------------------------------------------------------------
    // JOnAS Service Implementation
    // -------------------------------------------------------------------

    private static EJBService unique = null;

    /**
     * Private constructor to force only 1 instance.
     */
    private EJBServiceImpl() {
	cont = new Container(this);
	loader = new MultiClassLoader(null);
	tm = TransactionServiceImpl.getInstance().getTransactionManager();
    }

    /**
     * Get the unique instance. Create it at first call.
     * Return an interface of the service.
     */
    public static EJBService getInstance() {
	if (unique == null) {
	    unique = new EJBServiceImpl();
	}
	return unique;
    }

    /**
     * Start the Service
     * Configuration information is passed thru a Context object.
     */
    public void start(Context ctx) throws NamingException {

	Trace.outln(trace, "Starting EJBService");

	// Get the list of beans to load in JNDI context.
	String dds = (String) ctx.lookup("descriptors");
	if (dds != null) {
	    StringTokenizer st = new StringTokenizer(dds, ",");
	    while (st.hasMoreTokens()) {
		addBeans(st.nextToken().trim());
	    }
	}
    }

    // -------------------------------------------------------------------
    // EJBService Implementation
    // -------------------------------------------------------------------

    /**
     * Add beans from a .xml or a .jar file
     */
    public void addBeans(String fileName) {
	
	Trace.outln(trace, "EJBService: addBeans from "+fileName);

	// Get the deployment descriptor from file
	DeploymentDesc dd = null;
	try {
	    if (fileName.endsWith(".jar")) {
		// ejb-jar file
		loader.addClassPath(fileName);
		dd = DeploymentDesc.getInstance(loader);
	    } else {
		// xml file
		dd = DeploymentDesc.getInstance(fileName, BeanNaming.getJonasXmlName(fileName));
	    }
	} catch (DeploymentDescException e) {
	    Trace.errln("Cannot read the deployment descriptors '" + fileName + "': " + e);
	    // throw e XXX
	}

	// Process each bean from the DD
	BeanDesc[] beans = dd.getBeanDesc();
	for (int i = 0; i < beans.length; i++) {
	    cont.addBeans(beans[i]);
	}
    }


    /**
     * Remove beans from the EJB Service
     */
    public void removeBeans(String fileName) {
	
	Trace.outln(trace, "EJBService: removeBeans from "+fileName);

	JBeanHome bbh = (JBeanHome) homeList.get(fileName);
	bbh.unregister();
	homeList.remove(fileName);
	bbh = null;
    }

    /**
     * Unbinds all the EJBHome names from JNDI
     *
     */
    public void removeAllBeans() {

	Trace.outln(trace, "EJBService: removeAllBeans");

	try {
	    for (Enumeration lk = homeList.elements(); lk.hasMoreElements() ;) {
		JBeanHome bbh = (JBeanHome)lk.nextElement();
		bbh.unregister();
	    }
	} catch (Exception e) {
	    Trace.errln("EJBService:  " + e);
	    return;
	}
    }

    /**
     * List beans loaded in EJB Service
     */
    public String [] listBeans() {
	
	Trace.outln(trace, "EJBService: listBeans");

	String [] list = new String[homeList.size()];
	int i = 0;
	for (Enumeration lk = homeList.keys(); lk.hasMoreElements() ;) {
	    list[i++] = (String) lk.nextElement();
	}
	return list;
    }
	
    // -------------------------------------------------------------------
    // Other methods
    // -------------------------------------------------------------------

    /**
     * Add a new Home in the list
     */
    public void addHome(String ejbHomeClassName, Object ejbHomeInstance) {
	homeList.put(ejbHomeClassName, ejbHomeInstance);
    }
}
