/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.container; 

import java.lang.IllegalStateException;
import java.lang.Error;
import java.lang.RuntimeException;
import java.rmi.RemoteException;
import java.util.Properties;
import javax.ejb.EntityBean;
import javax.ejb.EJBContext;
import javax.ejb.EJBHome;
import javax.ejb.EJBObject;
import javax.ejb.EJBException;
import javax.naming.Context;
import javax.ejb.EntityContext;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import javax.transaction.UserTransaction;
import javax.transaction.Status;
import javax.transaction.Synchronization;
import javax.transaction.SystemException;
import org.objectweb.jonas.common.Trace;


/**
 * This class implements javax.ejb.EntityContext interface.
 *
 * This class implements the  context associated to an EJBObject Entity
 * in order to allow concurrent access from multiple transactions 
 */

public class EntityContextImpl extends EJBContextImpl implements EntityContext, Synchronization {

    private JBeanEntity ejbObject  = null; // the remote object
    private EntityBean enterpriseBean = null; // the bean instance
    
    // the following field was previously in SynchroImpl for EntitySynchro
    Transaction beanCoord = null;

    //the following fields  was previously in EntitySynchroImpl
    boolean isContainerManaged = false;  // true if container managed persistence
    boolean ismarkedremoved = false;     // true between a remove and the commit 

    // ------------------------------------------------------------------
    // constructors
    // ------------------------------------------------------------------

    /**
     * Constructs an EntityContext 
     * 
     * @param obj  the EJB Object created
     * @param home the enterprise bean's home interface
     * @param eb   the Enterprise Bean
     * @param sync the Synchronization Object
     *
     */
    public EntityContextImpl(JBeanEntity obj, 
			     JBeanHome home, 
			     EntityBean eb, 
			     boolean containerManaged) {

	super(home);
	Trace.outln(trace, "EntityContextImpl: " ,home);
	
	ejbObject = obj;
	enterpriseBean = eb;
	isContainerManaged = containerManaged; 
    }
  

    /**
     * Constructs a minimal EntityContext used by Finder Methods
     * in interposition classes
     */
    public EntityContextImpl(JBeanHome home) {
	super(home);
	Trace.outln(trace, "EntityContextImpl: ",home);
    }

    // ------------------------------------------------------------------
    // EntityContext implementation
    // ------------------------------------------------------------------


    /**
     * Obtains a reference to the EJB object that is currently associated with the instance. 
     *  An instance of a session enterprise Bean can call this method at anytime between the
     *   ejbCreate() and ejbRemove() methods, including from within the ejbCreate() and
     *   ejbRemove() methods.
     *  An instance can use this method, for example, when it wants to pass a reference to 
     *   itself in a method argument or result.
     * @return The EJB object currently associated with the instance. 
     * @exception IllegalStateException: Thrown if the instance invokes this 
     *            method while the instance is in a state that does not allow the 
     *            instance to invoke this method.
     */
    public  EJBObject getEJBObject() throws IllegalStateException {
	Trace.outln(trace, "EntityContextImpl.getEJBObject");

	if (isMarkedRemoved())
	    throw new IllegalStateException("EJB Object is removed");
	return ejbObject;
    }


    /**
     * Obtains the primary key of the EJB object that is currently associated 
     *  with this instance. 
     * An instance of an entity enterprise Bean can call this method only when the 
     *  instance is associated with an EJB object identity, i.e. in the ejbActivate, 
     *  ejbPassivate, ejbPostCreate method, ejbRemove, ejbLoad, ejbStore, and 
     *  business methods. 
     * Note: The result of this method is the same as the result of 
     *  getEJBObject().getPrimaryKey(). 
     *
     * @return  The EJB object currently associated with the instance. 
     *
     * @exception  IllegalStateException  Thrown if the instance invokes this 
     *             method while the instance is in a state that does not allow 
     *             the instance to invoke this method.     
     */
    public Object getPrimaryKey() throws IllegalStateException {
	Trace.outln(trace, "EntityContextImpl.getPrimaryKey ");
	try{
	    Object pk = ejbObject.getPrimaryKey();
	    if (pk == null)
		throw new IllegalStateException();
	    return pk;
	} catch(java.rmi.RemoteException e) {
	    Trace.errln("EntityContextImpl.getPrimaryKey() RemoteException caught :" +e);
	    throw new IllegalStateException();
	}
    }


    // ------------------------------------------------------------------
    // Other methods
    // ------------------------------------------------------------------

    /**
     *
     * Sets the reference to the EJBObject currently associated with this instance.
     *
     * @param pk - the EJB object currently associated with the instance.
     *
     */
    public void setEJBObject(JBeanEntity ejbObject) {
	this.ejbObject = ejbObject;	
    }
   

   /**
     * Returns the bean instance of this context
     *
     * @return the bean instance
     *
     */
    public EntityBean getInstance() {
	return enterpriseBean;
    }


   // -------------------------------------------------------------------
    // Synchronization implementation
    // -------------------------------------------------------------------

   /** 
     * This beforeCompletion method is called by the transaction 
     * manager prior to the start of the transaction completion process.
     * This method executes in the transaction context of the calling 
     * thread.
     * The Bean's state must be stored on the persistent storage before
     * the completion of the transaction.
    */
    public void beforeCompletion() {

	Trace.outln(trace, "EntityContextImpl.beforeCompletion");

	// Set the JNDI component context (for ejbStore)
	Context bnctx = home.setComponentContext();

	try {
	    storeIfModified();
	} catch (RemoteException e) {
	    // beforeCompletion must not report any exception
	} finally {
	    home.resetComponentContext(bnctx);
	}
    }


    /** 
     * The afterCompletion method is called by the transaction 
     * manager after the transaction is committed or rolled back. 
     * This method executes without a transaction context.
     * Here we can update container tables used for handling multiple instances
     * of a bean.
     * If the bean has bean marked to be remove and the status of the transaction
     * is STATUS_COMMITTED we can remove the EJBObject in the table primarykey-EJBObject
     *
     * @param status The status of the transaction completion.
     */
    public void afterCompletion(int status) {
	Trace.outln(trace, "EntityContextImpl.afterCompletion: Status= ",status);

	// The main thing to do here is to release the instance
	try {
	    ejbObject.releaseInstance(beanCoord);
	} catch(Exception e) {
	    Trace.errln("EntityContextImpl.afterCompletion : releaseInstance fails:"+e);
	}

	// If this instance has been marked to be removed
	// during a transaction committed we can actually remove the EJBObject
	if (ismarkedremoved && (status == Status.STATUS_COMMITTED)) {
	    try {
		((JEntityHome)home).removeEJBObject(ejbObject.getPrimaryKey());
	    } catch (RemoteException e) {
		Trace.errln("EntityContextImpl.afterCompletion : removeEJBObject fails:"+e);
	    }
	    ejbObject.noMoreInUse(true);
	}
	ismarkedremoved = false;
    }


    // -------------------------------------------------------------------
    // Other methods
    // -------------------------------------------------------------------

    /**
     *  Persistence: write data on storage
     */
    public void storeIfModified() throws RemoteException {

	Trace.outln(trace, "EntityContextImpl.storeIfModified");

       boolean isModified = ((JBeanEntity)ejbObject).isModifiedData(enterpriseBean);
	try {
	    enterpriseBean.ejbStore();
	    if (isContainerManaged && isModified) {
		ejbObject.storeData();
	    }
	} catch(RemoteException e) {
	    Trace.errln("EntityContextImpl.storeIfModified : " + e);
	    onException(); 
	    throw e;
	} catch(RuntimeException e) {
	    Trace.errln("EntityContextImpl.storeIfModified : " + e);
	    onException();
	    throw new RemoteException("Execption while storing data", e);
 	} catch(Error e) {
	    Trace.errln("EntityContextImpl.storeIfModified : " + e);
	    onException();
	    throw new RemoteException("Error while storing data", e);
	}
    }

    /**
     * A possible method for implementing Rules for container provider on Exceptions
     * cf spec 12.3.3 & 12.3.4
     *
     */
    public void onException () {

	Trace.outln(Trace.DB_31, "EntityContextImpl.onException");

	// If we are inside a transaction, it must be rolled back
	try {
	    Transaction tr = tm.getTransaction();
	    if (tr != null) {
		tr.setRollbackOnly();
	    }
	} catch (SystemException e) {
	    Trace.errln("onException: cannot get Transaction:"+e);
	}
	return;
    }

   /** 
    * This method considers the incoming bean request, and manages the 
    * synchrinization jobs.
    * On the first request we have to load the state of instance
    * May be used for Bean managed or Container managed persistence
    * We implements Option B for the three alternative commit options
    * described p 83 of the specs: 
    * "the container must synchronize the instance's state from the persistent
    * storage at the beginning of the next transaction"
    *
    * @throws RemoteException if a system problem occured. This method
    * must never raise EJBException because it's not catched upper.
    */
    public void considerRequest() throws RemoteException {
	Trace.outln(trace, "EntityContextImpl.considerRequest ");
	
	// If this is not the first request for this Tx, return.
	// Note that if the request is called outside any Tx, it's like
	// if each request were in a separate Tx.
	if (beanCoord != null) {
	    return;
	}

	// Container managed persistence
	if (isContainerManaged) {
	    Trace.outln(trace, "Container managed persistence: call to loadData");
	    try {
		ejbObject.loadData();
	    } catch (EJBException e) {
		onException(); 
		throw new RemoteException("Pb loading data ", e);
	    }
	}

	// Persistence (always called)
	try {
	    enterpriseBean.ejbLoad();  // cf diagramm p86
	} catch( RemoteException e){
	    Trace.errln("EntityContextImpl.considerRequest: " + e);
	    onException(); 
	    throw e;
	} catch(RuntimeException e){
	    Trace.errln("EntityContextImpl.considerRequest: " + e);
	    onException(); 
	    throw new RemoteException("Get " +e.toString()+" thrown by ejbLoad ", e);
	} catch(Error e){
	    Trace.errln("EntityContextImpl.considerRequest : " + e);
	    onException();
	    throw new RemoteException("Get " +e.toString()+" thrown by ejbLoad ", e);
	}
	
	try {
	    // just to remember that next requests are not the first request.
	    beanCoord = tm.getTransaction();
	} catch (SystemException e) {
	    Trace.errln("EntityContextImpl: cannot get Transaction:"+e);
	}
	if (beanCoord == null) {
	    // No transaction => no synchronization.
	    Trace.outln(trace, "EntityContextImpl.considerRequest:  beanCoord == null");
	    return;
	}
	
	// register this object as a synchronization
	try {
	    beanCoord.registerSynchronization(this);
	    Trace.outln(trace, "EntityContextImpl.considerRequest: registerSynchronization= ",this);
	} catch (Exception e) {
	    Trace.errln("EntityContextImpl.considerRequest: registerSynchronization(this) : "+e);
	    throw new RemoteException("Container cannot register synchronization Object", e);
	}
    }


    /**
     * Tests if the instance is marked to be removed
     *
     * @return false if the instance must be removed
     *
     */
    public boolean isMarkedRemoved() {
	return ismarkedremoved;
    }

    /**
     * Set the flag marked to be removed
     *
     * @param flg - true if a remove occurs on this instance
     *
     */
    public void setMarkedRemoved(boolean flg) {
	ismarkedremoved = flg;
    }

    /** 
     * reset the state of the  object in order to be reusable
     */
    public void reset() {
	Trace.outln(trace, "SynchroImpl reset");
	beanCoord = null;
	ejbObject = null;
	ismarkedremoved = false;
    }

}
