/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): Robert Krueger (bug fix)__________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.container;

import java.util.Hashtable;
import java.util.Stack;
import java.util.EmptyStackException;
import javax.transaction.Transaction;
import org.objectweb.jonas.common.Trace;


/**
 * This class handles contexts associated to EJBObject access from multiple 
 * transactions
 * This class is responsible to return the correct context for a given transaction scope
 * It handles a hastable(transaction id/ EntityContextImpl) for the current transaction
 *  and handles a stack of free EntityContextImpls.
 */

public class IContextTable {

    private static final int trace = Trace.DB_21;

    // Hastable for handling association transaction/EntityContextImpl
    // notxec is used because you cannot have a null key in hashtable.
    protected Hashtable ht = new Hashtable();
    protected EntityContextImpl notxec = null;

    // To get access to the pool
    protected JEntityHome home;

   /**
     * constructor
     */

    public IContextTable(JEntityHome home){
	this.home = home;
    }

    /**
     * Associates the specific EntityContextImpl object to the transaction
     *
     * @param tx - the Transaction object
     * @param ic - the EntityContextImpl Object
     */
    public synchronized void put(Transaction tx, EntityContextImpl ic) {

	Trace.outln(trace, "IContextTable.put");

	if (tx != null) {
	    ht.put(tx, ic);
	} else {
	    notxec = ic;
	}
    }


    /**
     * Returns the EntityContextImpl object corresponding to the current transaction scope.
     *
     * @param tx - the Transaction object (transaction scope)
     *
     * @return the EntityContextImpl object for this transaction if it exists, null if not.
     *
     */
    public synchronized EntityContextImpl get(Transaction tx) {

	Trace.outln(trace, "IContextTable.get");

	EntityContextImpl ret = null;

	if (tx != null) {
	    ret = (EntityContextImpl) ht.get(tx);
	} else {
	    ret = notxec;
	}

	return ret;
    }

    /**
     * Releases an EntityContextImpl when  a transaction is completed
     * or when object is removed.
     * the EntityContextImpl is removed from the hastable and pushed in the 
     * pooledinstances  stack.
     *
     * @param tx - the Transaction object (transaction scope)
     *
     */
    
    public synchronized void release(Transaction tx) throws Exception {

	Trace.outln(trace, "IContextTable.release");

	// remove it from the hashtable
	EntityContextImpl ec = null;
	if (tx != null) {
	    ec = (EntityContextImpl) ht.remove(tx);
	} else {
	    // We can go here during object removing only.
	    Trace.outln(trace, "IContextTable.release: release instance outside tx");
	    ec = notxec;
	    notxec = null;
	}

	// reinit objects
	ec.reset();

	// push instance on the pool of instances available
	home.toThePool(ec);
    }
}
