/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.container;

import java.rmi.RemoteException;
import java.rmi.NoSuchObjectException;
import java.util.Vector;
import javax.ejb.RemoveException;
import javax.ejb.EnterpriseBean;
import javax.ejb.EntityBean;
import javax.ejb.EntityContext;
import javax.ejb.Handle;
import javax.ejb.EJBObject;
import javax.naming.Context;
import javax.transaction.NotSupportedException;
import javax.transaction.RollbackException;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionRequiredException;
import javax.transaction.TransactionRolledbackException;

import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.deployment.api.MethodDesc;
import org.objectweb.jonas.jtm.Current;


/**
 * JBeanEntity is the base class for all Entity Beans.
 */
abstract public class JBeanEntity extends JBean {

    private final static int trace = Trace.DB_17;

    /**
     * persistence is container managed
     * @serial
     */
    private boolean isContainerManaged;

    /**
     * @serial
     */
    protected IContextTable ict;

    /*
     * Those abstract methods are implemented in the JOnASEntityXXX class (generated by GenIC)
     * They are defined abstract here, because there are called in EntitySynchroImpl.
     */
    abstract public void createData() throws RemoteException;
    abstract public void loadData()   throws RemoteException;
    abstract public void storeData()  throws RemoteException;
    abstract public void removeData() throws RemoteException;
    abstract public boolean isModifiedData(EntityBean bean);



    /**
     * constructor
     */
    public JBeanEntity(JEntityHome home, boolean cmp) throws RemoteException {
	super(home); 
	isContainerManaged = cmp;
	ict = new IContextTable(home);
    }

    // --------------------------------------------------------------------------
    // EJBObject implementation
    //
    // getEJBHome	in JBean
    // getHandle	genearated
    // getPrimaryKey	ok
    // isIdentical	in JBean
    // remove		ok
    // --------------------------------------------------------------------------

     /**
     * Obtains the primary key of the EJB object.
     *
     * @return    The EJB's object primary key.
     * @exception RemoteException: Thrown when the method failed due to a system-level failure. 
     *
     */
    public Object getPrimaryKey() throws RemoteException{
	if( nosuchobject){
	    throw new NoSuchObjectException("EJBObject already removed");
	}
	return primarykey;
    }
 
    /**
     * Removes the EJB object.
     *
     * @exception RemoteException: Thrown when the method failed due to a system-level failure. 
     * @exception RemoveException: The enterprise Bean or the container does not allow 
     *            destruction of the object.
     *
     */
    public void remove() throws RemoteException, RemoveException {
	Trace.outln(trace, "JBeanEntity.remove");

	if (nosuchobject)
	    throw new NoSuchObjectException("Object Removed");

	Transaction tx = null;
	try {
	    tx = tm.getTransaction();
	} catch(SystemException e) {
	    Trace.errln("remove: "+e);
	    throw new RemoveException("Cannot get transaction");
	}
	EntityContextImpl ec = getContext(tx);

	// JNDI context must be OK. (see EJB spec. p.60)
	home.setComponentContext();

	try {
	    ec.getInstance().ejbRemove();
	} catch(RemoteException e) {
	    Trace.errln("Get the following exception on ejbRemove:"+e);
	    throw new RemoveException("RemoteException on ejbRemove");
	} catch(RuntimeException e) {
	    Trace.errln("Get the following exception on ejbRemove:"+e);
	    throw new RemoveException("RuntimeException on ejbRemove");
	} catch(Error e) {
	    Trace.errln("Get the following exception on ejbRemove:"+e);
	    throw new RemoveException("Error Exception on ejbRemove");
	}
	if (isContainerManaged) {
	    Trace.outln(trace, "Container managed persistence : call to removeData()");
	    removeData();
	}

	if (tx == null) {
	    try {
		releaseInstance(tx); 
		getHome().removeEJBObject(primarykey);
	    } catch(Exception ex) {
		Trace.errln("releaseInstance fails");
	    }
	    noMoreInUse(true);
	} else {
	    // effective remove is differed to commit time
	    ec.setMarkedRemoved(true);
	}  
    }


    // --------------------------------------------------------------------------
    // Other methods
    // --------------------------------------------------------------------------

    /** 
     * Creates a new Entity Context for the proper transactional context
     * used by getContext
     *
     * @param tx - Transaction object corresponding to the current transaction
     * @return the Entity Context object
     */
    public synchronized EntityContextImpl createContext(Transaction tx) throws RemoteException {
	Trace.outln(Trace.DB_21, "JBeanEntity.createContext(tx)");

	String cn = null;
	Class c  = null;
	EntityContextImpl ec = null;
	
	try {
	    cn = home.getEnterpriseBeanClassName();

	    // load the bean with the same class loader of the Home
	    c = home.getClass().getClassLoader().loadClass(cn);

	} catch (ClassNotFoundException e) {
	    Trace.errln("createInstance: Failed to find class " +cn);
	    throw new RemoteException(" Container fails to find class " +cn, e);
	} catch (Exception e) {
	    Trace.errln("createInstance: Failed  " +cn);
	    throw new RemoteException(" createInstance failed " +cn, e);
	}	    

	try {
	    EntityBean eb = (EntityBean)c.newInstance();
	    ec = new EntityContextImpl(this, home, eb, isContainerManaged);

	    // Give entity context to the instance.
	    eb.setEntityContext(ec);

	    ict.put(tx, ec);

	} catch (RemoteException ex) {
	    Trace.errln("Remote Exception caught on setEntityContext");
	    throw ex;
	} catch (Exception e) {
	    Trace.errln("createInstance: "+ cn+" Fails to create a newInstance "+ e);
	    throw new RemoteException("Container cannot create a new instance of " +cn, e);
	}
	return ec;
    }

    
    /** 
     * Creates a new Entity Context for the proper transactional context
     * used only by interposition classes on methods create
     *
     */
     public synchronized EntityContextImpl createContext() throws RemoteException {
	 Transaction tx = null;
	 try {
	     tx = tm.getTransaction();
	 } catch(SystemException e) {
	     Trace.errln("CreateContext: cannot get transaction:"+e);
	 }
	 return createContext(tx);
     }


    /** 
     * Gets the Instance Context for a given transactional context
     * if it does not exist get a new instance from the pool
     *
     * @param tx - Transaction object corresponding to the current transaction
     * @return the EntityContextImpl object
     *
     */
    public synchronized EntityContextImpl getContext(Transaction tx) throws RemoteException {
	Trace.outln(trace, "JBeanEntity.getContext: ",tx);

	EntityContextImpl ec = ict.get(tx);
	if (ec == null) {
	    ec = getHome().fromThePool();
	    if (ec == null) { // Pool Empty create a new 
		ec = createContext(tx);
	    }else{
		ec.setEJBObject(this);
		ict.put(tx, ec);
	    }
	}
	return ec;
    }
    
    /** 
     * Gets the Instance Context for the proper transactional context
     * if it does not exist a new Instance Context is created and activated
     *
     * @return the EntityContextImpl object
     *
     */
    public synchronized EntityContextImpl getContext() throws RemoteException {
	 Transaction tx = null;
	 try {
	     tx = tm.getTransaction();
	 } catch(SystemException e) {
	     Trace.errln("getContext: cannot get transaction:"+e);
	 }
	 return getContext(tx);
    }
    

    /**
     * Releases an instance context on the completion of a transaction
     *
     * @param tx - coordinator of the transaction
     *
     */
    public synchronized void releaseInstance(Transaction tx) throws Exception {
	Trace.outln(trace, "JBeanEntity.releaseInstance ");
	ict.release(tx);
    }


    /**
     * Gets the proper instance of entity bean corresponding to the transactional Context
     * Used inthe generated code.
     * @@return the EnterpriseBean object
     */
    public synchronized EntityBean getBean() throws RemoteException{
	Transaction tx = null;
	try {
	    tx = tm.getTransaction();
	} catch (SystemException e) {
	    Trace.errln("getBean: cannot get Transaction: "+e);
	}
	EntityContextImpl ec = getContext(tx);
	return ec.getInstance();
    }

 

     /**
     * Sets the primary Key of the Entity Bean 
     *
     * @param pk - primary key
     *
     */
    public void setPrimaryKey(Object pk) {
	primarykey = pk;
    }


    /**
     * Returns the JEntityHome
     */
    public JEntityHome getHome() {
	return (JEntityHome)home;
    }

	
    /**
     * This method makes the container's job before the invocation of the business method
     * for a EntityBean
     * Here are treated transaction attributes in order to execute the business.
     *  method in the correct transactional context. A new transaction may be created.
     *  It is important that this method is reentrant, because instance of bean
     *  may be accessed from multiple transactions.
     *
     * @param     transAttribute Transaction attribute of the method that will be invoked.
     * @return    the selected context of the entityBean 
     * 
     * @exception RemoteException: Thrown when the method failed due to a system-level failure. 
     *
     */
    public RequestCtx preinvoke(int txattr) throws RemoteException {

	Trace.outln(trace, "JBeanEntity.preinvoke");
	Trace.outln(trace, "Transaction Attribute = ",MethodDesc.getTxAttributeName(txattr));

	if (nosuchobject)
	    throw new NoSuchObjectException("Object Removed");

	// First of all, get the current transaction
	Transaction currtx = null;
	try {
	    currtx = tm.getTransaction();
	} catch (SystemException e) {
	    Trace.errln("SystemException while getting transaction:\n"+e);
	}

	// Set bean context for JNDI
	Context bnctx = home.setComponentContext();

	// Set the connectionList in threadLocal
	Vector savelist = Current.setConnectionList(home.getConnectionList());

	// Check transaction attribute
	// must be conform to the EJB 1.1 spec ( 11.6.2)
	boolean mustCommit = false;
	Transaction cltx = null;

	switch (txattr) {
 
	case MethodDesc.TX_NOT_SUPPORTED:
	    // The enterprise Bean does not support a global transaction. The container must 
	    //  not invoke the enterprise Bean's method in the scope of a global transaction.

	    try {
		cltx = tm.suspend();
		currtx = null;
	    } catch (SystemException e) {
		Trace.errln("Cannot suspend transaction:\n"+e);
		throw new RemoteException("Cannot suspend transaction", e);
	    }
	    break;

	case MethodDesc.TX_NEVER:
	    // The enterprise Bean must be called without a transaction context.
	    if (currtx != null) {
		Trace.errln("Method with TX_NEVER : the bean must be called without a transaction context");
		throw new RemoteException("Method must be invoked without transaction context");
	    }
	    break;

	case MethodDesc.TX_MANDATORY:
	    // The enterprise Bean requires that the client invocation includes a global 
	    //  transaction scope. The container is responsible for managing transaction 
	    //  boundaries for the enterprise Bean as follow. 
	    // If the caller is associated with a transaction, the execution of the 
	    //  enterprise Bean method will be associated with the caller's transaction. 
	    // If the caller is not associated with a transaction, the container throws the
	    // TransactionRequiredException to the caller.

	    if (currtx == null) {
		throw new TransactionRequiredException("Mandatory attribute = caller must be in a transaction");
	    }
	    break;

	case MethodDesc.TX_REQUIRED:
	    // The enterprise Bean requires that the method be executed in a global transaction. 
  	    // The container is responsible for managing transaction boundaries for the
	    //  enterprise Bean as follow. 
	    // If the caller is associated with a transaction, the execution of the enterprise 
	    //  Bean method will be associated with the caller's transaction. 
	    // If the caller is not associated with a transaction, the container starts a 
	    //  new global transaction, executes the enterprise Bean's method in the scope of 
	    //  the transaction, and commits the transaction when the enterprise Bean's method 
	    //  has completed.

	    if (currtx == null) {
		try {
		    tm.begin();
		    mustCommit = true;
		    currtx = tm.getTransaction();
		} catch (NotSupportedException e) {
		    Trace.errln("Cannot start a transaction: NotSupportedException");
		    throw new RemoteException("Nested Transactions Not Supported");
		} catch (SystemException e) {
		    Trace.errln("Cannot start a transaction:\n"+e);
		    throw new RemoteException("Cannot start a transaction", e);
		}
	    }
	    break;
	    
	case MethodDesc.TX_REQUIRES_NEW:
	    // The enterprise Bean requires that a method is executed in a new global 
	    //  transaction scope. 
	    // The container is responsible for managing transaction boundaries for the 
	    //  enterprise Bean as follow. 
	    // The container starts a new transaction, executes the enterprise Bean's 
	    //  method in the scope of the new transaction, and commits the new transaction 
	    //  when the enterprise Bean's method has completed. 
	    // If the caller is associated with a transaction, the association of the 
	    //  current thread with the caller's transaction is suspended during the execution 
	    //  of the enterprise Bean's method, and resumed when the enterprise Bean's method 
	    //  has completed.

	    try {
		cltx = tm.suspend();
	    } catch (SystemException e) {
		Trace.errln("Cannot suspend transaction:\n"+e);
		throw new RemoteException("Cannot suspend transaction", e);
	    }

	    try {
		tm.begin();
		mustCommit = true;
		currtx = tm.getTransaction();
	    } catch (NotSupportedException e) {
		Trace.errln("Cannot start transaction:"+e);
		throw new RemoteException("Not supported exception");
	    } catch (SystemException e) {
		Trace.errln("Cannot start transaction:"+e);
		throw new RemoteException("Cannot start transaction");
	    }
	    break;
	}

	EntityContextImpl ec = getContext(currtx);
	ec.considerRequest();

	// Build a RequestCtx to save information about this request
	// until postinvoke.
	RequestCtx rctx = new RequestCtx(cltx, mustCommit);
	rctx.setOldJNDICtx(bnctx);
	rctx.setOldConnectionList(savelist);
	rctx.setEJBContext(ec);
	rctx.setTxAttr(txattr);
	return rctx;
    }	


    /**
     * This method is called after the invocation of the business method.
     *
     * @param     rctx context of the current request
     * @exception RemoteException: Thrown when the method failed due to a system-level failure.
     */
    public void postinvoke(RequestCtx rctx) throws RemoteException {
	Trace.outln(trace, "JBeanEntity.postinvoke ");

	// If the method has been called out of the scope of a transaction
	// we have to do the persistence's job 
	Transaction currentTx = null;
	try {
	    currentTx = tm.getTransaction();
	} catch (SystemException e) {
	    Trace.errln("postinvoke: cannot get Transaction"+e);
	}
	if (currentTx == null) {
	    // may raise RemoteException
	    EntityContextImpl ec = (EntityContextImpl) rctx.getEJBContext();
	    ec.storeIfModified();
	}	    	    
	    
	// common part
	doPostInvoke(rctx);

    }
}

