/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.container;

import java.util.Vector;
import java.rmi.RemoteException;
import java.rmi.NoSuchObjectException;
import java.rmi.NoSuchObjectException;
import javax.ejb.EnterpriseBean;
import javax.ejb.Handle;
import javax.ejb.RemoveException;
import javax.ejb.SessionBean;
import javax.ejb.SessionSynchronization;
import javax.ejb.EJBObject;
import javax.naming.Context;
import javax.transaction.NotSupportedException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import javax.transaction.TransactionRequiredException;
import javax.transaction.SystemException;

import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.deployment.api.MethodDesc;
import org.objectweb.jonas.jtm.Current;


/**
 * JBeanSession is the base class for all Session Beans
 */
public class JBeanSession extends JBean {

    private final static int trace = Trace.DB_17;

    /**
     * @serial
     */
    protected boolean isSessionSynchro = false;

    /**
     * @serial
     */
    private boolean isStateFul = false;

    /**
     * @serial
     */
    private boolean isTxBeanManaged = false;

    /**
     * @serial
     */
    protected SessionContextImpl sc = null;


    /**
     * Constructor
     */
    public JBeanSession(JSessionHome home, boolean sync) throws RemoteException {
	super(home); 
	isSessionSynchro = sync;
	isStateFul = home.isStateFul();
	isTxBeanManaged = home.isTxBeanManaged();
    }

    // --------------------------------------------------------------------------
    // EJBObject implementation
    //
    // getEJBHome	in JBean
    // getHandle	ok
    // getPrimaryKey	ok
    // isIdentical	in JBean
    // remove		ok
    // --------------------------------------------------------------------------

    /**
     * Obtains a handle for the EJB object. The handle can be used at later time to re-obtain
     *  a reference to the EHB object, possibly in a different Java Virtual Machine
     *
     * @return    A handle for the EJB object. 
     * @exception RemoteException: Thrown when the method failed due to a system-level failure. 
     *
     */
    public Handle getHandle() throws RemoteException {
	Trace.outln(trace, "JBeanSession.getHandle");
	if( nosuchobject)
	    throw new NoSuchObjectException("EJBObject already removed");
	if (handleEJBObject == null){
	    handleEJBObject = new SessionHandleImpl(this);
	}
	return (Handle) handleEJBObject;
    }

    /**
     * Obtains the primary key of the EJB Object
     * must throw RemoteException for sessionBean
     * (5.5 Spec EJB1.1)
     */
    public Object getPrimaryKey() throws RemoteException{
	Trace.errln("getPrimaryKey: No primary Key for Session Bean");
	throw new RemoteException("No primary Key for Session Bean");
    }
 
    /**
     * Removes the EJB object.
     * It is an error for a client to invoke remove on a session bean participating
     * to a transaction
     * @exception RemoteException: Thrown when the method failed due to a system-level failure. 
     * @exception RemoveException: The enterprise Bean or the container does not allow 
     *            destruction of the object.
     *
     */
    public void remove() throws RemoteException, RemoveException {
	Trace.outln(trace, "JBeanSession.remove");

	if (nosuchobject)
	    throw new NoSuchObjectException("EJBObject already removed");

	if (isSynchronized() && sc.synchro.inTransaction()) {
	   Trace.errln("it's an error to remove a session bean participating in a transaction"); 
	   throw new RemoveException();
	}

	// JNDI context must be OK. (see EJB spec. p.60)
	home.setComponentContext();

	try {
	    sc.getInstance().ejbRemove();
	}catch(RemoteException e){
	    Trace.errln("Get the following exception on ejbRemove");
	    Trace.errln(e);
	    throw new RemoveException();
	}catch(RuntimeException e){
	    Trace.errln("Get the following exception on ejbRemove");
	    Trace.errln(e);
	    throw new RemoveException();
	}catch(Error e) {
	    Trace.errln("Get the following exception on ejbRemove");
	    Trace.errln(e);
	    throw new RemoveException();
	}
	noMoreInUse(true);
    }
 

    // --------------------------------------------------------------------------
    // Other methods
    // --------------------------------------------------------------------------

    /**
     * Tests if a bean have a synchronization object associated.
     * Actually, all StateFul SessionBean are Synchronized.
     */
    public boolean isSynchronized() {
	return isStateFul;
    }

    /**
     * Creates the instance with its possible SessionSynchro and its SessionContext
     *
     * @return the SessionContextImpl object
     */
    public synchronized SessionContextImpl createContext() throws RemoteException {

	Trace.outln(trace, "JBeanSession.createContext");
	String cn = null;
	Class  c  = null;
	SessionSynchroImpl sync = null;

	try {
	    cn = home.getEnterpriseBeanClassName();

	    // load the bean with the same class loader of the Home
	    c = home.getClass().getClassLoader().loadClass(cn);

	} catch (ClassNotFoundException e) {
	    Trace.errln("createInstance: Failed to find class " +cn);
	    throw new RemoteException(" Container fails to find class " +cn, e);
	}

	try {
	    SessionBean sb = (SessionBean)c.newInstance();
	    if (isStateFul) {
		if (isSessionSynchro) {
		    sync = new SessionSynchroImpl(home, (SessionSynchronization) sb);
		} else {
		    sync = new SessionSynchroImpl(home, null);
		}
	    }
	    sc = new SessionContextImpl(this, home, sb, sync);

	    // Give session context to the instance.
	    sb.setSessionContext(sc);
	} catch(RemoteException ex) {
	    Trace.errln("Remote Exception caught on setSessionContext");
	    throw ex;
	} catch (Exception e) {
	    Trace.errln("createInstance: "+ cn+"Fails to create a newInstance "+ e);
	    throw new RemoteException("Container cannot create a new instance of " +cn, e);
	}
	return sc;
    }


   /**
     * This method makes the container's job before the invocation of the business method.
     * Here are treated transaction attributes in order to execute the business.
     *  method in the correct transactional context. A new transaction may be created.
     *
     * @param     transAttribute Transaction attribute of the method that will be invoked.
     * @exception RemoteException: Thrown when the method failed due to a system-level failure. 
     *
     * @return    the context of the SessionBean 
     *
     */
    public RequestCtx preinvoke(int txattr) throws RemoteException {

	Trace.outln(trace, "JBeanSession.preinvoke");
	Trace.outln(trace, "Transaction Attribute = ",MethodDesc.getTxAttributeName(txattr));

	if (nosuchobject)
	    throw new NoSuchObjectException("Object Removed");

	// First of all, get the current transaction
	Transaction currtx = null;
	try {
	    currtx = tm.getTransaction();
	} catch (SystemException e) {
	    Trace.errln("SystemException while getting transaction:\n"+e);
	}

	// Set bean context for JNDI
	Context bnctx = home.setComponentContext();

	// Set the connectionList in threadLocal
	Vector savelist = Current.setConnectionList(home.getConnectionList());

	// Check transaction attribute
	boolean mustCommit = false;
	Transaction cltx = null;

	switch (txattr) {

	case MethodDesc.TX_BEAN_MANAGED:
	    try {
		cltx = tm.suspend();
		if (sc.beanAssociatedTransaction != null) {
		    tm.resume(sc.beanAssociatedTransaction);
		} 
	    } catch (SystemException e) {
		Trace.errln("Exception in suspend/resume:"+e);
		throw new RemoteException("preinvoke failed");
	    }
	    break;

	case MethodDesc.TX_NOT_SUPPORTED:
	    // The enterprise Bean method does not support a global transaction. The container must 
	    //  not invoke the enterprise Bean's method in the scope of a global transaction.
		
	    // If the SessionBean is already participating in a Transaction, it is an error
	    // to invoke a method in a different transaction context or no transaction context.
	    if (isSynchronized() && sc.synchro.inTransaction()) {
		Trace.errln("SessionBean is already participating in a Transaction: error.");
		throw new RemoteException("Session bean can participate in at most one tx at a time");
	    }

	    // Suspend any incoming transaction
	    try {
		cltx = tm.suspend();
	    } catch (SystemException e) {
		Trace.errln("Exception in suspend:"+e);
		throw new RemoteException("preinvoke failed");
	    }
	    break;

	case MethodDesc.TX_NEVER:
	    // The enterprise Bean must be called without a transaction context.
	    if (isSynchronized() && sc.synchro.inTransaction()) {
		Trace.errln("SessionBean is already participating in a Transaction: error.");
		throw new RemoteException("Method must be invoked without transaction context");
	    }
	    if (currtx != null) {
		Trace.errln("Method with TX_NEVER : the bean must be called without a transaction context");
		throw new RemoteException("Method must be invoked without transaction context");
	    }
	    break;

	case MethodDesc.TX_MANDATORY:
	    // The enterprise Bean requires that the client invocation includes a global 
	    //  transaction scope. The container is responsible for managing transaction 
	    //  boundaries for the enterprise Bean as follow. 
	    // If the caller is associated with a transaction, the execution of the 
	    //  enterprise Bean method will be associated with the caller's transaction. 
	    // If the caller is not associated with a transaction, the container throws the
	    // TransactionRequiredException to the caller.
	
	    if (currtx == null) {
		throw new TransactionRequiredException("Mandatory attribute = caller must be in a transaction");
	    }
	    break;
		
	case MethodDesc.TX_REQUIRED:
	    // The enterprise Bean requires that the method be executed in a global transaction. 
	    // The container is responsible for managing transaction boundaries for the
	    //  enterprise Bean as follow. 
	    // If the caller is associated with a transaction, the execution of the enterprise 
	    //  Bean method will be associated with the caller's transaction. 
	    // If the caller is not associated with a transaction, the container starts a 
	    //  new global transaction, executes the enterprise Bean's method in the scope of 
	    //  the transaction, and commits the transaction when the enterprise Bean's method 
	    //  has completed.
	    if (currtx == null) {
		try {
		    tm.begin();
		    mustCommit = true;
		} catch (NotSupportedException e) {
		    Trace.errln("preinvoke: cannot start transaction:"+e);
		    throw new RemoteException("Nested Transactions Not Supported");
		} catch (SystemException e) {
		    Trace.errln("Cannot start a transaction:\n"+e);
		    throw new RemoteException("Cannot start a transaction", e);
		}
	    } else {
		Trace.outln(trace, "    Method called inside a transaction scope");
	    }	
	    break;
		
	case MethodDesc.TX_REQUIRES_NEW:
	    // The enterprise Bean requires that a method is executed in a new global 
	    //  transaction scope. 
	    // The container is responsible for managing transaction boundaries for the 
	    //  enterprise Bean as follow. 
	    // The container starts a new transaction, executes the enterprise Bean's 
	    //  method in the scope of the new transaction, and commits the new transaction 
	    //  when the enterprise Bean's method has completed. 
	    // If the caller is associated with a transaction, the association of the 
	    //  current thread with the caller's transaction is suspended during the execution 
	    //  of the enterprise Bean's method, and resumed when the enterprise Bean's method 
	    //  has completed.
		
	    // If the SessionBean is already participating in a Transaction, it is an error
	    // to invoke a method in a different transaction context or no transaction context.
	    if (isSynchronized() && sc.synchro.inTransaction()) {
		Trace.errln("SessionBean is already participating in a Transaction: error.");
		throw new RemoteException("Session bean can participate in at most one tx at a time");
	    }
		
	    try {
		cltx = tm.suspend();
	    } catch (SystemException e) {
		Trace.errln("Cannot suspend transaction:\n"+e);
		throw new RemoteException("Cannot suspend transaction", e);
	    }

	    try {
		tm.begin();
		mustCommit = true;
	    } catch (NotSupportedException e) {
		Trace.errln("Cannot start transaction:"+e);
		throw new RemoteException("Not supported exception");
	    } catch (SystemException e) {
		Trace.errln("Cannot start transaction:"+e);
		throw new RemoteException("Cannot start transaction");
	    }
	    break;
	}

	// Synchronization 
	// not useful if BEAN_MANAGED, NOT_SUPPORTED, NEVER
	if (isSynchronized()) 
	    sc.synchro.considerRequest();

	// Build a RequestCtx to save information about this request
	// until postinvoke.
	RequestCtx rctx = new RequestCtx(cltx, mustCommit);
	rctx.setOldJNDICtx(bnctx);
	rctx.setOldConnectionList(savelist);
	rctx.setEJBContext(sc);
	rctx.setTxAttr(txattr);
	return rctx;
    }


    /**
     * This method is called after the invocation of the business method.
     *
     * @param     rctx context of the current request
     * @exception RemoteException: Thrown when the method failed due to a system-level failure.
     */
    public void postinvoke(RequestCtx rctx) throws RemoteException {
	Trace.outln(trace, "JBeanSession.postinvoke");

	if (rctx.getTxAttr() == MethodDesc.TX_BEAN_MANAGED) {
	    // tx initiated by the bean must be suspended in any case
	    Transaction tx = null;
	    try {
		tx = tm.suspend();
	    } catch(SystemException e) {
		Trace.errln("postinvoke: cannot suspend transaction:"+e);
	    }
	    if (tx != null) {
		if (isStateFul) {
		    // keep it for resume in next request
		    sc.beanAssociatedTransaction = tx;
		} else {
		    // Stateless Session beans are not allowed to retain a
		    // transaction between 2 requests.
		    try {
			Trace.errln("Rollback stateless session bean transaction");
			tx.rollback();
		    } catch (SystemException e) {
			Trace.errln("postinvoke: cannot rollback transaction:"+e);
		    }
		}
	    }
	}  	    
	    
	// common part
	doPostInvoke(rctx);
    }


    /**
     * Gets the proper instance of session bean 
     * @return the SessionBean object
     */
    public synchronized SessionBean getBean() throws RemoteException {
	return sc.getInstance();
    }

}

