/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */



package org.objectweb.jonas.container; 
 
import java.util.Hashtable;
import java.util.Stack;
import java.util.EmptyStackException;
import java.rmi.RemoteException;
import javax.ejb.EJBObject;
import javax.ejb.EntityBean;
import javax.ejb.FinderException;
import javax.ejb.RemoveException;
import javax.ejb.Handle;
import javax.naming.Context;
import javax.sql.DataSource;
import javax.transaction.NotSupportedException;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionRequiredException;

import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.deployment.api.EntityDesc;
import org.objectweb.jonas.deployment.api.MethodDesc;
import org.objectweb.jonas.jtm.Current;


/**
 * Class inherited by all the Home object for entity beans
 */
abstract public class JEntityHome extends JBeanHome {
    
    private final static int trace = Trace.DB_16;

    /**
     * @serial
     */
    private Hashtable ejbobjects = new Hashtable();

    /**
     * @serial
     */
    private Object datasource = null;

    /**
     * @serial
     */
    private EntityBean available  = null;

    /**
     * @serial
     */
    protected String primaryKeyClassName = null;

    /**
     * @serial
     */
    protected Stack pooledinstances = new Stack();



    /**
     * constructor
     * The DataSourceImpl param value may be null
     */
    public JEntityHome(EntityDesc desc, String beanName, Object ds) 
	throws RemoteException {

	super(desc, beanName);

	Trace.outln(trace, "JEntityHome constructor for ",beanName);

	this.datasource = ds;
	this.primaryKeyClassName = desc.getPrimaryKeyClass().getName();

	JBeanMetaData bbdm = new JBeanMetaData(this,
					       desc.getHomeClass().getName(),
					       desc.getRemoteClass().getName(),
					       primaryKeyClassName,
					       false,
					       false);
	this.setEJBMetaData(bbdm);
    }


    /**
     * Maps the  specified primary key to the EJBObject (only id the key does not exist)
     *
     * @param pk - the primary key 
     * @param ejbObject - the EJBObject 
     *
     */
    public synchronized void putEJBObject(Object pk, JBeanEntity ejbObject) {

	Trace.outln (trace, "JEntityHome.putEJBObject");

	// The key already exists -> return.
	if (ejbobjects.containsKey(pk))
	    return;

	ejbobjects.put(pk, ejbObject);
	try {
	    Transaction tx = tm.getTransaction();
	    if (tx != null) {
		HomeSynchro hs = new HomeSynchro(this, pk);
		// register this object as a synchronization
		tx.registerSynchronization(hs);
		Trace.outln(trace, "JEntityHome.putEJBObject: registerSynchronization ");
	    }
	} catch (javax.transaction.RollbackException e) {
	    Trace.errln("JEntityHome.putEJBObject: Transaction is rolled back");
	    ejbobjects.remove(pk);
	} catch (Exception e) {
	    Trace.errln("JEntityHome.putEJBObject: cannot register synchronization Object: "+e);
	    ejbobjects.remove(pk);
	}
    }


    /**
     * Returns the EJBObject to which the specified primary key is mapped
     *
     * @param  pk - the primary key 
     * @return the value to which the key is mapped in this hashtable; 
     *         null if the key is not mapped to any value in this hashtable. 
     *
     */
    public JBeanEntity getEJBObject(Object pk) {

	Trace.outln (trace, "JEntityHome.getEJBObject ",pk);

	JBeanEntity bb = (JBeanEntity) ejbobjects.get(pk);
	if (bb == null) {
	    Trace.outln (trace, "There is no EJBObject for this primary key ",pk);
	}
	// Do not check isMarkedRemoved() anymore: not really useful and it
	// leads to memory leaks
	return bb;
    }
  

    /**
     * 
     */
    public synchronized void removeEJBObject(Object pk) {

	Trace.outln (trace, "JEntityHome.removeEJBObject");
	Object obj = ejbobjects.remove(pk);
    }


    /**
     * Returns the DataSource associated to the bean
     * JOnAS dependant, used mainly by generated classes.
     *
     * @return  the DataSource associated to the bean or null if there is no associated DataSource
     *
     */
    public Object getDataSource() {
	Trace.outln (trace, "JEntityHome.getDataSource");
	if (datasource == null) {
	    Trace.errln("JOnAS configuration error!");
	    Trace.errln("    JEntityHome.getDataSource() returning null");
	}
	return datasource;
    }


    /**
     * Get a pseudo instance of the bean for finder methods
     * This instance is never destroyed and can be reused.
     */
    public EntityBean getAvailable() throws RemoteException {

	if (available != null) {
	    return available;
	}
	String cn = null;
	Class c = null;
	try {
	    cn = getEnterpriseBeanClassName();
	    c = Class.forName(cn);
	} catch (ClassNotFoundException e) {
	    Trace.errln("createInstance: Failed to find class "+cn);
	    throw new RemoteException(" Container fails to find class "+cn, e);
	}

	try {
	    available = (EntityBean)c.newInstance();
	} catch (Exception e) {
	    Trace.errln("createInstance: "+ cn+" Fails to create a newInstance "+ e);
	    throw new RemoteException("Container cannot create a new instance of "+cn, e);
	}

	EntityContextImpl ctx = new EntityContextImpl(this);

	// Every instance should have a correct entity context
	// - JNDI context must be OK. (see EJB spec. p.60)
	// -> this should not be necessary if preinvoke were called before 
	//    getAvailable... (See GenIC)
	setComponentContext();
	available.setEntityContext(ctx);

	return available;
    }

    /**
     * Removes an EJB object identified by its primary key.
     */
    public void remove(Object primaryKey) throws RemoteException, RemoveException {

	Trace.outln (trace, "JEntityHome.remove(pk):");
	try {
	    JBeanEntity bb = (JBeanEntity)findByPK(primaryKey);
	    bb.remove();
	} catch(javax.ejb.FinderException ex){
	    throw new RemoveException("ObjectNotFound");
	}
    }


    /**
     * Removes an EJB object identified by its handle. 
     */
    public void remove(Handle handle) throws RemoteException, RemoveException {

	Trace.outln(trace,"remove(handle)");

	EJBObject ejbobj = handle.getEJBObject();
	ejbobj.remove();
    }

    /**
     * consider transaction for create and finder methods
     * Entity beans must be container managed for transations,
     * so we needn't to test this.
     */
    protected RequestCtx preinvoke(int txattr) throws RemoteException {

	Trace.outln(trace, "JEntityHome.preinvoke");
	Trace.outln(trace, "Transaction Attribute = " ,MethodDesc.getTxAttributeName(txattr));

	// First of all, get the current transaction
	Transaction currtx = null;
	try {
	    currtx = tm.getTransaction();
	} catch (SystemException e) {
	    Trace.errln("SystemException while getting transaction:\n"+e);
	}

	// Check transaction attribute
	// must be conform to the EJB 1.1 spec ( 11.6.2)
	boolean mustCommit = false;
	Transaction cltx = null;

	// Raises exception if "Never" and we are in a transaction
	if (txattr == MethodDesc.TX_NEVER && currtx != null) {
	    Trace.outln(Trace.DB_31, "Never and transaction not null");
	    throw new RemoteException("Never attribute = caller must not be in a transaction");
	}

	// Raises exception if "Mandatory" and we are not in a transaction
	if (txattr == MethodDesc.TX_MANDATORY && currtx == null) {
	    Trace.outln(Trace.DB_31, "Mandatory and not in transaction");
	    throw new TransactionRequiredException("Mandatory attribute = caller must be in a transaction");
	}

	// Suspend transaction if "NotSupported" or "RequiresNew"
	if (currtx != null &&
	    (txattr == MethodDesc.TX_REQUIRES_NEW || txattr == MethodDesc.TX_NOT_SUPPORTED)) {
	    try {
		cltx = tm.suspend();
	    } catch (SystemException e) {
		Trace.errln("Cannot suspend transaction:\n"+e);
		throw new RemoteException("Cannot suspend transaction", e);
	    }
	}

	// Start a new transaction in 2 cases:
	// - "RequiresNew" attribute
	// - "Required" and no current transaction
	// This transaction will be closed at postinvoke.
	if (txattr == MethodDesc.TX_REQUIRES_NEW || 
	    (txattr == MethodDesc.TX_REQUIRED && currtx == null)) {
	    try {
		tm.begin();
		mustCommit = true;
	    } catch (NotSupportedException e) {
		Trace.errln("Cannot start a transaction: NotSupportedException");
		throw new RemoteException("Nested Transactions Not Supported");
	    } catch (SystemException e) {
		Trace.errln("Cannot start a transaction:\n"+e);
		throw new RemoteException("Cannot start a transaction: SystemException");
	    }
	}

	// Build a RequestCtx to save information about this request
	// until the postinvoke time.
	RequestCtx rctx = new RequestCtx(cltx, mustCommit);

	// Set bean context for JNDI
	rctx.setOldJNDICtx(setComponentContext());

	// Set the connectionList in threadLocal
	// and save old value in RequestContext
	rctx.setOldConnectionList(Current.setConnectionList(connectionList));

	return rctx;
    }

    

    /**
     * Generic "findByPrimaryKey" method called from the container
     */
    abstract public Object findByPK(java.lang.Object pk) 
	throws java.rmi.RemoteException, javax.ejb.FinderException;


    public EntityContextImpl fromThePool(){
	Trace.out(Trace.DB_21, "JEntityHome.fromThePool: ");
	EntityContextImpl ec = null;
	try{
	    ec = (EntityContextImpl)pooledinstances.pop();
	}catch(EmptyStackException ex){
	    return ec;
	}
	return ec;
    }

    public void toThePool(EntityContextImpl ec){
	Trace.outln(Trace.DB_21, "JEntityHome.toThePool");
	EntityContextImpl eci  = (EntityContextImpl)pooledinstances.push(ec);
    }


}
