/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999,2000 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * 00/01/08 Markus Fritz
 *          make connection pooling thread safe (with same tx)
 *          introduce MaxAge for connections
 *          optimize pooling
 *          add some 'paranoia code': test conns before returning
 *	    them from the pool.
 * 00/18/04 Jun Inamori (j-office@osa.att.ne.jp)
 *          For closing correctly all connections at server shutdown.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.dbm;

import java.io.Serializable;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.Hashtable;
import java.util.Vector;
import javax.naming.Context;
import javax.naming.NamingException;
import javax.naming.Reference;
import javax.naming.Referenceable;
import javax.naming.StringRefAddr;
import javax.sql.ConnectionEvent;
import javax.sql.ConnectionEventListener;
import javax.sql.DataSource;
import javax.sql.XAConnection;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.xa.XAResource;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.jdbc_xa.XADataSourceImpl;
import org.objectweb.jonas.jtm.CompletedTransactionListener;
import org.objectweb.jonas.jtm.Current;

/**
 * JOnAS generic implementation of DataSource to manage an underlying
 * JDBC-XA driver (XADataSource, XAConnection, XAResource, ...)
 * This class acts as a factory for Connection objects.
 * Its main goal is to manage a pool of XAConnection objects, to avoid 
 * closing the actual Connection on DBMS, and to deal with distributed
 * transactions.
 * This DataSource is registered in JNDI (=> implements Referenceable)
 */
public class ConnectionManager implements DataSource, 
					  Referenceable, 
					  Serializable,
					  CompletedTransactionListener,
                                          ConnectionEventListener {

    private static final int trace = Trace.DB_8;

    // The static table of all the ConnectionManager objects
    // Each ResourceManager has its own ConnectionManager (i.e. DataSource)
    private static Hashtable cmList = new Hashtable();

    /**
     * unique Current object, kept here for efficiency
     * @serial
     */
    private Current current = Current.getCurrent();

    /**
     * underlaying XA DataSource
     * @serial
     */
    private XADataSourceImpl xads = null;

    /**
     * The pool of XAConnections managed by this Connection Manager
     * @serial
     */
    private Pool pool = new Pool(this);

    // ----------------------------------------------------------
    // properties
    // ----------------------------------------------------------

    /**
     * @serial for JNDI
     */
    private String dataSourceName = null;
    /**
     * @serial url for database
     */
    private String url = null;
    /**
     * @serial JDBC driver Class
     */
    private String className = null;
    /**
     * @serial default user
     */
    private String userName = null;
    /**
     * @serial default passwd
     */
    private String password = null;

    // -----------------------------------------------------------------
    // Constructors
    // -----------------------------------------------------------------

    /**
     * Constructor for Factory
     */
    public ConnectionManager() {
        Trace.outln(trace, "ConnectionManager.ConnectionManager()");

        // Create the underlaying JDBC-XA driver
        // We do not use JNDI here. (should we do ?)
        xads = new XADataSourceImpl();
    }

    // -----------------------------------------------------------------
    // Properties
    // -----------------------------------------------------------------
    public String getDataSourceName() {
        return dataSourceName;
    }
    public void setDataSourceName(String s) {
        Trace.outln(trace, "ConnectionManager.setDataSourceName: "+s);
        dataSourceName = s;
        xads.setDataSourceName(s); // for getRMID!

        // Add it to the list
        cmList.put(s, this);
    }

    public String getUrl() {
        return url;
    }
    public void setUrl(String s) {
        Trace.outln(trace, "ConnectionManager.setUrl(String "+s+")");
        url = s;
        // Forward it to JDBC-XA driver
        xads.setUrl(s);
    }

    public String getClassName() {
        return className;
    }
    public void setClassName(String s) {
        Trace.outln(trace, "ConnectionManager.setClassName(String "+s+")");
        className = s;
        // Forward it to JDBC-XA driver
        xads.setClassName(s);
    }

    public String getUserName() {
        return userName;
    }
    public void setUserName(String s) {
        Trace.outln(trace, "ConnectionManager.setUserName(String "+s+")");
        userName = s;
        // Forward it to JDBC-XA driver
        xads.setUserName(s);
    }

    public String getPassword() {
        return password;
    }
    public void setPassword(String s) {
        Trace.outln(trace, "ConnectionManager.setPassword(String "+s+")");
        password = s;
        // Forward it to JDBC-XA driver
        xads.setPassword(s);
    }

    /**
     * Configure the Connection pool. Called by the Container at init.
     * Configuration can be set in datasource.properties files.
     */
    public void poolConfigure(String connchecklevel,
			      String connmaxage,
			      String connmaxthreads,
			      String connexcltimeout,
			      String connteststmt) {
	pool.jdbcConnCheckLevel = (new Integer(connchecklevel)).intValue();
	pool.jdbcConnMaxAge = (new Integer(connmaxage)).intValue();
	pool.jdbcConnMaxThreads = (new Integer(connmaxthreads)).intValue();
	pool.jdbcConnExclTimeout = (new Integer(connexcltimeout)).intValue();
	pool.jdbcTestStmt = connteststmt;
	Trace.outln(trace, "ConnectionManager configured with:");
	Trace.outln(trace, "   jdbcConnCheckLevel  = "+pool.jdbcConnCheckLevel);
	Trace.outln(trace, "   jdbcConnMaxAge      = "+pool.jdbcConnMaxAge);
	Trace.outln(trace, "   jdbcConnMaxThreads  = "+pool.jdbcConnMaxThreads);
	Trace.outln(trace, "   jdbcConnExclTimeout = "+pool.jdbcConnExclTimeout);
	Trace.outln(trace, "   jdbcTestStmt        = "+pool.jdbcTestStmt);
	// set times in milliseconds
	pool.jdbcConnMaxAge = pool.jdbcConnMaxAge * 60L * 1000L;
	pool.jdbcConnExclTimeout = pool.jdbcConnExclTimeout * 1000L;
    }

    // -----------------------------------------------------------------
    // DataSource implementation
    // -----------------------------------------------------------------

    /**
     * Attempt to establish a database connection.
     *
     * @return a Connection to the database
     *
     * @exception java.sql.SQLException - if a database-access error occurs.
     */
    public Connection getConnection() throws SQLException {
        Trace.outln(trace, "ConnectionManager.getConnection()");
        return getConnection(userName, password);
    }

    /**
     * Attempt to establish a database connection.
     * The pool management must be "synchronized".
     *
     * @param user - the database user on whose behalf the Connection is being made
     * @param password - the user's password
     *
     * @return a Connection to the database
     *
     * @exception SQLException - if a database-access error occurs.
     */
    public synchronized Connection getConnection(String username, String password) 
        throws SQLException {

        Trace.outln(trace, "ConnectionManager.getConnection("+username+")");

        XAConnection xac = null;
        Transaction tx = null;
        try {
            tx = current.getTransaction();
            Trace.outln(trace, "\tTx = " + String.valueOf(tx));
	} catch (NullPointerException n) {
	    // current is null: we are not in EJBServer.
	    Trace.errln("ConnectionManager: should not be used outside an EJBServer");
        } catch (SystemException e) {
            Trace.errln("ConnectionManager: getTransaction failed:"+e);
        }

        // Check if available XAConnection in the pool for this user

        PoolItem pi = pool.openConnection(username, (Transaction) tx);
        if (pi == null) {
            Trace.outln(trace, "ConnectionManager: No more XAConnection in the pool");
            // Get a new XAConnection
            xac = xads.getXAConnection(username, password);
            // Put it in the pool as in-use

	    pi = pool.newConnection(xac, username, (Transaction) tx);
            // Register me as a ConnectionEventListener
            xac.addConnectionEventListener(this);
        } else {
            // We found a reusable XAConnection
            // We are still registered as ConnectionEventListener
            Trace.outln(trace, "ConnectionManager: Reuse an XAConnection from the pool");
            xac = pi.xaConn;
        }

        synchronized (xac) {
	    Connection ret = xac.getConnection();
	    // Enlist XAResource if we are actually in a transaction
	    if (tx != null) {
		if (pi.open == 1) { // Only if first/only thread
		    // We are inside a transaction: no autocommit!
		    ret.setAutoCommit(false);
		    try {
			tx.enlistResource(xac.getXAResource());
		    } catch (Exception e) {
			Trace.errln("Cannot enlist XAResource:"+e);
			// What shall we do here ? XXX
		    }
		}
	    } else {
		// We are not in a transaction yet: we just make a pre-registration.
		ret.setAutoCommit(true);    // in case we do not start a Tx
		current.addConnection(xac);
	    }

	    // return a Connection object
	    return ret;
	}
    }

    /**
     * Get the log writer for this data source. 
     * 
     * The log writer is a character output stream to which all logging and 
     * tracing messages for this data source object instance will be printed. 
     * This includes messages printed by the methods of this object, messages 
     * printed by methods of other objects manufactured by this object, and so on.
     * Messages printed to a data source specificlog writer are not printed to the 
     * log writer associated with the Drivermanager class. When a data
     * source object is created the log writer is initially null, in other words, 
     * logging is disabled.
     *
     * @return the log writer for this data source, null if disabled
     *
     * @exception SQLException - if a database-access error occurs.
     */
    public java.io.PrintWriter getLogWriter() throws SQLException {
        Trace.outln(trace, "ConnectionManager.getLogWriter()");
        return xads.getLogWriter();
    }


    /**
     * Set the log writer for this data source. 
     * 
     * The log writer is a character output stream to which all logging and 
     * tracing messages for this data source object instance will be printed. 
     * This includes messages printed by the methods of this object, messages printed
     * by methods of other objects manufactured by this object, and so on. 
     * Messages printed to a data source specific log writer are not printed to 
     * the log writer associated with the Drivermanager class. When a data
     * source object is created the log writer is initially null, in other words, 
     * logging is disabled.
     * 
     * @param out - the new log writer; to disable, set to null
     *
     * @exception SQLException - if a database-access error occurs.
     */
    public void setLogWriter(java.io.PrintWriter out) throws SQLException {
        Trace.outln(trace, "ConnectionManager.setLogWriter()");
        xads.setLogWriter(out);
    }


    /**
     * Sets the maximum time in seconds that this data source will wait while 
     * attempting to connect to a database.
     * A value of zero specifies that the timeout is the default system timeout 
     * if there is one; 
     * otherwise it specifies that there is no timeout. 
     * When a data source object is created the login timeout is initially zero.
     *
     * @param seconds - the data source login time limit
     *
     * @exception SQLException - if a database access error occurs.
     */
    public void setLoginTimeout(int seconds) throws SQLException {
        Trace.outln(trace, "ConnectionManager.setLoginTimeout(int"+seconds+")");
        xads.setLoginTimeout(seconds);
    }

    /**
     * Gets the maximum time in seconds that this data source can wait while attempting 
     * to connect to a database.
     * A value of zero means that the timeout is the default system timeout if there is one; 
     * otherwise it means that there is no timeout. When a data source object is created 
     * the login timeout is initially zero.
     *
     * @return the data source login time limit
     *
     * @exception SQLException - if a database access error occurs.
     */
    public int getLoginTimeout() throws SQLException {
        Trace.outln(trace, "ConnectionManager.getLoginTimeout()");
        return xads.getLoginTimeout();
    }

    // -----------------------------------------------------------------
    // ConnectionEventListener implementation
    // -----------------------------------------------------------------

    /**
     * Invoked when the application calls close() on its representation 
     * of the connection.
     * @param event - an event object describing the source of the event
     */
    public void connectionClosed(ConnectionEvent event) {
        Trace.outln(trace, "ConnectionManager.connectionClosed(ConnectionEvent)");

        XAConnection xac = (XAConnection) event.getSource();
        synchronized (xac) {
	    // First of all, always mark the Connection as closed.
            PoolItem pi = pool.closeConnection(xac);
	    if (pi == null) {
		// was already closed: just return.
		return;
	    }
	    // Check if inside a transaction 
	    Transaction pitx = (Transaction)pi.tx;
	    Transaction tx = null;
	    try {
		tx = current.getTransaction();
		Trace.outln(trace, "\tTx = " + String.valueOf(tx));
	    } catch (NullPointerException n) {
		// current is null: we are not in EJBServer.
		Trace.errln("ConnectionManager: should not be used outside an EJBServer");
	    } catch (SystemException e) {
		Trace.errln("ConnectionManager: getTransaction failed:"+e);
	    }

	    // delist Resource if in transaction
	    // We must keep the connection till commit or rollback
	    if (tx != null) {
		if (pi.open < 1) try {
                    tx.delistResource(xac.getXAResource(), XAResource.TMSUCCESS);
		} catch (Exception e) {
		    Trace.errln("Exception in connectionClosed:"+e);
		}
            }

	    // remove it from the list of open connections for this thread
	    // only if it was opened outside a tx.
	    if (pitx == null) {
                current.removeConnection(xac);
            }
	}
    }


    /**
     * Invoked when a fatal connection error occurs, just before an
     * SQLException is 
     * thrown to the application.
     *
     * @param event - an event object describing the source of the event
     */
    public void connectionErrorOccurred(ConnectionEvent event) {
        Trace.outln(trace, "ConnectionManager.connectionErrorOccurred(ConnectionEvent)");

        // First of all, always mark the Connection as closed.
        XAConnection xac = (XAConnection) event.getSource();

        synchronized (xac) {
            PoolItem pi = pool.closeConnection(xac);
	    if (pi == null) {
		// was already closed: just return.
		return;
	    }

	    // Check if inside a transaction 
	    Transaction pitx = (Transaction)pi.tx; 
	    Transaction tx = null;
	    try {
		tx = current.getTransaction();
		Trace.outln(trace, "\tTx = " + String.valueOf(tx));
	    } catch (NullPointerException n) {
		// current is null: we are not in EJBServer.
		Trace.errln("ConnectionManager: should not be used outside an EJBServer");
	    } catch (SystemException e) {
		Trace.errln("ConnectionManager: getTransaction failed:"+e);
	    }

            // delist Resource if in transaction
            if (tx != null) {
                if (pi.open < 1) try {
                    tx.delistResource(xac.getXAResource(), XAResource.TMFAIL);
                } catch (Exception e) {
                    Trace.errln("Exception in connectionErrorOccurred:"+e);
                }
            }

	    // remove it from the list of open connections for this thread
	    // only if it was opened outside a tx.
	    if (pitx == null) {
                current.removeConnection(xac);
            }
        }
    }

    // -----------------------------------------------------------------
    // CompletedTransactionListener interface
    // -----------------------------------------------------------------

    /**
     * Forget all about this Transaction (after the commit or rollback)
     * @param tx the transaction
     * @param st its status
     */
    public void endTransaction(Transaction tx, int st) {

        Trace.outln(trace, "ConnectionManager.endTransaction");

        pool.freeConnections(tx);
    }

    // -----------------------------------------------------------------
    // Referenceable implementation
    // -----------------------------------------------------------------

    /**
     * Retrieves the Reference of this object. Used at binding time by JNDI
     * to build a reference on this object.
     * 
     * @return  The non-null Reference of this object. 
     * @exception  NamingException  If a naming exception was encountered while 
     * retrieving the reference.  
     */
    public Reference getReference() throws NamingException {

        Reference ref = new Reference (this.getClass().getName(), 
                                       "org.objectweb.jonas.dbm.DataSourceFactory",
                                       null);
        // Actually, only the datasource.name is used! (See DataSourceFactory.java)
        ref.add(new StringRefAddr("datasource.name", getDataSourceName()));
        ref.add(new StringRefAddr("datasource.url", getUrl()));
        ref.add(new StringRefAddr("datasource.classname", getClassName()));
        ref.add(new StringRefAddr("datasource.username", getUserName()));
        ref.add(new StringRefAddr("datasource.password", getPassword()));
        return ref;
    }

   /**
    * get the ConnectionManager matching the DataSource name
    */
   public static ConnectionManager getConnectionManager(String dsname) {
       
       ConnectionManager cm = (ConnectionManager) cmList.get(dsname);
       return cm;
   }

   public void closeAllConnection(){
	Trace.outln(trace,"ConnectionManager.closeAllConnection()");
	pool.closeAllConnection();
    }

}
