/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999,2000 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * 00/18/04 Jun Inamori (j-office@osa.att.ne.jp)
 *          New implementation of unbindDataSources for closing correctly 
 *          all connections at server shutdown.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.dbm;

import java.util.Enumeration;
import java.util.Properties;
import java.util.Vector;
import javax.naming.NamingException;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.transaction.Transaction;
import org.objectweb.jonas.common.Env;
import org.objectweb.jonas.common.Trace;

/**
 * Database Manager acts as a factory for the ConnectionManager objects.
 * Its goal is to create such objects and to register them in JNDI
 *
 */

public class  DataBaseManager {

    private static final int trace = Trace.DB_8;

    private static final String DATASOURCE_NAME = "datasource.name";
    private static final String URL = "datasource.url";
    private static final String CLASS_NAME = "datasource.classname";
    private static final String USER_NAME = "datasource.username";
    private static final String PASSWORD = "datasource.password";
    private static final String CONNECTION_CHECK_LEVEL = "jdbc.connchecklevel";
    private static final String CONNECTION_MAX_AGE = "jdbc.connmaxage";
    private static final String CONNECTION_MAX_THREADS = "jdbc.connmaxthreads";
    private static final String CONNECTION_TIMEOUT = "jdbc.connexcltimeout";
    private static final String CONNECTION_STATEMENT = "jdbc.connteststmt";

    private static final String DEFAULT_CHECK_LEVEL = "1"; // check physical connection is still open
    private static final String DEFAULT_MAX_AGE = "30";    // 30 minutes
    private static final String DEFAULT_MAX_THREADS = "4"; // 4 concurrent threads on same tx/connection
    private static final String DEFAULT_TIMEOUT = "30";    // 30 seconds
    private static final String DEFAULT_STATEMENT  = "SELECT 1";

    // List of connection managers (DataSource) in use.
    private Vector dsList = null;

    // Initial Context for Naming
    private Context ictx = null;
    
    private static DataBaseManager unique = null;

    /*
     * We manage a unique instance
     *
     */
    private DataBaseManager() {
	dsList = new Vector();
	// Avoid using NamingManager here: performance is not a goal here.
	try {
	    ictx = new InitialContext();
	} catch (NamingException e) {
	    Trace.errln("Cannot create initial context");
	}
    }

    public static DataBaseManager getInstance() {
	if (unique == null)
	    unique = new DataBaseManager();
	return unique;
    }

    /**
     * Creates a ConnectionManager (implementation of sql.dataSource).
     *
     *  @param dsd - a set of properties that describes a dataSource and the ConnectionPool:
     *              datasource.name
     *              datasource.url
     *              datasource.classname
     *              datasource.username
     *              datasource.password
     *              jdbc.connchecklevel
     *              jdbc.connmaxage
     *              jdbc.connmaxthreads
     *              jdbc.connexcltimeout
     *              jdbc.connteststmt
     *
     *
     */
    public void createConnectionManager(Properties dsd) throws Exception{
	Trace.outln(trace, "DataBaseManager.createConnectionManager()");
	String dataSourceName = null;
	//traceProperties(dsd);             
	try {
	    
	    ConnectionManager ds = new ConnectionManager();
	    dataSourceName = dsd.getProperty(DATASOURCE_NAME);
	    ds.setDataSourceName(dataSourceName);
	    ds.setUrl(dsd.getProperty(URL));
	    ds.setClassName(dsd.getProperty(CLASS_NAME));
	    ds.setUserName(dsd.getProperty(USER_NAME, ""));
	    ds.setPassword(dsd.getProperty(PASSWORD, ""));
	    ds.poolConfigure(dsd.getProperty(CONNECTION_CHECK_LEVEL, DEFAULT_CHECK_LEVEL),
			     dsd.getProperty(CONNECTION_MAX_AGE, DEFAULT_MAX_AGE),
			     dsd.getProperty(CONNECTION_MAX_THREADS, DEFAULT_MAX_THREADS),
			     dsd.getProperty(CONNECTION_TIMEOUT, DEFAULT_TIMEOUT),
			     dsd.getProperty(CONNECTION_STATEMENT, DEFAULT_STATEMENT));

	    dsList.addElement(ds);
	    Trace.outln(Trace.DB_31, "Mapping datasource "+ds.getUrl()+" on "+dataSourceName);
	    ictx.rebind(dataSourceName, ds);
	} catch (Exception e) {
	    Trace.errln(" Cannot create ConnectionManager for " + dataSourceName);
	    throw e;
	}
    }

    /**
     * Creates a set of connectionManager
     *
     * @param dsds - a vector of Properties
     */
    public void start() throws Exception {
	Vector dsds = Env.getDataSources();
	for (Enumeration vp = dsds.elements(); vp.hasMoreElements(); ) {
	    Properties dsd = (Properties) vp.nextElement();
	    createConnectionManager(dsd);
	}
    }
    
   /**
    * Return the names of dataSources used
    *
    */
    public Vector  getDataSourceNames() {
	Trace.outln(trace, "DataBaseManager.getDataSourceNames()");
	return dsList;
    }

    /*
     * Unbind dataSource names from the Naming
     */
    public void unbindDataSources() {
	Trace.outln(trace, "DataBaseManager.unbindDataSources()");
	try {
	    for (Enumeration lk = dsList.elements(); lk.hasMoreElements(); ) {
		ConnectionManager cm = (ConnectionManager)lk.nextElement();
		cm.closeAllConnection();
		String dsn = cm.getDataSourceName();
		ictx.unbind(dsn);
	    }
	} catch (Exception e) {
	    Trace.errln("cannot unbind DataSources : " + e);
	    return;
	}
    }

    void traceProperties(Properties dsd) {
	Trace.outln(trace,DATASOURCE_NAME +" = " +dsd.getProperty(DATASOURCE_NAME));
	Trace.outln(trace,URL +" = " +dsd.getProperty(URL));
	Trace.outln(trace,CLASS_NAME +" = " +dsd.getProperty(CLASS_NAME));
	Trace.outln(trace,USER_NAME +" = " +dsd.getProperty(USER_NAME));
	Trace.outln(trace,PASSWORD +" = " +dsd.getProperty(PASSWORD));
	Trace.outln(trace,CONNECTION_CHECK_LEVEL +" = " +dsd.getProperty(CONNECTION_CHECK_LEVEL));
	Trace.outln(trace,CONNECTION_MAX_AGE +" = " +dsd.getProperty(CONNECTION_MAX_AGE));
	Trace.outln(trace,CONNECTION_MAX_THREADS +" = "+ dsd.getProperty(CONNECTION_MAX_THREADS));
	Trace.outln(trace,CONNECTION_TIMEOUT +" = " +dsd.getProperty(CONNECTION_TIMEOUT));
	Trace.outln(trace,CONNECTION_STATEMENT +" = " +dsd.getProperty(CONNECTION_STATEMENT));
    }
}
	
