/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999,2000 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.dbm;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.LinkedList;
import javax.sql.XAConnection;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.jtm.TransactionImpl;

public class Pool {
    private static final int trace = Trace.DB_8;
    
    // Liste of free available items
    private LinkedList oFreeItems = new LinkedList();
    // requests with no tx are identified by thread
    private int iItems_w_o_Tx = 0;
    // Requests with same tx get always the same conn
    private Hashtable hTx2Item = new Hashtable();
    
    // map XAConn => Item (to speedup CloseConn)
    private Hashtable hXAC2Item = new Hashtable();

    private ConnectionManager myConnMgr = null;
    
    // ---------------------------------------------------------------------
    // Pool Configuration
    // ---------------------------------------------------------------------
    // Max age of a Connection in milliseconds.
    long jdbcConnMaxAge;
    // Max. concurrent threads on same tx/conn
    int jdbcConnMaxThreads;
    // Max wait time (in milliseconds) if more than jdbcConnMaxThreads threads request conn
    long jdbcConnExclTimeout;
    // Level of checking on connections
    int jdbcConnCheckLevel;
    // test statement
    String jdbcTestStmt;

    public Pool(ConnectionManager cmgr) {
	myConnMgr = cmgr;
	(new PoolKeeper(this)).start();
    }
    
    /**
     * Look up a free XAConnection for this user and this transaction.
     * If found, set the item to "busy".
     *
     * @return the free PoolItem or null if no item available
     */
    public PoolItem openConnection(String user, Transaction  tx) throws SQLException {
        PoolItem item = null;
	boolean newTx = false;
        synchronized(this) {
            if (tx != null) {
                Object i;
                while ((i = hTx2Item.get(tx)) != null) {
                    if (i instanceof PoolItem)
                        break;
                    else try {
                        wait(); // for placeholder to be replaced
                    } catch (InterruptedException ign) {
                    }
		}
                item = (PoolItem)i;
            }
            if (item == null) {
		while (true) {
		    synchronized (oFreeItems) {
			if (oFreeItems.isEmpty()) {
			    Trace.outln(trace, 
					 "  Pool: cannot find a reusable Connection for Tx = " 
					 + String.valueOf(tx));
			    if (tx != null)
				hTx2Item.put(tx, new Object()); // Placeholder
			    return null; // => newConnection
			}
			item = (PoolItem)oFreeItems.removeFirst();
		    }
		    if (jdbcConnCheckLevel > 0) {
			try {
			    Connection conn = item.xaConn.getConnection();
			    if (conn.isClosed()) { 
				destroyItem(item); 
				continue; 
			    }
			    if (jdbcConnCheckLevel > 1) {
				Trace.outln(trace, "Retrying connection: "+jdbcTestStmt);
				java.sql.Statement stmt = conn.createStatement();
				stmt.execute(jdbcTestStmt); // => exception
				stmt.close();
			    }
			} catch (Exception ex) {
			    Trace.errln("  Pool: Remove invalid item.\n\t"+ex);
			    destroyItem(item); 
			    continue;
			}
		    }
		    break;
		}
		item.tx = tx;

                if (tx == null) {
                    iItems_w_o_Tx++;
                } else {
                    hTx2Item.put(tx, item);
		    newTx = true;
		}
            }
        }
        
        if (tx != null) {
            Trace.outln(trace, "  Pool: reuse a Connection for Tx = "+ String.valueOf(tx));
	    if (newTx) {
		try {
		    ((TransactionImpl)tx).addCompletedTransactionListener(myConnMgr);
		} catch (SystemException e) {
		    Trace.errln("openConnection: cannot addCompletedTransactionListener:"+e);
		}
	    }
	}
        // Try to get Connection open max jdbcConnMaxThreads times concurrent
        long w, t = System.currentTimeMillis() + jdbcConnExclTimeout;
        synchronized (item) {
            // Handle State "busy" (open>jdbcConnMaxThreads, tx!=null) => wait
            while (item.open >= jdbcConnMaxThreads) {
                w = t - System.currentTimeMillis();
                if (w < 0) {
                    Trace.errln("WARNING: Cannot 'exclusively' get connection");
                    break; // return non-exclusive conn
                }
                
                Trace.outln(trace, "  Pool: waiting for exclusive connection");
                try {
                    item.wait(w);
                } catch (InterruptedException ign) {
                    Trace.errln("WARNING: Interupted while waiting for"+
                                 " exclusive connection");
                    break; // return non-exclusive conn
                }
            }
            item.open++;
        }
	dump();
        return item;
    }
    
    /**
     * Add a new XAConnection in the pool
     */

    public PoolItem  newConnection(XAConnection xac, String user, Transaction tx) {
        PoolItem item = new PoolItem(xac, user, jdbcConnMaxAge);
        hXAC2Item.put(xac, item);
        item.tx = tx;
        item.open++;
        synchronized(this) {
            if (tx == null) {
                iItems_w_o_Tx++;
            } else {
                hTx2Item.remove(tx); // placeholder
                hTx2Item.put(tx, item);
                notifyAll();
            }
        }
        Trace.outln(trace, "  Pool: add a Connection for Tx = " + String.valueOf(tx));
	if (tx != null) {
	    try {
		((TransactionImpl)tx).addCompletedTransactionListener(myConnMgr);
	    } catch (SystemException e) {
		Trace.errln("newConnection: cannot addCompletedTransactionListener:"+e);
	    }
	}
	dump();
        return item;
    }
    
    /**
     * The transaction has committed (or rolled back). We can return its
     * connections to the pool of available connections.
     */
    public void freeConnections(Transaction tx) {

        // tx is guarantied to be not null
        PoolItem item = (PoolItem)hTx2Item.remove(tx);
        if (item == null) {
            Trace.errln("  Pool: no connection found to free for this Tx");
	    return;
        }

	synchronized(item) {
            Trace.outln(trace, "  Pool: free a Connection for this Tx");
            item.tx = null;

            if (item.open > 0) {
		// XXX Better to force a logical close here ?
                Trace.errln("WARNING: Connection not closed by caller");
		Trace.errln("Force a physical close to avoid reusing it in another transaction");
		item.open = 0;
		destroyItem(item);
		return;
	    }
            item.open = 0;
            
            // Check if this conn has reached it's max age
            if (item.lMaxAgeMillis < System.currentTimeMillis()) {
                Trace.outln(trace, 
                             "  Pool: remove a Connection from this pool"+
                             " while its max age is reached");
		destroyItem(item);
            } else synchronized (oFreeItems) {
                // Add it to Free list
                oFreeItems.addLast(item);
            }
        }
	dump();
    }
    

    /**
     * Mark a specific Connection in the pool as closed.
     * If it is no longer associated to a Tx, we can free it.
     *
     * @return the PoolItem the connection is in (=> tx)
     * or null if error.
     */
    public PoolItem closeConnection(XAConnection xac) {
        PoolItem item = (PoolItem)hXAC2Item.get(xac);
        if (item == null) {
            Trace.errln("Cannot resolve XAC => PoolItem");
            return null;
        }
        
        // The connection will be available only if not associated
        // to a transaction. Else, it will be reusable only for the
        // same transaction.
        synchronized(item) {
            item.open--;
            if (item.open < 0) {
                item.open = 0;
                Trace.outln(Trace.DB_31, "Connection was already closed");
		return null;
            } else if (item.tx == null && item.open > 0) {
                item.open = 0;
                Trace.errln("connection-open counter overflow");
            }

            if (item.tx != null) {
                Trace.outln(trace, "  Pool: keep the Connection for this tx");
            } else {
                Trace.outln(trace, "  Pool: free the Connection");

                // while freeConnections is not called, do it here
                iItems_w_o_Tx--;

                // Check if this conn has reached its max age
                if (item.lMaxAgeMillis < System.currentTimeMillis()) {
                    Trace.outln(trace, 
                                 "  Pool: remove a Connection from this pool"+
                                 " while it's max age is reached");
		    destroyItem(item);
                } else synchronized (oFreeItems) {
                    // else add it to FreeList for reuse
                    oFreeItems.addLast(item);
                }
            }
            item.notify();
        }
	dump();
        return item;
    }

    private void destroyItem(PoolItem item) {
       hXAC2Item.remove(item.xaConn);
       try {
           item.xaConn.close();
       } catch (java.sql.SQLException ign) {
           Trace.errln(ign.toString());
       }
       item.xaConn = null;
    }

    /**
     * Check all connections not in a transaction if their maxage
     * is reached and if so kill them.
     */
    public void check4MaxAge() {
        PoolItem item;
        synchronized (oFreeItems) {
            Iterator i = oFreeItems.iterator();
            while (i.hasNext()) {
                item = (PoolItem) i.next();
                synchronized(item) {
                    if (item.tx != null || item.open > 0)
                        Trace.errln("Inconsistent PoolItem removed from FreeList");
                    // Ceck if this conn has reached it's max age
                    if (item.lMaxAgeMillis < System.currentTimeMillis()) {
                        Trace.outln(trace, 
                                     "  Pool: remove a Connection from this pool"+
                                     " while it's max age is reached");
                        i.remove();
			destroyItem(item);
                    }
                }
            }
        }
	dump();
    }

    /**
     * Dump the pool 
     * For debugging: print statistics
     */
    private void dump() {
        PoolItem item;
        if (!Trace.isLevelSet(Trace.DB_11)) {
	    return;
	}
	synchronized(this) {
	    Trace.errln("------------ ConnectionPool -----------------");
	    Trace.errln("Nb of item not associated with tx = "+iItems_w_o_Tx);
	    Trace.errln("+++ Hashtable XAC -> item:");
	    Object k;
	    Enumeration e = hXAC2Item.keys();
	    while (e.hasMoreElements()) {
		k = e.nextElement();
		item = (PoolItem)hXAC2Item.get(k);
		Trace.errln(item.toString()+"\topen="+item.open+"\ttx="+item.tx);
	    }
                
	    Iterator i = oFreeItems.iterator();
	    Trace.errln("+++ FreeList:");
	    while (i.hasNext()) {
		item = (PoolItem)i.next();
		Trace.errln(item.toString()+"\topen="+item.open+"\ttx="+item.tx);
	    }
                
	    
	    e = hTx2Item.keys();
	    Trace.errln("+++ Associated with tx:");
	    while (e.hasMoreElements()) {
		k = e.nextElement();
		item = (PoolItem)hTx2Item.get(k);
		Trace.errln(item.toString()+"\topen="+item.open+"\ttx="+item.tx);
	    }
	    if (hXAC2Item.size() != oFreeItems.size() + iItems_w_o_Tx + hTx2Item.size()) {
		Trace.errln("  Pool: Inconsistency detected !");
		System.exit(1);	// debugging!
	    }
	    Trace.errln("---------------------------------------------");
	}
    }

    /**
     * Adm.stopServer()
     * -->DataBaseManager.unbindDataSources()
     * -->ConnectionManager.closeAllConnection()
     * -->Pool.closeAllConnection()
     */
    public void closeAllConnection() {
	Trace.outln(trace, "Pool.closeAllConnection()");

	Enumeration e = hXAC2Item.keys();
	try {
	    while (e.hasMoreElements()) {
		XAConnection xc = (XAConnection) e.nextElement();
		xc.close();
	    }
	} catch (java.sql.SQLException ex) {
	    Trace.errln(ex.toString());
	}
	dump();
    }
}
  
