/*
 * The contents of this file are subject to the JOnAS Public License Version
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): Lutris Technologies Inc http://www.lutris.com
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.deployment.api;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.io.Reader;
import java.io.InputStreamReader;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.io.ByteArrayOutputStream;

import org.xml.sax.SAXParseException;
import org.exolab.castor.xml.ValidationException;
import org.exolab.castor.xml.MarshalException;

import org.objectweb.jonas.deployment.xml.EjbJar;
import org.objectweb.jonas.deployment.xml.AssemblyDescriptor;
import org.objectweb.jonas.deployment.xml.Session;
import org.objectweb.jonas.deployment.xml.Entity;
import org.objectweb.jonas.deployment.xml.JonasEjbJar;
import org.objectweb.jonas.deployment.xml.JonasSession;
import org.objectweb.jonas.deployment.xml.JonasEntity;

/**
 * Class to hold extended(*) deployment descriptor information <BR>
 * (*) extended deployment descriptor information includes Jonas specific
 * deployment descriptor file information
 * @author Christophe Ney
 */
public class DeploymentDesc {

  private Vector beanDescs = new Vector();

  public static final String EJB_JAR_FILE_NAME = "META-INF/ejb-jar.xml";
  public static final String JONAS_EJB_JAR_FILE_NAME ="META-INF/jonas-ejb-jar.xml";

  /**
   * constructor is private, call one of the static methods instead
   */
  private DeploymentDesc(ClassLoader classLoader,EjbJar ejbJar,JonasEjbJar jonasEjbJar) throws DeploymentDescException {

    // test classloader
    if (classLoader==null) {
      throw new DeploymentDescException("Classloader is null");
    }

    // assembly descriptor
    AssemblyDescriptor asd = ejbJar.getAssemblyDescriptor();

    // hashtable of jonas-session
    Hashtable jonasSession = new Hashtable();
    Enumeration e = jonasEjbJar.enumerateJonasSession();
    while( e.hasMoreElements()) {
      JonasSession jSes = (JonasSession)e.nextElement();
      jonasSession.put(jSes.getEjbName().getContent(),jSes);
    }

    // session beans
    e = ejbJar.getEnterpriseBeans().enumerateSession();
    while( e.hasMoreElements()) {
      BeanDesc bd = null;
      Session ses = (Session)e.nextElement();
      // find corresponding jonas session
      JonasSession jSes = (JonasSession)jonasSession.get(ses.getEjbName().getContent());
      if (jSes==null)
      throw new DeploymentDescException("Jonas descriptor missing for bean "+ses.getEjbName().getContent());
      // stateful
      if (ses.getSessionType().getContent().equals("Stateful")) {
        bd = new SessionStatefulDesc(classLoader,ses,asd,jSes);
        // stateless
      } else if (ses.getSessionType().getContent().equals("Stateless")) {
        bd = new SessionStatelessDesc(classLoader,ses,asd,jSes);
        // error
      } else {
        throw new DeploymentDescException("Invalid session-type content for bean "+ses.getEjbName().getContent());
      }
      bd.check();
      beanDescs.addElement(bd);
    }

    // hashtable of jonas-entity
    Hashtable jonasEntity = new Hashtable();
    e = jonasEjbJar.enumerateJonasEntity();
    while( e.hasMoreElements()) {
      JonasEntity jEnt = (JonasEntity)e.nextElement();
      jonasEntity.put(jEnt.getEjbName().getContent(),jEnt);
    }
    // entity beans
    e = ejbJar.getEnterpriseBeans().enumerateEntity();
    while( e.hasMoreElements()) {
      BeanDesc bd = null;
      Entity ent = (Entity)e.nextElement();
      // find corresponding jonas session
      JonasEntity jEnt = (JonasEntity)jonasEntity.get(ent.getEjbName().getContent());
      if (jEnt==null)
      throw new DeploymentDescException("jonas-entity missing for bean "+ent.getEjbName().getContent());
      // bean managed
      if (ent.getPersistenceType().getContent().equals("Bean")) {
        bd =new EntityBmpDesc(classLoader,ent,asd,jEnt);
        // container managed (always jdbc)
      } else if (ent.getPersistenceType().getContent().equals("Container")) {
        bd = new EntityJdbcCmpDesc(classLoader,ent,asd,jEnt);
        // error
      } else {
        throw new DeploymentDescException("Invalid persistence-type content for bean "+ent.getEjbName().getContent());
      }
      bd.check();
      beanDescs.addElement(bd);
    }
  }

  /**
   * Factory method with given classloader
   * @param classLoader classloader used to load xml deployment descriptors and bean classes
   * @return instance of the corresponding DeploymentDesc
   * @exception DeploymentDescException when DeploymentDesc cannot be created with
   * given files.
   */
  public static DeploymentDesc getInstance(ClassLoader classLoader)
  throws DeploymentDescException {
    InputStream is;
    EjbJar ejbJar;
    JonasEjbJar jonasEjbJar;

    // load deployment descriptor data
    is = classLoader.getResourceAsStream(EJB_JAR_FILE_NAME);
    if (is==null) {
      throw new DeploymentDescException(EJB_JAR_FILE_NAME+" resource not found");
    }
    ejbJar = loadEjbJar(is,EJB_JAR_FILE_NAME);
    try {
      is.close();
    } catch(IOException e){
    }

    // load jonas deployment descriptor data
    is = classLoader.getResourceAsStream(JONAS_EJB_JAR_FILE_NAME);
    if (is==null) {
      throw new DeploymentDescException(JONAS_EJB_JAR_FILE_NAME+" resource not found");
    }
    jonasEjbJar = loadJonasEjbJar(is,JONAS_EJB_JAR_FILE_NAME);
    try {
      is.close();
    } catch(IOException e){
    }

    // instantiate deployment descriptor
    return new DeploymentDesc(classLoader,ejbJar,jonasEjbJar);
  }


  /**
   * Factory method using deployment descriptor and Jonas deployment descriptor
   * file names.
   * @return instance of the corresponding DeploymentDesc
   * @exception DeploymentDescException when DeploymentDesc cannot be created with
   * given files.
   * @deprecated use for integration only (jdk1.2 required)
   */
  public static DeploymentDesc getInstance(String ejbJarFileName, String jonasEjbJarFileName)
  throws DeploymentDescException {
    InputStream is;
    EjbJar ejbJar;
    JonasEjbJar jonasEjbJar;

    // load deployment descriptor data
    try {
      is = new FileInputStream(ejbJarFileName);
    } catch (FileNotFoundException e) {
      throw new DeploymentDescException(ejbJarFileName+" file not found");
    }
    ejbJar = loadEjbJar(is,ejbJarFileName);
    try {
      is.close();
    } catch(IOException e){
    }

    // load jonas deployment descriptor data
    try {
      is = new FileInputStream(jonasEjbJarFileName);
    } catch (FileNotFoundException e) {
      throw new DeploymentDescException(jonasEjbJarFileName+" file not found");
    }
    jonasEjbJar = loadJonasEjbJar(is,jonasEjbJarFileName);
    try {
      is.close();
    } catch(IOException e){
    }

    // instantiate deployment descriptor
    ClassLoader cl = DeploymentDesc.class.getClassLoader();
    return new DeploymentDesc(cl,ejbJar,jonasEjbJar);
  }


  /**
   * Load the EjbJar file
   */
  private static EjbJar loadEjbJar(InputStream is,String name)
  throws DeploymentDescException {
    Reader reader = new InputStreamReader(is);
    try {
      return EjbJar.unmarshal(reader);
    } catch (ValidationException e) {
      throw new DeploymentDescException("XML Validation Error in "+name,e);
    } catch (MarshalException e) {
      String detailMsg = "Invalid XML in " + name;
      if (e.getException() instanceof SAXParseException) {
        detailMsg=getSAXMsg(name,((SAXParseException)e.getException()),"Invalid XML");
      }
      throw new DeploymentDescException(detailMsg,e);
    } finally {
      try {
        reader.close();
      } catch (IOException e) {
      }
    }
  }


  /**
   * Load the JonasEjbJar file
   */
  private static JonasEjbJar loadJonasEjbJar(InputStream is,String name)
  throws DeploymentDescException {
    Reader reader = new InputStreamReader(is);
    try {
      return JonasEjbJar.unmarshal(reader);
    } catch (ValidationException e) {
      throw new DeploymentDescException("XML Validation Error in "+name,e);
    } catch (MarshalException e) {
      String detailMsg = "Invalid XML in " + name;
      if (e.getException() instanceof SAXParseException) {
        detailMsg=getSAXMsg(name,((SAXParseException)e.getException()),"Invalid XML");
      }
      throw new DeploymentDescException(detailMsg,e);
    } finally {
      try {
        reader.close();
      } catch (IOException e) {
      }
    }
  }


  /**
   * build a message from SAX Exception in a consistent style, consistent
   * with emacs compile mode (same as grep, cc, javac, etc).
   */
  private static String getSAXMsg(String fileName, SAXParseException exception, String msg) {
    String ret= fileName
      + ":" + exception.getLineNumber()
      + ":"+ exception.getColumnNumber()
      + ": ";
      ret+=msg;
      return ret;
  }
  
  /**
   * Get descriptors for all beans contained in jar file
   * @return Array of bean's descriptors
   */
  public BeanDesc[] getBeanDesc(){
    BeanDesc ret[] = new BeanDesc[beanDescs.size()];
    beanDescs.copyInto(ret);
    return ret;
  }

  /**
   * String representation of the object for test purpose
   * @return String representation of this object
   */
  public String toString() {
    StringBuffer ret = new StringBuffer();
    BeanDesc[] b = getBeanDesc();
    for (int i=0; i<b.length; i++) {
      ret.append("\ngetBeanDesc("+i+")="+b[i].getClass().getName());
      ret.append(b[i].toString());
    }
    return ret.toString();
  }


}
