/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A. 
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): Lutris Technologies Inc http://www.lutris.com
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
*/


package org.objectweb.jonas.deployment.api;

import java.lang.String;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

import org.objectweb.jonas.deployment.xml.AssemblyDescriptor;
import org.objectweb.jonas.deployment.xml.Entity;
import org.objectweb.jonas.deployment.xml.CmpField;
import org.objectweb.jonas.deployment.xml.JonasEntity;

/**
 * Base class for descriptor of an entity bean with container managed persistence 
 * @author Christophe Ney
 */
public abstract class EntityCmpDesc extends EntityDesc {

    protected Hashtable fieldDesc = new Hashtable();
    protected Field primaryKeyField = null;    
    protected Method isModifiedMethod = null;
    
    /**
     * Package protected constructor used by API
     */
    EntityCmpDesc(ClassLoader classLoader,Entity ent, AssemblyDescriptor asd,JonasEntity jEnt)
            throws DeploymentDescException {
        super(classLoader,ent,asd,jEnt);

	// isModifiedMethod
	if (jEnt.getIsModifiedMethodName() != null) {
	    String mName = jEnt.getIsModifiedMethodName().getContent();
	    try {
		isModifiedMethod = this.ejbClass.getMethod(mName, new Class[0]);
	    } catch (NoSuchMethodException e) {
		isModifiedMethod = null;
		throw new DeploymentDescException(mName+" is not a method of "+this.ejbClass.getName());
	    } catch (SecurityException e) {
                throw new DeploymentDescException("Cannot use java reflexion on "+this.ejbClass.getName());
	    }
	}

        // primary key field
        if (ent.getPrimkeyField()!=null) {
            String fn = ent.getPrimkeyField().getContent();
            try {
                primaryKeyField=ejbClass.getField(fn);
            } catch (NoSuchFieldException e){
                throw new DeploymentDescException(fn+" is not a field of "+this.ejbClass.getName());
            } catch (SecurityException e){
                throw new DeploymentDescException("Cannot use java reflexion on "+this.ejbClass.getName());
            }
        }

        // field descriptors for persistent fields
        CmpField[] cf = ent.getCmpField();
        for (int i=0; i<cf.length; i++){
            String fn = cf[i].getFieldName().getContent();
            // check for valid field name            
            try {
                Field f = ejbClass.getField(fn);
            } catch (NoSuchFieldException e) {
                throw new DeploymentDescException("Invalid field name in field-name in bean "+this.ejbName,e);
            } catch (SecurityException e) {
                throw new DeploymentDescException("Cannot use java reflexion on "+this.ejbClass.getName());
            }
            fieldDesc.put(fn,new FieldDesc(false));
        }
         
        // primary key fields
        if (primaryKeyField!=null) {
            // primary key field
            if (!fieldDesc.containsKey(primaryKeyField.getName())) 
                throw new DeploymentDescException("primkey-field is not listed as cmp-field in bean "+this.ejbName);
            fieldDesc.put(primaryKeyField.getName(),new FieldDesc(true));
        } else {
            // public fields of primary key class
            Field[] pcf = primaryKeyClass.getFields();
            for (int i=0; i<pcf.length; i++) {
                if (Modifier.isPublic(pcf[i].getModifiers())) {
                    String pn = pcf[i].getName();
                    // exclude serialVersionUID field for jdk1.2.1 on solaris
                    if (!pn.equals("serialVersionUID")) {
                        if (!fieldDesc.containsKey(pn)) 
                            throw new DeploymentDescException("public field "+pn+" of primkey-class is not listed as cmp-field in bean "+this.ejbName);
                        fieldDesc.put(pn,new FieldDesc(true));
                    }
                }
            }
        }
    }
    
    /**
     * Assessor for a CMP field
     * @return Descriptor for the given field
     */
    public boolean hasCmpFieldDesc(Field field) {
        return fieldDesc.containsKey(field.getName());
    }
    
    /**
     * Get descriptor for a given field
     * Used by GenIC
     * @param field of the bean class
     * @return Descriptor for the given field
     */
    public FieldDesc getCmpFieldDesc(Field field) {
        FieldDesc ret = (FieldDesc)fieldDesc.get(field.getName());
        if (ret==null)
            throw new Error("Field "+field.getName()+" is not a cmp field of class "+this.ejbClass.getName());
        return ret;
    }
    

    /**
     * Assessor for primary key being a field
     * @return true if primary key is a field
     */
    public boolean hasPrimaryKeyField(){
        return primaryKeyField!=null;
    }    
    
    /**
     * Get the primary key field for the entity.
     * Used by GenIC
     * @return Field for the primary key
     */
    public Field getPrimaryKeyField(){
        if (primaryKeyField==null)
            throw new Error("No primary key field defined for bean "+this.ejbName);
        return primaryKeyField;
    }

    /**
     * Assessor for existence of a isModified methoe
     * @return true of isModified method exist for the bean
     */
    public boolean hasIsModifiedMethod(){
        return isModifiedMethod!=null;
    }


    /**
     * Get the 'isModified' method name implemented in the bean class.
     * (This information is JOnAS specific).
     * @return Name of the isModified method
     */
    public Method getIsModifiedMethod(){
        if (isModifiedMethod==null)
            throw new Error("No isModified method defined for bean "+this.ejbName);
        return isModifiedMethod;
    }
        

    /**
     * String representation of the object for test purpose
     * @return String representation of this object
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
        ret.append(super.toString());
        Enumeration e = fieldDesc.keys();
        while (e.hasMoreElements()) {
            String f = (String)e.nextElement();
            FieldDesc fd = (FieldDesc)fieldDesc.get(f);
            ret.append("\ngetCmpFieldDesc("+f+")="+fd.getClass().getName());
            ret.append(fd.toString());
        }
        if (hasPrimaryKeyField())
            ret.append("\ngetPrimaryKeyField()="+getPrimaryKeyField().toString());
        if (hasIsModifiedMethod())
            ret.append("getIsModifiedMethod()="+getIsModifiedMethod().toString());
        return ret.toString();
    }
}

