/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A. 
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): Lutris Technologies Inc http://www.lutris.com
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
*/


package org.objectweb.jonas.deployment.api;


import java.lang.Class;
import java.lang.String;
import java.lang.Object;


/**
 * The EnvEntry class corresponds to the 'env-entry' element in the XML DTD
 * for EJB 1.1 deployment descriptor.
 *
 * The env-entry element contains the declaration of an enterprise bean's environment entry.
 * The declaration consists of:
 *   - an optional description (which is not taken into account here),
 *   - the name of the environment entry,
 *   - the type of the environment entry,
 *   - and an optional value.
 * @author Helene Joanin, Christophe Ney
 */
public class EnvEntryDesc {

    String name;
    Class type;
    Object value;

    /**
     * Package protected constructor used by API
     */
    EnvEntryDesc(org.objectweb.jonas.deployment.xml.EnvEntry env) 
            throws DeploymentDescException {
        name = env.getEnvEntryName().getContent();
        String t = env.getEnvEntryType().getContent();
        String v = (env.getEnvEntryValue()!=null) ? env.getEnvEntryValue().getContent() : null;
        try {        
            if (t.equals(Boolean.class.getName())) {
                type = Boolean.class;
                if (v!=null){
                    if (v.equals("True"))
                        value = Boolean.TRUE;
                    else if (v.equals("False")) 
                        value = Boolean.FALSE;
                    else 
                        throw new DeploymentDescException(v+" is not a valid value for env-entry "+name);
                }
            } else if (t.equals(String.class.getName())) {
                type = String.class;
                value = v;
            } else if (t.equals(Integer.class.getName())) {
                type = Integer.class;
                value = (v!=null)? new Integer(v):null;
            } else if (t.equals(Double.class.getName())) {
                type = Double.class;
                value = (v!=null)? new Double(v):null;
            } else if (t.equals(Byte.class.getName())) {
                type = Byte.class;
                value = (v!=null)? new Byte(v):null;
            } else if (t.equals(Short.class.getName())) {
                type = Short.class;
                value = (v!=null)? new Short(v):null;
            } else if (t.equals(Long.class.getName())) {
                type = Long.class;
                value = (v!=null)? new Long(v):null;
            } else if (t.equals(Float.class.getName())) {
                type = Float.class;
                value = (v!=null)? new Float(v):null;
            } else {
                throw new DeploymentDescException(t+" is not a valid type for env-entry "+name);
            }
        } catch (NumberFormatException e) {
            throw new DeploymentDescException(v+" is not a valid value for env-entry "+name,e);
        }
        
    }


    /**
     * Get the name of the environemt entry.
     * @return Name for environment entry
     */
    public String getName() {
    	return name;
    }

    /**
     * Get the fully-qualified Java type of the environemt entry.
     * Type is needed since value is optional
     * The possibles values are:
     *   java.lang.Boolean
     *   java.lang.String
     *   java.lang.Integer
     *   java.lang.Double
     *   java.lang.Byte
     *   java.lang.Short
     *   java.lang.Long
     *   java.lang.Float
     */
    public Class getType() {
	    return type;
    }

    /**
     * Assessor for existence of value for the descriptor
     * @return true if a value is available
     */
    public boolean hasValue() {
    	return value!=null;
    }

    /**
     * Get the value of the environment entry.
     * @return value for the environment entry (must be set)
     */
    public Object getValue() {
        if (value==null)
            throw new Error("Value not set for env-entry "+name);
    	return value;
    }

    /**
     * String representation of the object for test purpose
     * @return String representation of this object
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
        ret.append("\ngetName()="+getName());
        ret.append("\ngetType()="+getType());
        if (hasValue())
            ret.append("\ngetValue()="+getValue().toString());
        return ret.toString();
    }
    
}

