/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A. 
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): Lutris Technologies Inc http://www.lutris.com
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
* several changes 13.3.00
*/


package org.objectweb.jonas.deployment.api;

import java.lang.reflect.Method;

import org.objectweb.jonas.deployment.xml.TransAttribute;
import org.objectweb.jonas.deployment.xml.MethodName;
import org.objectweb.jonas.deployment.xml.MethodParam;
import org.objectweb.jonas.deployment.xml.MethodParams;

 /**
  * Class for deployment descriptors of methods<BR>
  * Holds description of transaction attributes
  * @author Helene Joanin, Christophe Ney
 */
public class MethodDesc {

    /**
     * Set of constants for method transaction attribute
     */
    public static final int TX_NOT_SUPPORTED = 0;
    // cney TODO remove TX_BEAN_MANAGED
    public static final int TX_BEAN_MANAGED  = 1;
    static final int TX_NOT_SET       = 1;
    public static final int TX_REQUIRED      = 2;
    public static final int TX_SUPPORTS      = 3;
    public static final int TX_REQUIRES_NEW  = 4;
    public static final int TX_MANDATORY     = 5;
    public static final int TX_NEVER         = 6;

    private static final String[] ATTR = {"TX_NOT_SUPPORTED","TX_BEAN_MANAGED","TX_REQUIRED","TX_SUPPORTS","TX_REQUIRES_NEW","TX_MANDATORY","TX_NEVER","TX_NOT_SET"};

    /**
     * value for undefined transaction attribute
     */
    private int txAttribute = TX_NOT_SET;
    private int txAttributeStatus = APPLY_TO_BEAN;
    
    public static final int APPLY_TO_NOTHING = 0;
    public static final int APPLY_TO_BEAN = 1;
    public static final int APPLY_TO_CLASS = 2;
    public static final int APPLY_TO_METHOD_NAME = 3;
    public static final int APPLY_TO_METHOD = 4;

    protected static final String[] APPLY_TO = {"APPLY_TO_NOTHING","APPLY_TO_BEAN","APPLY_TO_CLASS","APPLY_TO_METHOD_NAME","APPLY_TO_METHOD"};
        
    /**
     * package protected default constructor
     */
    MethodDesc() {
    }

    /**
     * Protected copy constructor used by API
     */
    protected MethodDesc(MethodDesc object) {
        txAttribute = object.txAttribute;
        txAttributeStatus = object.txAttributeStatus;
    }
    
    /**
     * Overwrite TxAttribute 
     * @param transAttribute new value for txAttribute
     * @param status applicability of given transAttribute parameter
     */
    void overwriteTxAttribute(TransAttribute transAttribute,int status)
            throws DeploymentDescException {
        // overwrite only if numerical value greater than existing one
        if (status<this.txAttributeStatus)
            return;
        setTxAttribute(transAttribute);
        txAttributeStatus = status;
    }

    /**
     * Set TxAttribute with given value
     */
    void setTxAttribute(TransAttribute transAttribute)
            throws DeploymentDescException {
        String tr = transAttribute.getContent();
        if (tr.equals("NotSupported")) 
            txAttribute = TX_NOT_SUPPORTED;
        else if (tr.equals("Required")) 
            txAttribute = TX_REQUIRED;
        else if (tr.equals("Supports")) 
            txAttribute = TX_SUPPORTS;
        else if (tr.equals("RequiresNew")) 
            txAttribute = TX_REQUIRES_NEW;
        else if (tr.equals("Mandatory")) 
            txAttribute = TX_MANDATORY;
        else if (tr.equals("Never")) 
            txAttribute = TX_NEVER;
        else {  
            throw new DeploymentDescException(tr+" is not a valid trans-attribute value");
        }
    }

    /**
     * Get the status of applicability for a given pattern to a method
     * @return status of applicability APPLY_TO_NOTHING,APPLY_TO_BEAN,APPLY_TO_CLASS,APPLY_TO_METHOD_NAME,APPLY_TO_METHOD
     */
    public static int matchPattern(java.lang.reflect.Method meth, Class patternExcludeClass, 
                                 MethodName patternMethodName, MethodParams patternMethodParams) {
        String mName = patternMethodName.getContent();
        // classes don't match
        if (meth.getDeclaringClass()==patternExcludeClass)
            return APPLY_TO_NOTHING;
        //class is enough
        if (mName.equals("*")) 
            return (patternExcludeClass==null)?APPLY_TO_BEAN:APPLY_TO_CLASS;
        // method name does not match
        if (!mName.equals(meth.getName()))
            return APPLY_TO_NOTHING;
        // no params specified (name test is enough)
        if (patternMethodParams==null)
            return APPLY_TO_METHOD_NAME;
        Class pars[] = meth.getParameterTypes();
        MethodParam[] pattPars = patternMethodParams.getMethodParam();
        // number of parameters does not match
        if (pars.length != pattPars.length)
            return APPLY_TO_NOTHING;
        for (int i=0; i<pars.length;i++) {
            if (!getClassName(pars[i]).equals(pattPars[i].getContent()))
                return APPLY_TO_NOTHING;
        }
        return APPLY_TO_METHOD;
    }

    /**
     * Returns common name of a given type <BR>
     * For exmple it returns int[] for an array of int
     * @return String with the name of the given type
     */
    protected static String getClassName(Class c) {
    	String name;
        if (c.isArray()) {
	        name = getClassName(c.getComponentType())+"[]";
        } else {
    	    name = c.getName();
	    }
        return(name);
    }

    
    
    /**
     * Get the container transaction attribute that match the method
     * @return Constant value within list :
     * TX_NOT_SUPPORTED,TX_REQUIRED,TX_SUPPORTS,TX_REQUIRES_NEW,TX_MANDATORY,TX_NEVER,TX_NOT_SET
     */
    public int getTxAttribute() {
	    return txAttribute;
    }
    
    /**
     * Get the container transaction attribute that match the method
     * @return Constant value within list :
     * TX_NOT_SUPPORTED,TX_REQUIRED,TX_SUPPORTS,TX_REQUIRES_NEW,TX_MANDATORY,TX_NEVER,TX_NOT_SET
     */
    public int getTxAttributeStatus() {
	    return txAttributeStatus;
    }
    
    
    /**
     * String representation of the object for test purpose
     * @return String representation of this object
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
        ret.append("\ngetTxAttribute()="+ATTR[getTxAttribute()]);
        ret.append("\ngetTxAttributeStatus()="+APPLY_TO[getTxAttributeStatus()]);
        return ret.toString();
    }
    

    /**
     * String representation of the transactionnal attribute
     * @return String representation of this transactionnal attribute
     */
    public static String getTxAttributeName(int value) {
	    if ((value < 0) && (value > ATTR.length)) {
    	    throw new Error(value+" is not a valid TxAttribute");
	    } 
        return ATTR[value];
	}
        
    /**
     * String representation of the transactionnal attribute
     * @return String representation of this transactionnal attribute
     */
    public String getTxAttributeName() {
        return ATTR[txAttribute];
	}
        
}
 


