/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A. 
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): Lutris Technologies Inc http://www.lutris.com
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
*/


package org.objectweb.jonas.deployment.api;

import java.lang.Class;

import org.objectweb.jonas.deployment.xml.ResourceRef;
import org.objectweb.jonas.deployment.xml.JonasResource;


/**
 * The ResourceRef class corresponds to the 'resource-ref' element in the XML DTD
 * for EJB 1.1 deployment descriptor.
 *
 * The resource-ref element contains a declaration of enterprise bean's reference
 * to an external resource.
 * It consists of:
 *   - an optional description (which is not taken into account here),
 *   - the resource manager connection factory reference name,
 *   - the indication of the resource manager connection factory type expected by the
 *     enterprise bean code,
 *   - and the type of authentication (bean or container).
 * @author Helene Joanin, Christophe Ney
 */
public class ResourceRefDesc {

    public static final int APPLICATION_AUTH = 0;
    public static final int CONTAINER_AUTH   = 1;
    private static final String[] AUTH = {"APPLICATION_AUTH","CONTAINER_AUTH"};

    String name;
    Class type;
    int authentication;
    String jndiName;

    /**
     * Package protected constructor used by API
     */
    ResourceRefDesc(ClassLoader classLoader,ResourceRef res, JonasResource jRes) 
            throws DeploymentDescException {
        name = res.getResRefName().getContent();
        try {
            type = classLoader.loadClass(res.getResType().getContent());
        } catch(ClassNotFoundException e) {
            throw new DeploymentDescException("res-type class not found for resource-ref "+name,e);
        }
        String auth = res.getResAuth().getContent();
        if (auth.equals("Application")) {
            authentication = APPLICATION_AUTH;
        } else if (auth.equals("Container")) {
            authentication = CONTAINER_AUTH;
        } else 
            throw new DeploymentDescException("res-auth not valid for resource-ref "+name);
        
        jndiName = jRes.getJndiName().getContent();
    }

    /**
     * Get resource ref. name
     * @return Name of the resource ref.
     */
    public String getName() {
    	return name;
    }

    /**
     * Get resource ref. type
     * @return Class of the resource ref.
     */
    public Class getType() {
	    return type;
    }

    /**
     * Get the authentication of the resource ref.
     * @return Authentication value within APPLICATION_AUTH, CONTAINER_AUTH
     */
    public int getAuthentication() {
    	return authentication;
    }
    
    /**
     * Assessor for JDBC resource
     * @return true if the resource is Jdbc compliant
     */
    public boolean isJdbc() {
        Class jdbcClass = javax.sql.DataSource.class;
        return jdbcClass.isAssignableFrom(type);
    }

    /**
     * Get the jndi name of the resource ref.
     * @return String representation of the JNDI name
     */
    public String getJndiName() {
        return jndiName;
    }
    
    /**
     * String representation of the object for test purpose
     * @return String representation of this object
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
        ret.append("\ngetName()="+getName());
        ret.append("\ngetType()="+getType());
        ret.append("\ngetAuthentication()="+AUTH[getAuthentication()]);
        ret.append("\nisJdbc()="+new Boolean(isJdbc()).toString());
        ret.append("\ngetJndiName()="+getJndiName());
        return ret.toString();
    }
    
    
}

