/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.jtm;

import java.rmi.RemoteException;
import java.rmi.NoSuchObjectException;
import org.objectweb.jonas.common.RemoteObject;
import java.util.Vector;
import javax.transaction.Status;
import javax.transaction.TransactionRolledbackException;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.timer.TimerEvent;
import org.objectweb.jonas.timer.TimerEventListener;
import org.objectweb.jonas.timer.TimerManager;

/**
 * Implementation of the object that represents a transaction.
 * This remote object has been created by a TransactionFactory.
 *
 * @see	   org.objectweb.jonas.jtm.TransactionFactory
 *
 */
public class ControlImpl extends RemoteObject 
    implements Control, Coordinator, RecoveryCoordinator, Terminator, Resource, TimerEventListener {
	
    private static int trace = Trace.DB_4;

    // ===============================================================
    // Private data
    // ===============================================================
    /**
     * @serial
     */
    private Vector resourceList = new Vector();
    /**
     * @serial
     */
    private Vector synchronizationList = new Vector();
    /**
     * @serial
     */
    private int mystatus = Status.STATUS_UNKNOWN;
    /**
     * @serial
     */
    private Coordinator supcoord;
    /**
     * @serial
     */
    private PropagationContext txcontext;
    /**
     * @serial
     */
    private RecoveryCoordinator recov;
    /**
     * @serial
     */
    private long mytimeout;
    /**
     * @serial
     */
    private TimerEvent mytimer = null;
    /**
     * @serial
     */
    private Otid myotid;
    /**
     * @serial
     */
    private Log mylog;

    // ---------------------------------------------------------------
    // Constructors
    // ---------------------------------------------------------------

    /**
     * Constructor for create or recreate
     *
     * @param timeout		Timeout in seconds for this transaction
     * @param otid		Otid allocated to this transaction
     * @param coord		Superior coordinator (null if create)
     */
    ControlImpl(long timeout, Otid otid, Coordinator supco) throws RemoteException {
		
	Trace.outln(trace, "ControlImpl constructor");

	// init object state
	mystatus = Status.STATUS_ACTIVE;
	myotid = otid;
	supcoord = supco;	// sub-coordinator
	recov = null;

	// XXX In case of sub-coordinator, should we arm a timer for rollback ???
	mytimeout = timeout;
	mytimer = TimerManager.getInstance().addTimer(this, timeout, null);

	// If sub-coord, register resource
	if (supco != null) {
	    try {
		recov = supco.register_resource(this);
	    } catch (Exception e) {
		Trace.errln("Cannot register sub-coordinator:\n" + e);
	    }
	}
    }


    // ---------------------------------------------------------------
    // Control Interface 
    // ---------------------------------------------------------------

    /**
     * Gets the Terminator object for this transaction
     *
     * @return		Terminator for this transaction
     */
    public Terminator get_terminator() 
	throws RemoteException {
	
	Trace.outln(trace, "get_terminator");
	return this;
    }
	
    /**
     * Gets the Coordinator object for this transaction
     *
     * @return		Coordinator for this transaction
     */
    public Coordinator get_coordinator() 
	throws RemoteException {
		
	Trace.outln(trace, "get_coordinator");
	return this;
    }
	

    // ---------------------------------------------------------------
    // Coordinator Interface 
    // ---------------------------------------------------------------

    /**
     * Gets the current status of this transaction
     *
     * @return		current transaction status
     */
    public int get_status() 
	throws RemoteException {

	Trace.outln(trace, "get_status");
	return mystatus;
    }

    /**
     * Tests if the given coordinator represents this transaction
     *
     * @param Coordinator	Coordinator to be compared to this transaction
     * @return			true if it is the same transaction
     */
    public boolean is_same_transaction(Coordinator tc) 
	throws RemoteException {

	Trace.outln(trace, "is_same_transaction");

	String other = null;

	try {
	    other = tc.get_transaction_name();
	} catch (Exception e) {
	    Trace.errln("is_same_transaction raised exception:\n" + e);
	    return false;
	}
	return other.equals(get_transaction_name());
    }

    /**
     * Registers a Resource object for this transaction
     *
     * @param res		Resource to be registered
     * @return			RecoveryCoordinator used for replay_completion
     */
    public synchronized RecoveryCoordinator register_resource(Resource res)  
	throws RemoteException {

	switch (mystatus) {
			
	case Status.STATUS_ACTIVE:
	    // Normal case
	    resourceList.addElement(res);
	    Trace.outln(trace, "Resource registered");
	    break;
			
	case Status.STATUS_MARKED_ROLLBACK:
	case Status.STATUS_ROLLEDBACK:
	case Status.STATUS_ROLLING_BACK:
	    Trace.errln("register_resource: Transaction Rolled back");
	    throw new TransactionRolledbackException();
			
	default:
	    Trace.errln("register_resource: Transaction Inactive");
	    throw new InactiveException();
	}
	return this;
    }
	
    /**
     * Registers a Synchronization object for this transaction
     *
     * @param sync		RemoteSynchro to be registered
     */
    public synchronized void register_synchronization(RemoteSynchro sync) 
	throws RemoteException {
		
	switch (mystatus) {
			
	case Status.STATUS_ACTIVE:
	    // Normal case
	    synchronizationList.addElement(sync);
	    Trace.outln(trace, "RemoteSynchro registered");
	    break;
			
	case Status.STATUS_MARKED_ROLLBACK:
	case Status.STATUS_ROLLEDBACK:
	case Status.STATUS_ROLLING_BACK:
	    Trace.errln("register_synchronization: Transaction Rolled back");
	    throw new TransactionRolledbackException();
			
	default:
	    Trace.errln("register_synchronization: Transaction Inactive");
	    throw new InactiveException();
	}
    }

    /**
     * Asks to rollback the transaction
     *
     */
    public synchronized void rollback_only() throws RemoteException {
		
	switch (mystatus) {
			
	case Status.STATUS_MARKED_ROLLBACK:
	    // nothing to do: Already marked rolledback.
	    Trace.errln("rollback_only: Already marked rolledback");
	    break;
			
	case Status.STATUS_ACTIVE:
	    mystatus = Status.STATUS_MARKED_ROLLBACK;
	    Trace.outln(trace, "rollback_only: Marked rollback");
	    break;
			
	default:
	    // already prepared
	    Trace.errln("rollback_only: Inactive");
	    throw new InactiveException();
	}
    }
	
    /**
     * Gets a String that represents the transaction name.
     * Only the Format Id and the Global Id are used to build it :
     * The Branch Qualifier is not used.
     *
     * @return			Transaction Name
     */
    public String get_transaction_name() throws RemoteException {
		
	Trace.outln(trace, "get_transaction_name");

	String gtrid = new String(myotid.getGlobalTransactionId());
	String ret = Long.toHexString(myotid.getFormatId()) + ":" + gtrid;

	return ret;
    }

    /**
     * Get the propagation context for this transaction
     *
     * @return	PropagationContext
     * @exception InactiveException			transaction inactive
     */
    public synchronized PropagationContext get_txcontext() throws RemoteException {

	if (txcontext == null) {
	    txcontext = new PropagationContext(mytimeout, get_coordinator(), myotid);
	}
	if (txcontext == null) {
	    Trace.outln(Trace.DB_31, "No transaction context");
	    throw new InactiveException();
	}
	return txcontext;
    }


    // ---------------------------------------------------------------
    // Terminator Interface 
    // ---------------------------------------------------------------

    /**
     * Commits this transaction
     *
     * @param report_heuristics want to report heuristics if any
     * 
     * @exception TransactionRolledbackException the transaction has been rolled back
     * @exception HeuristicMixed Resources have rolled back
     * @exception HeuristicHazard Resources may have rolled back
     */
    public synchronized void commit(boolean report_heuristics) throws RemoteException {
			
	Trace.outln(trace, "Terminator commit");

	// Invalids the Timer.
	if (mytimer != null) {
	    mytimer.unset();
	}
	
	// Checks status
	switch (mystatus) {
			
	case Status.STATUS_ACTIVE:
	    // normal case
	    break;
			
	case Status.STATUS_COMMITTED:
	    Trace.errln("commit: already done");
	    return;
			
	case Status.STATUS_ROLLEDBACK:
	    Trace.errln("commit: already rolled back");
	    throw new TransactionRolledbackException();
			
	case Status.STATUS_MARKED_ROLLBACK:
	    // Synchronization objects
	    int errors = do_before_completion();
	    if (errors > 0) {
		Trace.outln(Trace.DB_31, "before completion error at rollback");
	    }
	    do_rollback(report_heuristics);
	    throw new TransactionRolledbackException();
			
	default:
	    Trace.errln("commit: bad status");
	    throw new HeuristicMixed();
	}
	
	// Only one resource -> one phase commit.
	// If Resource raises an heuristic, forward it to the client.
	if (resourceList.size() == 1) {
	    Trace.outln(trace, "commit: 1 resource");
			
	    // Synchronization objects
	    int errors = do_before_completion();
	    if (errors > 0) {
		Trace.outln(Trace.DB_31, "before completion error -> rollback");
		do_rollback(report_heuristics);
		throw new TransactionRolledbackException();
	    }
			
	    mystatus = Status.STATUS_COMMITTING;
	    try {
		Resource res = (Resource) resourceList.elementAt(0);
		res.commit_one_phase();
		mystatus = Status.STATUS_COMMITTED;
	    } catch (TransactionRolledbackException e) {
		Trace.outln(Trace.DB_31, "commit_one_phase = TransactionRolledbackException");
		mystatus = Status.STATUS_ROLLEDBACK;
	    } catch (HeuristicHazard e) {
		Trace.outln(Trace.DB_31, "commit_one_phase = HeuristicException");
		mystatus = Status.STATUS_UNKNOWN;
	    } catch (NoSuchObjectException e) {
				// nothing done: can rollback transaction
		Trace.outln(Trace.DB_31, "commit_one_phase = NoSuchObjectException");
		mystatus = Status.STATUS_ROLLEDBACK;
	    } catch (Exception e) {
		Trace.errln("commit_one_phase = Unexpected exception: "+e);
		mystatus = Status.STATUS_UNKNOWN;
	    }
			
	    // Synchronization objects
	    do_after_completion();
			
	    do_destroy();
			
	    // Possibly report heuristics to the caller
	    switch (mystatus) {
	    case Status.STATUS_COMMITTED:
		break;
	    case Status.STATUS_ROLLEDBACK:
		throw new TransactionRolledbackException();
	    case Status.STATUS_UNKNOWN:
		if (report_heuristics) {
		    throw new HeuristicHazard();
		}
	    }
	    return;
	}
		
	// 2PC - phase 1
	int v = do_prepare(report_heuristics);
		
	// Depending on vote, commits or rollbacks transaction
	switch (v) {
	case Resource.VOTE_COMMIT:
	    Trace.outln(trace, "commit: committing Tx");
	    break;
	case Resource.VOTE_ROLLBACK:
	    Trace.outln(trace, "commit: rolling back Tx");
	    do_rollback(report_heuristics);
	    return;
	case Resource.VOTE_READONLY:
	    // No commit phase.
	    Trace.outln(trace, "commit: readonly resources");
	    mystatus = Status.STATUS_COMMITTED;
	    do_destroy();
	    return;
	}
		
	// Prepare was OK: Decision to commit.
	mylog.flushLog(Log.DECISION_TO_COMMIT);

	// 2PC - phase 2
	if (do_commit(report_heuristics) == 0)
	    do_destroy();
    }
	
    // ---------------------------------------------------------------
    // Resource or Terminator Interface 
    // rollback() has the same signature in Terminator and Resource!
    // ---------------------------------------------------------------

    /**
     * Rolls back this transaction branch. Can be a sub-coordinator or
     * a normal coordinator.
     *
     */
    public synchronized void rollback()  throws RemoteException {

	Trace.outln(trace, "Terminator rollback");

	// Checks status
	switch (mystatus) {
			
	case Status.STATUS_ACTIVE:
	case Status.STATUS_MARKED_ROLLBACK:
	    // normal case
	    break;
			
	case Status.STATUS_ROLLEDBACK:
	    Trace.errln("rollback: already rolled back");
	    return;
			
	default:
	    Trace.errln("rollback: bad status");
	    throw new HeuristicMixed("rollback: bad status");
	}
	
	// Superior coordinator.
	if (supcoord == null) {
	    // Synchronization objects
	    int errors = do_before_completion();
	    if (errors > 0) {
		Trace.outln(Trace.DB_31, "before completion error at rollback");
	    }
	    // Invalids the Timer. 
	    if (mytimer != null) {
		mytimer.unset();
	    }
	}
	
	// rollback Tx
	try {
	    do_rollback(false);
	} catch (Exception e) {
	    Trace.errln("rollback raised exception "+e);
	}
    }


    // ---------------------------------------------------------------
    // Resource Interface (for sub-coordinators)
    // ---------------------------------------------------------------

    /**
     * Sub-coordinator has received prepare from its superior.
     * It must more or less do the same things that the phase 1 of the 2PC.
     *
     * @return		Vote : commit, roollback or read-only.
     */
    public synchronized int prepare() throws RemoteException {

	Trace.outln(trace, "sub-coordinator prepare");

	// Checks status
	switch (mystatus) {
			
	case Status.STATUS_ACTIVE:
	    // normal case
	    break;	

	case Status.STATUS_COMMITTED:
	    // Should not occur ?
	    Trace.errln("transaction already commited");
	    return Resource.VOTE_COMMIT;
	
	case Status.STATUS_ROLLEDBACK:
	    // Should not occur ?
	    Trace.errln("transaction already rolled back");
	    return Resource.VOTE_ROLLBACK;

	case Status.STATUS_MARKED_ROLLBACK:
	    do_rollback(false);
	    return Resource.VOTE_ROLLBACK;
				
	default:
	    // Don't know what to do here...
	    Trace.errln("prepare: bad status");
	    return Resource.VOTE_ROLLBACK;
	}
			
	int ret = do_prepare(false);

	switch (ret) {
	case Resource.VOTE_COMMIT:
	    Trace.outln(trace, "prepare: vote commit");
	    break;
	case Resource.VOTE_ROLLBACK:
	    Trace.outln(trace, "prepare: vote rollback");
	    do_rollback(false);
	    return ret;
	case Resource.VOTE_READONLY:
	    Trace.outln(trace, "prepare: vote readonly");
	    mystatus = Status.STATUS_COMMITTED;
	    do_destroy();
	    return ret;
	}

	// Flush log + recovery coordinator
	// TODO

	return ret;
    }

    /**
     * Sub-coordinator received commit from its superior.
     * It must more or less do the same things that the phase 2 of the 2PC.
     */
    public synchronized void commit() throws RemoteException {

	Trace.outln(trace, "sub-coordinator commit");

	// check status
	switch (mystatus) {
	case Status.STATUS_PREPARED:
	    // normal case
	    break;		
	default:
	    // Don't know what to do here...
	    Trace.errln("commit: bad status");
	    return;
	}

	// send commit to resources
	if (do_commit(true) == 0)
	    do_destroy();
    }

    /**
     * Sub-coordinator received commit_one_phase from its superior.
     * It is more or less a Terminator.commit().
     */
    public void commit_one_phase() throws RemoteException {

	Trace.outln(trace, "sub-coordinator commit_one_phase");

	// Just call Terminator.commit
	commit(true);
    }

    /**
     * forget transaction
     *
     */
    public synchronized void forget() throws RemoteException {

	Trace.outln(trace, "sub-coordinator forget");

	do_destroy();
    }

    // ---------------------------------------------------------------
    // RecoveryCoordinator Interface 
    // ---------------------------------------------------------------

    /**
     * Asks the status of this transaction, after recovery of a Resource
     *
     * @param res		Resource recovering
     */
    public int replay_completion(Resource res) 
	throws RemoteException {

	Trace.outln(trace, "replay_completion");
	return mystatus;
    }


    // ===============================================================
    // TimerEventListener implementation
    // ===============================================================

    /**
     * The transaction timeout hhas expired
     * Do not synchronize this method to avoid deadlocks!
     */
    public void timeoutExpired(Object arg) {
	Trace.outln(Trace.DB_31, "JTM: timeout expired");
	try {
	    do_rollback(false);
	} catch (Exception e) {
	    Trace.errln("rollback raised exception "+e);
	}
    }

    // ===============================================================
    // Private methods
    // ===============================================================
	
    /**
     * timeout expired on this transaction.
     * This method is not private, because it's used by Timers
     */
    void ding() throws InactiveException {
    }

    /**
     * Phase 1 of the 2-Phase-Commit:
     * Sends prepare to each registered resource
     * This internal routine should be used either for prepare on sub-coordinator
     *  or for commit on Terminator.
     *
     * @return			global vote
     */
    private int do_prepare(boolean report_heuristics) {

	int errors = 0;
	int ret = Resource.VOTE_READONLY;

	// Synchronization objects
	errors = do_before_completion();
	if (errors > 0) {
	    Trace.outln(trace, "before_completion failed -> rollback");
	    return Resource.VOTE_ROLLBACK;
	}
			
	// No resource -> just forget transaction.
	if (resourceList.size() == 0) {
	    Trace.errln("commit: no resource");
	    mystatus = Status.STATUS_COMMITTED;
	    do_after_completion();
	    do_destroy();
	    return ret;
	}
	
	// Creates a log for that transaction, where we will add all the
	// resources that replied VOTE_COMMIT to prepare.
	// Do not flush the log on disk before decision to commit.
	mylog = new Log();
		  
	// Sends prepare to each resource.
	// In case of prepare on sub-coord. we may have only 1 resource.
	// In case of phase 1 of the 2PC, we have several resources, because
	// the case of 1 resource is treated with commit_one_phase (optimization)
	mystatus = Status.STATUS_PREPARING;
	for (int i = 0; i < resourceList.size(); i++) {
	    Resource res = (Resource) resourceList.elementAt(i);
	    if (errors > 0) {
		// An error has occured: Send rollback to the Resource
		try {
		    Trace.outln(trace, "send rollback to resource");
		    res.rollback();
		} catch (Exception e) {
		    Trace.errln("Cannot rollback resource: " + e);
		}
	    } else {
		// No error yet: Send prepare to the resource.
		try {
		    Trace.outln(trace, "send prepare to resource");
		    switch (res.prepare()) {
		    case Resource.VOTE_COMMIT:
			// Log resource
			mylog.addResource(res);
			ret = Resource.VOTE_COMMIT;
			break;
		    case Resource.VOTE_ROLLBACK:
			Trace.outln(Trace.DB_31, "Resource replied rollback to prepare");
			ret = Resource.VOTE_ROLLBACK;
			errors++;
			break;
		    case Resource.VOTE_READONLY:
			Trace.outln(trace, "Resource replied rollback to prepare");
			break;
		    }
		} catch (HeuristicHazard e) { // Subcoordinator only
		    Trace.errln("HeuristicHazard on prepare");
		    ret = Resource.VOTE_ROLLBACK;
		    errors++;
		} catch (HeuristicMixed e) { // Subcoordinator only
		    Trace.errln("HeuristicMixed on prepare");
		    ret = Resource.VOTE_ROLLBACK;
		    errors++;
		} catch (Exception e) {
		    Trace.errln("exception on prepare: " + e);
		    ret = Resource.VOTE_ROLLBACK;
		    errors++;
		}
	    }
	}
	if (ret == Resource.VOTE_READONLY) {
	    Trace.outln(trace, "All resources returned Readonly");
	    mystatus = Status.STATUS_COMMITTED;
	    // Synchronization objects
	    do_after_completion();
	}
	if (ret == Resource.VOTE_COMMIT) {
	    mystatus = Status.STATUS_PREPARED;
	}
	return ret;
    }

    /**
     * Phase 2 of the 2-Phase-Commit:
     * Sends commit to each registered resource
     * This internal routine should be used either for commit on sub-coordinator
     *  or for commit on Terminator.
     *
     * @return		0 if commit OK.
     */
    private int do_commit(boolean report_heuristics)
	throws TransactionRolledbackException, HeuristicMixed, HeuristicHazard {

	Trace.outln(trace, "Committing transaction");

	// First check that a log is initialized
	if (mylog == null) {
	    Trace.errln("do_commit: no log");
	    return -1;
	}

	// Status Transaction = committing
	mystatus = Status.STATUS_COMMITTING;

	// Sends commit to each resource "prepared".
	int commitnb = 0;
	int heuristicnb = 0;
	int errors = 0;
	for (int i = 0; i < mylog.resourceLogged.size(); i++) {
	    ResourceInfo resinfo = (ResourceInfo) mylog.resourceLogged.elementAt(i);
	    if (resinfo.mystate != ResourceInfo.PREPARED) {
		Trace.outln(Trace.DB_31, "do_commit: resource not prepared");
		continue;
	    }
	    // aborts transaction if error and no resource committed yet
	    if (commitnb == 0 && errors > 0) {
		try {
		    Trace.outln(Trace.DB_31, "do_commit: Send rollback to resource");
		    resinfo.myres.rollback();
		    resinfo.mystate = ResourceInfo.ROLLEDBACK;
		} catch (HeuristicCommit e) {
		    Trace.errln("do_commit: Heuristic commit");
		    resinfo.mystate = ResourceInfo.HEURISTIC_COMMIT;
		    commitnb++;
		    heuristicnb++;
		} catch (Exception e) {
		    Trace.errln("do_commit: exception on rollback: " + e);
		}
		continue;
	    }
	    // commits resource
	    try {
		Trace.outln(trace, "do_commit: Send commit to resource");
		resinfo.myres.commit();
		resinfo.mystate = ResourceInfo.COMMITTED;
		commitnb++;
	    } catch (HeuristicRollback e) {
		Trace.errln("do_commit: Heuristic Rollback");
		resinfo.mystate = ResourceInfo.HEURISTIC_ROLLBACK;
		errors++;
		if (commitnb > 0)
		    heuristicnb++;
	    } catch (HeuristicMixed e) {
		Trace.errln("do_commit: Heuristic Mixed");
		resinfo.mystate = ResourceInfo.HEURISTIC_MIXED;
		errors++;
		commitnb++;
		heuristicnb++;
	    } catch (HeuristicHazard e) {
		Trace.errln("do_commit: Heuristic Hazard");
		resinfo.mystate = ResourceInfo.HEURISTIC_HAZARD;
		errors++;
		commitnb++;
		heuristicnb++;
	    } catch (NotPreparedException e) {
		Trace.errln("do_commit: Resource Not Prepared");
		resinfo.mystate = ResourceInfo.REGISTERED;
		errors++;
	    } catch (NoSuchObjectException e) {
		Trace.errln("do_commit: invalid objref - assume committed");
		resinfo.mystate = ResourceInfo.COMMITTED;
		commitnb++;
	    } catch (Exception e) {
		Trace.errln("do_commit: exception on commit: " + e);
		errors++;
	    }
	}
	if (errors == 0) {
	    // Everything's fine.
	    Trace.outln(trace, "do_commit: transaction committed");
	    mystatus = Status.STATUS_COMMITTED;
	    mylog.forgetLog();
	    // Synchronization Objects
	    do_after_completion();
	    return 0;
	}
	if (heuristicnb == 0) {
	    // Transaction has been eventually rolled back
	    Trace.outln(Trace.DB_31, "do_commit: transaction rolled back");
	    mystatus = Status.STATUS_ROLLEDBACK;
	    mylog.forgetLog();
	    // Synchronization Objects
	    do_after_completion();
	    throw new TransactionRolledbackException();
	}
	// Heuristics must be logged
	Trace.outln(Trace.DB_31, "do_commit: Heuristics must be logged");
	mystatus = Status.STATUS_UNKNOWN;
	mylog.updateLog();
	// Synchronization Objects
	do_after_completion();
	if (report_heuristics) {
	    throw new HeuristicHazard();
	}
	return -1;
    }

    /**
     * Rollbacks the transaction
     * This internal routine should be used either for rollback on sub-coordinator
     * or for commit on Terminator when prepare returned a vote RollBack.
     *
     */
    private void do_rollback(boolean report_heuristics) 
	throws HeuristicMixed {

	Trace.outln(trace, "Rolling back transaction");
	
	mystatus = Status.STATUS_ROLLEDBACK;

	// Rollback resources
	int commitnb = 0;
	for (int i = 0; i < resourceList.size(); i++) {
	    Resource res = (Resource) resourceList.elementAt(i);
	    try {
		Trace.outln(trace, "Send rollback to Resource");
		res.rollback();
	    } catch (HeuristicCommit e) {
		Trace.errln("Rollback raised HeuristicCommit");
		commitnb++;
	    } catch (Exception e) {
		Trace.errln("Cannot rollback resource: " + e);
	    }
	}

	// Synchronization objects
	do_after_completion();

	if (commitnb > 0 && report_heuristics) {
	    // May be should throw HeuristicCommit instead
	    throw new HeuristicMixed();
	}
    }

    /**
     * Destroys transaction object.
     */
    private void do_destroy() {
	// No object destruction in java. The garbage collector will do this.
    }

    /**
     * Sends before_completion to the registered synchronizations
     *
     * @return		0 if no error
     */
    private int do_before_completion() {
	int errors = 0;

	Trace.outln(trace, "do_before_completion");
	Trace.outln(trace, "synchronizationList.size()="+synchronizationList.size());
	for (int i = 0; i < synchronizationList.size(); i++) {

	    RemoteSynchro sync = (RemoteSynchro) synchronizationList.elementAt(i);
	    try {
		sync.before_completion(this);
	    } catch (Exception e) {
		Trace.errln("before_completion raised exception "+e);
		errors++;
	    }
	}
	return errors;
    }

    /**
     * Sends after_completion to the registered synchronizations
     */
    private void do_after_completion() {

	Trace.outln(trace, "do_after_completion "+mystatus);
	Trace.outln(trace, "synchronizationList.size()="+synchronizationList.size());	
	for (int i = 0; i < synchronizationList.size(); i++) {
	    RemoteSynchro sync = (RemoteSynchro) synchronizationList.elementAt(i);
	    try {
		sync.after_completion(this, mystatus);
	    } catch (Exception e) {
		Trace.errln("after_completion raised exception "+e);
	    }
	}
    }

}
