/*
 *
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.jtm;

import java.util.Properties;
import java.util.Hashtable;
import java.util.Vector;
import java.rmi.RemoteException;
import java.sql.SQLException;
import javax.naming.Referenceable;
import javax.naming.Reference;
import javax.naming.NamingException;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.StringRefAddr;
import javax.sql.XAConnection;
import javax.transaction.*;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.Xid;
import org.objectweb.jonas.timer.TimerManager;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.jtm.Coordinator;
import org.objectweb.jonas.jtm.PropagationContext;
import org.objectweb.jonas.jtm.TransactionFactory;


/**
 * Current is the common Implementation for UserTransaction and TransactionManager.
 * UserTransaction is used by clients that want manage transactions themselves.
 * It is referenceable via JNDI
 * TransactionManager is used by an EJBServer.
 * This object is unique in a VM, i.e. each EJBServer has ONE Current object
 * and each client program should normally issue only ONE lookup on JNDI.
 * Current also implements Referenceable, because of JNDI.
 */
public class Current implements UserTransaction, TransactionManager, Referenceable {

    // ------------------------------------------------------------------
    // Thread specific data
    // ------------------------------------------------------------------

    // transaction associated to the thread
    private static ThreadLocal threadTx = new ThreadLocal();
    // list of connections open outside a transaction
    private static ThreadLocal threadConnections = new ThreadLocal();

    // ------------------------------------------------------------------
    // Private data
    // ------------------------------------------------------------------

    private static final int trace = Trace.DB_5;

    // Static hashtable: Xid ---> transaction
    private static Hashtable txXids = new Hashtable();

    // Must be init at null, for clients that do not get UserTransaction
    private static Current unique = null;

    // private info	
    private static TimerManager timermgr = null;	// local
    private static TransactionFactory tm = null; 	// local or remote

    private static final int DEFAULT_TIMEOUT = 60;
    private int transactionTimeout = DEFAULT_TIMEOUT;

    // ------------------------------------------------------------------
    // Constructors
    // ------------------------------------------------------------------

    /**
     * Default constructor.
     * A client does not need the TMFactory.
     */
    public Current() {

	Trace.outln(trace, "Current.Current()");
	unique = this;
    }
	
    /**
     * Constructor for EJBServer. The TM factory is passed as argument.
     * Note that the TM factory can be either local or remote.
     */
    public Current(TransactionFactory tmfact) {

	Trace.outln(trace, "Current.Current(TransactionFactory tmfact)");
	unique = this;
	tm = tmfact;
	timermgr = TimerManager.getInstance();
    }

    // ------------------------------------------------------------------
    // UserTransaction implementation
    // ------------------------------------------------------------------

    /**
     * Create a new transaction and associate it with the current thread.
     *
     * @exception NotSupportedException Thrown if the thread is already
     *    associated with a transaction. (nested transaction are not
     *    supported)
     *
     * @exception SystemException Thrown if the transaction manager 
     *    encounters an unexpected error condition
     *
     */
    public void begin() throws NotSupportedException, SystemException {

	Trace.outln(trace, "Current.begin()");

	// Checks that no transaction is already associated with this thread.
	TransactionImpl tx = (TransactionImpl) threadTx.get();
	if (tx != null) {
	    throw new NotSupportedException("Nested transactions not supported");
	}

	// Build a new Xid
	XidImpl otid = new XidImpl();

	// Creates a new TransactionImpl object
	// - May raise SystemException
	tx = new TransactionImpl(otid, transactionTimeout);

	// Attach current thread to the new transaction
	doAttach(tx);

	// Associates this Tx with the Xid
	putTxXid(otid, tx);

	// Set a timer for the transaction
	if (timermgr != null) {
	    tx.setTimer(timermgr.addTimer(tx, transactionTimeout, null));
	}
    }

    /**
     * Complete the transaction associated with the current thread. When this
     * method completes, the thread becomes associated with no transaction.
     *
     * @exception RollbackException Thrown to indicate that
     *    the transaction has been rolled back rather than committed.
     *
     * @exception HeuristicMixedException Thrown to indicate that a heuristic 
     *    decision was made and that some relevant updates have been committed
     *    while others have been rolled back.
     *
     * @exception HeuristicRollbackException Thrown to indicate that a 
     *    heuristic decision was made and that some relevant updates have been
     *    rolled back.
     *
     * @exception SecurityException Thrown to indicate that the thread is
     *    not allowed to commit the transaction.
     *
     * @exception IllegalStateException Thrown if the current thread is
     *    not associated with a transaction.
     *
     * @exception SystemException Thrown if the transaction manager 
     *    encounters an unexpected error condition
     */
    public void commit() throws RollbackException, 
	HeuristicMixedException, HeuristicRollbackException, SecurityException,
	IllegalStateException, SystemException {

	Trace.outln(trace, "Current.commit()");

	// Get Transaction
	TransactionImpl tx = (TransactionImpl) getTransaction();
	if (tx == null) {
	    throw new IllegalStateException("Cannot get Transaction for commit");
	}
	XidImpl xid = tx.getXid();

	// Detach Transaction from current thread
	// Must be done at this level (see JTA spec. 3.3.3)
	// TMSUCCESS should be defined in Transaction.
	doDetach(XAResource.TMSUCCESS);

	// Commit Transaction. Exceptions may be raised!
	tx.commit();
    }

    /**
     * Roll back the transaction associated with the current thread. When this
     * method completes, the thread becomes associated with no transaction.
     *
     * @exception SecurityException Thrown to indicate that the thread is
     *    not allowed to roll back the transaction.
     *
     * @exception IllegalStateException Thrown if the current thread is
     *    not associated with a transaction.
     *
     * @exception SystemException Thrown if the transaction manager 
     *    encounters an unexpected error condition
     *
     */
    public void rollback() throws IllegalStateException, SecurityException, SystemException {

	Trace.outln(trace, "Current.rollback()");

	// Get Transaction
	TransactionImpl tx = (TransactionImpl) getTransaction();
	if (tx == null) {
	    throw new IllegalStateException("Cannot get Transaction for rollback");
	}
	XidImpl xid = tx.getXid();

	// Dissociates this Transaction from current thread
	// Must be done at this level (see JTA spec. 3.3.3)
	// TMFAIL should be defined in Transaction.
	doDetach(XAResource.TMFAIL);

	// Roll back the transaction. Exceptions may be raised!
	tx.rollback();
    }

    /**
     * Modify the transaction associated with the current thread such that
     * the only possible outcome of the transaction is to roll back the
     * transaction.
     *
     * @exception IllegalStateException Thrown if the current thread is
     *    not associated with a transaction.
     *
     * @exception SystemException Thrown if the transaction manager 
     *    encounters an unexpected error condition
     *
     */
    public void setRollbackOnly() throws IllegalStateException, SystemException {

	Trace.outln(trace, "Current.setRollbackOnly()");

	// Get Transaction
	TransactionImpl tx = (TransactionImpl) getTransaction();
	if (tx == null) {
	    throw new IllegalStateException("Cannot get Transaction for setRollbackOnly");
	}

	// Set transaction rollback only. Exceptions may be raised!
	tx.setRollbackOnly();
    }

    /**
     * Obtain the status of the transaction associated with the current thread.
     * 
     * @return The transaction status. If no transaction is associated with 
     *    the current thread, this method returns the Status.NoTransaction 
     *    value.
     *
     * @exception SystemException Thrown if the transaction manager 
     *    encounters an unexpected error condition
     *
     */
    public int getStatus() throws SystemException {

	Trace.outln(trace, "Current.getStatus()");

	// Get Transaction
	TransactionImpl tx = (TransactionImpl) getTransaction();
	if (tx == null) {
	    return Status.STATUS_NO_TRANSACTION;
	}

	// Get TX status. Exceptions may be raised!
	return tx.getStatus();
    }

    /**
     * Modify the value of the timeout value that is associated with the
     * transactions started by the current thread with the begin method.
     *
     * <p> If an application has not called this method, the transaction
     * service uses some default value for the transaction timeout.
     *
     * @param seconds The value of the timeout in seconds. If the value
     *    is zero, the transaction service restores the default value.
     *
     * @exception SystemException Thrown if the transaction manager 
     *    encounters an unexpected error condition
     *
     */
    public void setTransactionTimeout(int seconds) throws SystemException {

	Trace.outln(trace, "Current.setTransactionTimeout(int "+seconds+")");

	if (seconds > 0)
	    transactionTimeout = seconds;
	else
	    transactionTimeout = DEFAULT_TIMEOUT;
    }

    // ------------------------------------------------------------------
    // TransactionManager implementation
    // (only the methods that are not already in UserTransaction)
    // ------------------------------------------------------------------

    /**
     * Get the transaction object that represents the transaction 
     * context of the calling thread.
     *
     * @return the Transaction object representing the transaction 
     * associated with the calling thread. If the calling thread is
     * not associated with a transaction, a null object reference
     * is returned.
     *
     * @exception SystemException Thrown if the transaction manager 
     * encounters an unexpected error condition
     *
     */
    public Transaction getTransaction() throws SystemException {

	Trace.outln(trace, "Current.getTransaction()");

	return (Transaction) threadTx.get();
    }

    /** 
     * Resume the transaction context association of the calling thread
     * with the transaction represented by the supplied Transaction object. 
     * When this method returns, the calling thread is associated with the 
     * transaction context specified.
     * Warning: No XA start is done here. We suppose it is already done
     * after a getConnection. The supposed programming model is: getConnection,
     * SQL, close.
     *
     * @param tobj The <code>Transaction</code> object that represents the 
     *    transaction to be resumed.
     *
     * @exception InvalidTransactionException Thrown if the parameter 
     *    transaction object contains an invalid transaction
     *
     * @exception IllegalStateException Thrown if the thread is already
     *    associated with another transaction.
     *
     * @exception SystemException Thrown if the transaction manager 
     *    encounters an unexpected error condition
     */
    public void resume(Transaction tobj) 
	throws InvalidTransactionException, IllegalStateException, SystemException {

	Trace.outln(trace, "Current.resume()");

	// invalid Transaction
	if (tobj == null) {
	    Trace.errln("resume: null arg.");
	    throw new InvalidTransactionException("resume(null) is not valid");
	}

	// Check that the thread is not already associated to ANOTHER transaction
	Transaction mytx = (Transaction) threadTx.get();
	if (mytx != null) {
	    if (mytx.equals(tobj)) {
		Trace.outln(trace, "resume: nothing to do");
		return;
	    }
	    Trace.errln("resume: already associated with another transaction.");
	    throw new IllegalStateException("the thread is already associated with another transaction.");
	}

	// Makes the new association
	doAttach(tobj);
    }
	
    /** 
     * Suspend the transaction currently associated with the calling 
     * thread and return a Transaction object that represents the 
     * transaction context being suspended. If the calling thread is 
     * not associated with a transaction, the method returns a null 
     * object reference. When this method returns, the calling thread 
     * is associated with no transaction. 
     * Warning: No XA end is done here. We suppose it is already done
     * after a close. The supposed programming model is: getConnection,
     * SQL, close.
     *
     * @return Transaction object representing the suspended transaction.
     *
     * @exception SystemException Thrown if the transaction manager 
     *    encounters an unexpected error condition
     *
     * @exception SystemException Thrown if the transaction manager 
     *    encounters an unexpected error condition
     *
     */
    public Transaction suspend() throws SystemException {

	Trace.outln(trace, "Current.suspend()");
	
	// TMSUSPEND should be defined in Transaction.
	return doDetach(XAResource.TMSUSPEND);
    }


    // ------------------------------------------------------------------
    // Referenceable implementation
    // ------------------------------------------------------------------

    /**
     * Retrieves the Reference of this object. 
     * 
     * @return  The non-null Reference of this object. 
     * @exception  NamingException  If a naming exception was encountered while retrieving the reference.  
     */
    public Reference getReference() throws NamingException {

	Trace.outln(trace, "Current.getReference()");

	// create the reference
	Reference ref = new Reference(this.getClass().getName(), 
				      "org.objectweb.jonas.jtm.UserTransactionFactory", 
				      null);

	return ref;
    }

    // ------------------------------------------------------------------
    // Other public methods
    // ------------------------------------------------------------------

    /**
     * Returns the unique instance of the class or null if not initialized
     * in case of plain client.
     *
     * @return The Current object created 
     */
    public static Current getCurrent() {
	return unique;
    }

    /**
     * Returns the TMFactory (in JTM)
     */
    public static TransactionFactory getJTM() {
	if (tm == null) {
	    Trace.errln("Current: TMFactory is null!");
	}
	return tm;
    }

    /**
     * Associate to the current thread a transaction represented by its
     * propagation context. This is used internally by the implicit
     * propagation of the transactionnal context:
     * - in the skeleton, before calling the request
     * - in the stub, after receiving the reply (in case it was updated)
     */
    public void setPropagationContext(PropagationContext pctx) {

	if (pctx == null) {
	    Trace.outln(trace, "Current.setPropagationContext(null) -> detach");
	    // Detaches thread from any transaction
	    threadTx.set(null);
	    return;
	}

	Trace.outln(trace, "Current.setPropagationContext("+pctx.getXid()+")");

	// Get the Transaction matching this Xid if it exists
	XidImpl xid = (XidImpl) pctx.getXid();
	TransactionImpl tx = getTxXid(xid);
	if (tx == null) {
	    Trace.outln(trace, "Current: setPropagationContext -> new Tx");
	    // create a Transaction object
	    tx = new TransactionImpl(pctx);
	    putTxXid(xid, tx);
	} else {
	    // Just in case it has been changed
	    tx.setPropagationContext(pctx);
	}

	// Associates this Tx with the current thread (= resume)
	threadTx.set(tx);
    }

    /**
     * Get the propagation context associated with the current thread
     * or null if the thread is not involved in a transaction.
     */
    public PropagationContext getPropagationContext() {

	Trace.outln(trace, "Current.getPropagationContext()");

	try {
	    TransactionImpl tx = (TransactionImpl) getTransaction();
	    if (tx != null)
		// will return null if Tx not valid.
		return tx.getPropagationContext();
	} catch (SystemException e) {
	    Trace.errln("getPropagationContext system exception:"+e);
	}
	return null;
    }



    /**
     * Add a connection to the list of the connections used by the calling
     * thread.
     */
    public void addConnection(XAConnection xac) {
	Trace.outln(trace, "Current.addConnection");
	Vector connectionList = (Vector) threadConnections.get();
	if (connectionList == null) {
	    connectionList = new Vector();
	    threadConnections.set(connectionList);
	}
	connectionList.addElement(xac);
    }

    /**
     * Remove a connection in the list of open connections for this thread
     */
    public void removeConnection(XAConnection xac) {
	Trace.outln(trace, "Current.removeConnection");
	Vector connectionList = (Vector) threadConnections.get();
	if (connectionList == null) {
	    Trace.errln("removeConnection: connectionlist is empty");
	    return;
	}
	connectionList.removeElement(xac);
    }

    /**
     * Set the connectionList and return old value.
     */
    public static Vector setConnectionList(Vector newlist) {
	Vector oldlist = (Vector) threadConnections.get();
	threadConnections.set(newlist);
	return oldlist;
    }
	
    /**
     * Forget all about this transaction. We must destroy references to
     * TransactionImpl object to allow the garbage collector to free
     * memory allocated to this transaction.
     */
    public void forgetTx(XidImpl xid) {
	removeTxXid(xid);
    }

    // ------------------------------------------------------------------
    // private methods
    // ------------------------------------------------------------------
	
    /*
     * Attach thread to the given transaction (= resume)
     */
    private void doAttach(Transaction tx) {
	Trace.outln(trace, "Attach Tx to Thread");

	// Associates this Tx with the current thread
	threadTx.set(tx);
	
	// Enlist open connections and setAutoCommit to false
	Vector connectionList = (Vector) threadConnections.get();
	if (connectionList != null) {
	    for (int i = 0; i < connectionList.size(); i++) {
		XAConnection xac = (XAConnection) connectionList.elementAt(i);
		Trace.outln(trace, "doAttach: Enlist a connection open by this thread");
		try {
		    xac.getConnection().setAutoCommit(false);
		    tx.enlistResource(xac.getXAResource());
		} catch (RollbackException e) {
		    // never.
		    Trace.errln("doAttach: RollbackException");
		} catch (SQLException e) {
		    // never.
		    Trace.errln("doAttach:"+e);
		} catch (SystemException e) {
		    Trace.errln("doAttach:"+e);
		}
	    }
	}
    }

    /*
     * Detach thread from transaction
     *
     * @param flag passed to delistResource.
     */
    private Transaction doDetach(int flag) {
	Trace.outln(trace, "Detach Thread from Tx");

	// Get Transaction previously attached
	Transaction tx = (Transaction) threadTx.get();

	// Unlist connections open by this thread before transaction start
	// and set autocommit to true, in case the connection will be used
	// outside a transaction.
	if (tx != null) {
	    Vector connectionList = (Vector) threadConnections.get();
	    if (connectionList != null) {
		for (int i = 0; i < connectionList.size(); i++) {
		    XAConnection xac = (XAConnection) connectionList.elementAt(i);
		    Trace.outln(trace, "doDetach: Unlist a connection open by this thread");
		    try {
			// cannot setAutoCommit(true) now, because commit is not
			// done already!
			tx.delistResource(xac.getXAResource(), flag);
		    } catch (SQLException e) {
			// never.
			Trace.errln("doAttach:"+e);
		    } catch (SystemException e) {
			Trace.errln("doAttach:"+e);
		    }
		}
	    }
	}

	// Dissociates the transaction from current thread and return it.
	threadTx.set(null);
	return tx;
    }

    /*
     * putTxXid
     */
    private void putTxXid(XidImpl xid, TransactionImpl tx) {
	Trace.outln(trace, "Associate Tx to Xid:"+xid);
	txXids.put(xid, tx);
    }

    /*
     * getTxXid
     */
    private TransactionImpl getTxXid(XidImpl xid) {
	TransactionImpl tx = (TransactionImpl) txXids.get(xid);
	return tx;
    }

    /*
     * removeTxXid
     */
    private void removeTxXid(XidImpl xid) {
	Trace.outln(trace, "Remove Tx from Xid:"+xid);
	txXids.remove(xid);
    }
}

