/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.jtm;

import java.rmi.RemoteException;
import java.sql.SQLException;
import java.util.Vector;
import javax.sql.DataSource;
import javax.sql.XAConnection;
import javax.transaction.*;
import javax.transaction.xa.XAException;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.Xid;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.timer.TimerEvent;
import org.objectweb.jonas.timer.TimerEventListener;
import org.objectweb.jonas.jtm.Coordinator;
import org.objectweb.jonas.jtm.HeuristicMixed;
import org.objectweb.jonas.jtm.PropagationContext;
import org.objectweb.jonas.jtm.RecoveryCoordinator;
import org.objectweb.jonas.jtm.Terminator;

/**
 * TransactionImpl is the implementation of the Transaction interface,
 * defined in JTA specifications. This object is intended to be used
 * by the EJBServer for transaction management. It is used indirectly
 * by the UserTransaction implementation too, i.e. the Current object.
 * The goal is to use the JTA interface to hide the JTM interface to
 * the caller (EJBServer, Bean or Client).
 */
public class TransactionImpl implements Transaction, TimerEventListener {

    private static final int trace = Trace.DB_5;

    // ------------------------------------------------------------------
    // Private data
    // ------------------------------------------------------------------
    private SubCoordinator subcoord = null;
    private PropagationContext myCtx = null;
    private boolean interpose = false;
    private TimerEvent timer = null; // keep this to unvalidate timer
    private RecoveryCoordinator recoveryCoord = null;

    // ------------------------------------------------------------------
    // Constructors
    // ------------------------------------------------------------------

    /**
     * New transaction (begin).
     *
     * @param timeout The value of the timeout in seconds.
     *
     */
    public TransactionImpl(Xid xid, int timeout) throws SystemException {
	Trace.outln(trace, "TransactionImpl.TransactionImpl(Xid xid, int timeout)");

	// Build a propagation context local (no ref to JTM yet)
	myCtx = new PropagationContext(timeout, null, xid);
    }

    /**
     * New Transaction for this thread (setPropagationContext)
     *
     * @param pctx propagation context
     *
     */
    public TransactionImpl(PropagationContext pctx) {
	Trace.outln(trace, "TransactionImpl.TransactionImpl(PropagationContext pctx)");
	if (pctx == null) {
	    Trace.errln("TransactionImpl: null PropagationContext");
	}
	myCtx = pctx;
	// interposition will be done later, only if necessary.
	interpose = true;
    }

    // ------------------------------------------------------------------
    // Transaction implementation
    // ------------------------------------------------------------------

    /**
     * Complete the transaction represented by this Transaction object
     * The calling thread is not required to have the same transaction
     * associated with the thread. (JTA 3.3.3)
     *
     * @exception RollbackException Thrown to indicate that
     *    the transaction has been rolled back rather than committed.
     *
     * @exception HeuristicMixedException Thrown to indicate that a heuristic 
     *    decision was made and that some relevant updates have been committed
     *    while others have been rolled back.
     *
     * @exception HeuristicRollbackException Thrown to indicate that a 
     *    heuristic decision was made and that some relevant updates have been
     *    rolled back.
     *
     * @exception SecurityException Thrown to indicate that the thread is
     *    not allowed to commit the transaction.
     *
     * @exception IllegalStateException Thrown if the current thread is
     *    not associated with a transaction.
     *
     * @exception SystemException Thrown if the transaction manager 
     *    encounters an unexpected error condition
     */
    public void commit() throws RollbackException, 
                HeuristicMixedException, HeuristicRollbackException,
                SecurityException, SystemException {

	Trace.outln(trace, "TransactionImpl.commit()");

	// *** Distributed transaction.
	Terminator term = myCtx.getTerminator();
	if (term != null) {
	    // Commits the Transaction, with heuristic report
	    try {
		term.commit(true);
	    } catch (java.rmi.ServerException e) {
		if (e.detail instanceof TransactionRolledbackException) {
		    Current.getCurrent().forgetTx(getXid());
		    Trace.outln(Trace.DB_31, "Commit distributed transaction -> rolled back!");
		    throw new RollbackException();
		}
		if (e.detail instanceof HeuristicMixed) {
		    Trace.outln(Trace.DB_31, "Commit distributed transaction -> Heuristic mixed!");
		    throw new HeuristicMixedException();
		}
	    } catch (TransactionRolledbackException e) {
		Current.getCurrent().forgetTx(getXid());
		Trace.outln(Trace.DB_31, "Commit distributed transaction -> rolled back!");
		throw new RollbackException();
	    } catch (HeuristicMixed e) {
		Trace.outln(Trace.DB_31, "Commit distributed transaction -> Heuristic mixed!");
		throw new HeuristicMixedException();
	    } catch (Exception e) {
		Trace.errln("Unexpected Exception on commit:"+e);
		throw new SystemException("Unexpected Exception on commit");
	    }
	    Current.getCurrent().forgetTx(getXid());
	    return;
	}

	// *** Local transaction
	// commit_one_phase may raise remote exceptions. We must rethrow local exceptions.
	// if no coordinator, nothing to do.
	if (subcoord != null) {
	    try {
		subcoord.commit_one_phase();
	    } catch (TransactionRolledbackException e) {
		Trace.outln(Trace.DB_31, "Commit local transaction -> rolled back!");
		throw new RollbackException();
	    } catch (RemoteException e) {
		Trace.errln("Unexpected Exception on commit_one_phase:"+e);
		throw new SystemException("Unexpected Exception on commit_one_phase");
	    }
	}
    }
    
    /**
     * Delist the resource specified from the current transaction 
     * associated with the calling thread.
     *
     * @param xaRes The XAResource object representing the resource to delist
     *
     * @param flag One of the values of TMSUCCESS, TMSUSPEND, or TMFAIL.
     *
     * @exception IllegalStateException Thrown if the transaction in the 
     *    target object is inactive.
     *
     * @exception SystemException Thrown if the transaction manager 
     *    encounters an unexpected error condition
     *
     * @return true if the dissociation of the Resource is successful.
     */
    public boolean delistResource(XAResource xares, int flag)
        throws IllegalStateException, SystemException {

	Trace.outln(trace, "TransactionImpl.delistResource()");

	// Send the XA end to the XAResource
	try {
	    xares.end(myCtx.getXid(), flag);
	} catch (XAException e) {
		throw new SystemException("Cannot send XA end:"+e);
	}
	return true;
    }
    
    /**
     * Enlist the resource specified with the current transaction 
     * context of the calling thread
     *
     * @param xaRes The XAResource object representing the resource to delist
     *
     * @return <i>true</i> if the resource was enlisted successfully; otherwise 
     *    false.
     *
     * @exception RollbackException Thrown to indicate that
     *    the transaction has been marked for rollback only.
     *
     * @exception IllegalStateException Thrown if the transaction in the 
     *    target object is in prepared state or the transaction is inactive.
     *
     * @exception SystemException Thrown if the transaction manager 
     *    encounters an unexpected error condition
     *
     */
    public boolean enlistResource(XAResource xares)
	throws RollbackException, IllegalStateException,
	       SystemException {

	Trace.outln(trace, "TransactionImpl.enlistResource(XAResource xares)");

	// Check trivial cases
	if (xares == null) {
	    Trace.errln("enlistResource: null argument");
	    throw new SystemException("enlistResource: null argument");
	}
	if (myCtx == null) {
	    throw new SystemException("enlistResource: no Propagation Context");
	}

	// make a subCoordinator object if not existing yet
	if (subcoord == null) {
	    makeSubCoord();
	    if (subcoord == null) {
		Trace.errln("enlistResource: could not create subcoordinator");
		throw new SystemException("enlistResource: could not create subcoordinator");
	    }
	}

	// Add the Resource to the list
	// may raise exceptions
	boolean found = subcoord.addResource(xares);

	// Send the XA start to the XAResource
	// This code raises several issues:     XXX
	// 1) We should set the flag TMRESUME if the resource was previously
	//    suspended by end(TMSUSPEND) !!!
	// 2) A new Xid branch should be generated in case of new RM (if !found)
	//    See JTA Specifications, page 12/13.
	int flag = found ? XAResource.TMJOIN : XAResource.TMNOFLAGS;
	try {
	    xares.start(myCtx.getXid(), flag);
	} catch (XAException e) {
		throw new SystemException("Cannot send XA start:"+e);
	}
	return true;
    }

    /**
     * Obtain the status of the transaction associated with the current thread.
     * 
     * @return The transaction status. If no transaction is associated with 
     *    the current thread, this method returns the Status.NoTransaction 
     *    value.
     *
     * @exception SystemException Thrown if the transaction manager 
     *    encounters an unexpected error condition
     *
     */
    public int getStatus() throws SystemException {

	Trace.outln(trace, "TransactionImpl.getStatus()");

	// *** Distributed transaction
	Coordinator coord = myCtx.getCoordinator();
	if (coord != null) {
	    // Ask the transaction status to JTM 
	    int ret;
	    try {
		ret = coord.get_status();
	    } catch (Exception e) {
		return Status.STATUS_NO_TRANSACTION;
	    }
	    return ret;
	}

	// *** Local transaction
	// The status is kept in the subcoordinator
	if (subcoord == null) {
	    // The transaction has just been started
	    // No resource or synchro are registered yet.
	    return Status.STATUS_ACTIVE;
	}
	return subcoord.getStatus();
    }

    /** 
     * Register a synchronization object for the transaction currently
     * associated with the calling thread. The transction manager invokes
     * the beforeCompletion method prior to starting the transaction 
     * commit process. After the transaction is completed, the transaction
     * manager invokes the afterCompletion method.
     *
     * @param sync The javax.transaction.Synchronization object for the 
     *    transaction associated with the target object
     *
     * @exception RollbackException Thrown to indicate that
     *    the transaction has been marked for rollback only.
     *
     * @exception IllegalStateException Thrown if the transaction in the 
     *    target object is in prepared state or the transaction is inactive.
     *
     * @exception SystemException Thrown if the transaction manager 
     *    encounters an unexpected error condition
     *
     */
    public void registerSynchronization(Synchronization sync)
	throws RollbackException, IllegalStateException, 
	       SystemException {

	Trace.outln(trace, "TransactionImpl.registerSynchronization(Synchronization sync)");
	    
	// It's time to make the subcoordinator, if not existing yet.
	if (subcoord == null) {
	    makeSubCoord();
	}

	// Add Synchronization to the list.
	// may raise exceptions
	subcoord.addSynchronization(sync);
    }

    /**
     * Rollback the transaction represented by this Transaction object.
     *
     * @exception IllegalStateException Thrown if the transaction in the 
     *    target object is in prepared state or the transaction is inactive.
     *
     * @exception SystemException Thrown if the transaction manager 
     *    encounters an unexpected error condition
     *
     */
    public void rollback() throws IllegalStateException, SystemException {

	Trace.outln(trace, "TransactionImpl.rollback()");

	// *** Distributed transaction.
	Terminator term = myCtx.getTerminator();
	if (term != null) {
	    // Rollback the Transaction
	    try {
		term.rollback();
	    } catch (java.rmi.ServerException e) {
		// HeuristicCommit ????
		throw new IllegalStateException("Exception on rollback:"+e.detail);
	    } catch (Exception e) {
		throw new SystemException("Unexpected Exception on rollback");
	    } 
	    Current.getCurrent().forgetTx(getXid());
	    return;
	}

	// *** Local transaction.
	// if no coordinator, nothing to do.
	if (subcoord != null) {
	    try {
		subcoord.rollback();
	    } catch (RemoteException e) {
		throw new IllegalStateException("Exception on rollback:"+e);
	    }
	}
    }
	
    /**
     * Modify the transaction associated with the current thread such that
     * the only possible outcome of the transaction is to roll back the
     * transaction.
     *
     * @exception IllegalStateException Thrown if the current thread is
     *    not associated with any transaction.
     *
     * @exception SystemException Thrown if the transaction manager 
     *    encounters an unexpected error condition
     *
     */
    public void setRollbackOnly() throws IllegalStateException, SystemException {

	Trace.outln(trace, "TransactionImpl.setRollbackOnly()");

	// make a subCoordinator object if not existing yet
	if (subcoord == null) {
	    makeSubCoord();
	}

	// status is kept in subcoordinator
	subcoord.setRollbackOnly();
    }

    // ------------------------------------------------------------------
    // TimerEventListener implementation
    // ------------------------------------------------------------------

    /**
     * timeout for that transaction has expired
     */
    public void timeoutExpired(Object arg) {

	Trace.outln(trace, "TransactionImpl.timeoutExpired(Object arg)");
	if (interpose == false) {
	    if (subcoord != null) {
		try {
		    subcoord.rollback();
		} catch (RemoteException e) {
		    // Nothing to do here?
		}
	    }
	}
    }

    // ------------------------------------------------------------------
    // This object is used as an HashTable index
    // ------------------------------------------------------------------

    /**
     * return true if objects are identical
     */
    public boolean equals(Object obj2) {
	TransactionImpl tx2 = (TransactionImpl) obj2;

	// trivial cases
	if (tx2 == this) {
	    return true;
	}
	if (tx2 == null) {
	    Trace.outln(Trace.DB_31, "Null pointer in Transaction.equals");
	    return false;
	}

	// compare otids
	return getXid().equals(((TransactionImpl)tx2).getXid());
    }

    /**
     * return a hashcode value for this object
     */
    public int hashCode() {
	return getXid().hashCode();
    }

    // ------------------------------------------------------------------
    // Other methods
    // ------------------------------------------------------------------

    /**
     * Register a CompletedTransactionListener
     */
    public void addCompletedTransactionListener(CompletedTransactionListener l)
	throws IllegalStateException, SystemException {

	Trace.outln(trace, "TransactionImpl.addCompletedTransactionListener");
	    
	// It's time to make the subcoordinator, if not existing yet.
	if (subcoord == null) {
	    makeSubCoord();
	}

	// Add the listener to the list.
	// may raise exceptions
	subcoord.addCompletedTransactionListener(l);
    }

    /**
     * Return associated PropagationContext
     * Used for implicit Context propagation.
     *
     * @return PropagationContext associated with the transaction.
     */
    public PropagationContext getPropagationContext() {
	return myCtx;
    }

    /**
     * set a timer for the transaction
     */
    public void setTimer(TimerEvent timer) {
	Trace.outln(trace, "TransactionImpl.setTimer(TimerEvent timer)");
	this.timer = timer;
    }

    /**
     * unset the timer
     */
    public void unsetTimer() {
	Trace.outln(trace, "TransactionImpl.unsetTimer()");
	if (timer != null) {
	    timer.unset();
	    timer = null;
	}
    }

    /**
     * update the propagation context
     */
    public void setPropagationContext(PropagationContext pctx) {

	Trace.outln(trace, "TransactionImpl.setPropagationContext(PropagationContext pctx)");
	
	// Debugging - to be removed later XXX
	if (myCtx.getXid().equals(pctx.getXid()) == false) {
	    Trace.errln("setPropagationContext: Bad Xid");
	    return;
	}

	Coordinator remoteCoord = pctx.getCoordinator();

	// Debugging - to be removed later XXX
	if (remoteCoord == null && myCtx.getCoordinator() != null) {
	    Trace.errln("setPropagationContext: Bad Coordinator");
	    return;
	}

	// Interpose subCoordinator if newly distributed Tx
	if (remoteCoord != null && myCtx.getCoordinator() == null) {
	    myCtx.setCoordinator(remoteCoord);
	    if (subcoord != null) {
		// register the subCoordinator as a Resource.
		try {
		    recoveryCoord = remoteCoord.register_resource(subcoord);
		} catch (RemoteException e) {
		    Trace.errln("Cannot make interposition:"+e);
		    return;
		}
	    }
	}
    }

    /**
     * Get the Xid of the transaction
     */
    public XidImpl getXid() {
	return (XidImpl) myCtx.getXid();
    }

    /**
     * make a SubCoordinator for this Transaction object
     */
    private void makeSubCoord() {

	Trace.outln(trace, "TransactionImpl.makeSubCoord()");

	// Build the SubCoordinator object
	try {
	    subcoord = new SubCoordinator(this, myCtx.getXid());
	} catch (RemoteException e) {
	    // should never go here.
	    Trace.errln("new SubCoordinator raised exception: "+e);
	    return;
	}

	// If interposition must be done: do it now!
	// Each time we have a remoteCoord + a subCoord, we must interpose.
	Coordinator remoteCoord = myCtx.getCoordinator();

	// First of all, create the Control object on JTM
	// if it was not created before, if interpose flag is set.
	if (interpose && remoteCoord == null) {
	    try {
		// Note that the JTM will build another Xid
		// This must be clarified XXX
		Trace.outln(trace, "Creating a remote Control on JTM for a distributed transaction");
		remoteCoord = (Coordinator) Current.getJTM().create(myCtx.getTimeout());
	    } catch (RemoteException e) {
		Trace.errln("Cannot create distributed transaction:"+e);
		return;
	    }
	    myCtx.setCoordinator(remoteCoord);
	}

	// Achieve interposition if not already done:
	// - register the subCoordinator as a Resource.
	if (remoteCoord != null && recoveryCoord == null) {
	    try {
		recoveryCoord = remoteCoord.register_resource(subcoord);
	    } catch (RemoteException e) {
		Trace.errln("Cannot make interposition:"+e);
		return;
	    }
	}
    }
}
