/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.jtm;

import java.rmi.RemoteException;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.transaction.TransactionManager;
import javax.transaction.UserTransaction;
import org.objectweb.jonas.common.Trace;

/**
 * Transaction Service implementation.
 * This singleton class must exist in each jonas server.
 * This class manages a unique Current object that implements both
 * TransactionManager and UserTransaction interfaces.
 */
public class TransactionServiceImpl implements TransactionService {

    private static final int trace = Trace.DB_4;

    // TM factory may be local or remote.
    // The TM factory is used by the JTA implementation when transactions
    // are distributed among several JVM (EJBServer or applet client)
    private TransactionFactory tm = null;

    // Unique Current object implementing the standard TransactionManager
    // set to null while service is not started.
    private Current current = null;

    // -------------------------------------------------------------------
    // JOnAS Service Implementation
    // -------------------------------------------------------------------

    private static TransactionService unique = null;

    /**
     * Private constructor to force only 1 instance.
     */
    private TransactionServiceImpl() {
    }

    /**
     * Get the unique instance. Create it at first call.
     * Return an interface of the service.
     */
    public static TransactionService getInstance() {
	if (unique == null) {
	    unique = new TransactionServiceImpl();
	}
	return unique;
    }

    /**
     * Start the Service
     * Configuration information is passed thru a Context object.
     */
    public void start(Context ctx) throws NamingException, RemoteException {

	// Gets the configuration
	String remote = (String) ctx.lookup("remote");
	boolean jtmlocal = !remote.equalsIgnoreCase("true");

	InitialContext ictx = new InitialContext();

	// Gets the Distributed Transaction Manager (JTM)
	if (jtmlocal) {
	    // Creates a JTM locally in this server
	    Trace.outln(trace, "TransactionService: Working with a colocated Transaction Manager ");
	    
	    // TODO: Recovery of the JTM

	    // Creates the ControlFactory and register it in JNDI.
	    try {
		Trace.outln(trace, "Create and register TM factory");
		tm = new TransactionFactoryImpl();
		ictx.rebind("TMFactory", (Object) tm);
	    } catch (RemoteException e) {
		Trace.errln("TransactionService: Cannot create TransactionFactory:\n"+e);
		throw e;
	    } catch (NamingException e) {
		Trace.errln("TransactionService: Cannot rebind TM:\n"+e);
		throw e;
	    }
	} else {
	    // JTM is remote, finds it by JNDI
	    Trace.outln(trace, "TransactionService: Working with a remote Transaction Manager ");
	    try {
		tm = (TransactionFactory) ictx.lookup("TMFactory");
	    } catch (NamingException e) {
		Trace.errln("TransactionService: Cannot get TM factory:"+e);
		throw e;
	    }
	}
	
	// Create and init the unique Current object.
	// Current is the local TransactionManager. It must be present in every
	// jonas server and implements the JTA TransactionManager interface.
	current = new Current(tm);

	// Register a UserTransactionFactory in JNDI.
	// Only if we are inside the JTM
	if (jtmlocal) {
	    try {
		Trace.outln(trace, "Register UserTransactionFactory");
		ictx.rebind("java:comp/UserTransaction", current);
	    } catch (NamingException e) {
		Trace.errln("Cannot rebind UserTransaction:"+e);
		throw e;
	    }
	}

    }

    /**
     * Start the Service - deprecated
     */
    public void start(boolean jtmlocal) throws NamingException, RemoteException {

	InitialContext ictx = new InitialContext();

    }

    // -------------------------------------------------------------------
    // TransactionService Implementation
    // -------------------------------------------------------------------

    /**
     * Gets the Current object instance
     */
    public Current getCurrent() {
	return current;
    }

    /**
     * Gets the TransactionManager object instance
     */
    public TransactionManager getTransactionManager() {
	return (TransactionManager) current;
    }

    /**
     * Gets the UserTransaction object instance
     */
    public UserTransaction getUserTransaction() {
	return (UserTransaction) current;
    }

    /**
     * Gets the TransactionFactory object (JTM factory)
     */
    public TransactionFactory getTransactionFactory() {
	return tm;
    }
}

