/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.ldap;

import java.io.Serializable;
import java.rmi.RemoteException;
import org.objectweb.jonas.common.RemoteObject;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.naming.Context;
import javax.naming.NamingException;
import javax.naming.NamingEnumeration;
import javax.naming.NameClassPair;
import javax.naming.Reference;
import javax.naming.Referenceable;
import javax.naming.StringRefAddr;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;

import org.objectweb.jonas.common.Trace;


/**
 * This object acts as a factory for LDAP connections (tested on Netscape Directory Server 3.0)
 */

public class DataSourceLDAPImpl implements DataSourceLDAP,
					   javax.naming.Referenceable,
					   java.io.Serializable {

    private static final int trace = Trace.DB_9;

    // ----------------------
    // DataSource properties
    // ----------------------

    /**
     * @serial
     */
    private String pDataSourceName = null;
    /**
     * @serial
     */
    private String pPassword = null;
    /**
     * @serial
     */
    private String pUrl = null;
    /**
     * @serial
     */
    private String pUserName = null;

    /**
     * @serial
     * Be aware that we have 1 datasource for each bean (and may be more if
     * the bean make several lookups!) So, we can keep here the list of
     * attributes that are mapped as byte-array for this bean.
     */
    private String pBinAttribs = null;

    /**
     * @serial
     * Notice that we have always one DirContext per DataSource.
     */
    private DirContext ldapCtx = null;


    // -----------------------------------------------------------------
    // Constructors
    // -----------------------------------------------------------------

    /**
     * Constructor for DataSourceLDAPImplFactory
     */
    public DataSourceLDAPImpl() {
	Trace.outln(trace,"DataSourceLDAPImpl.DataSourceLDAPImpl()");
    }


    // DataSourceName property
    public String getDataSourceName() {
	return pDataSourceName;
    }
    public void setDataSourceName(String s) {
	pDataSourceName = s;
    }

    // UserName property
    public String getUserName() {
	return pUserName;
    }
    public void setUserName(String s) {
	pUserName = s;
    }

    // Password property
    public String getPassword() {
	return pPassword;
    }
    public void setPassword(String s) {
	pPassword = s;
    }

    // Url property
    public String getUrl() {
	return pUrl;
    }
    public void setUrl(String s) {
	pUrl = s;
    }

    // ----------------------------------------------------------------------------------
    // interface javax.naming.Referenceable
    // ----------------------------------------------------------------------------------

    /**
     * Retrieves the Reference of this object. 
     * 
     * @return  The non-null Reference of this object. 
     * @exception  NamingException  If a naming exception was encountered while retrieving the reference.  
     */
    public Reference getReference() throws NamingException {
	Reference ref = new Reference (this.getClass().getName(), 
				       "org.objectweb.jonas.ldap.DataSourceLDAPImplFactory",
				       null);
	ref.add(new StringRefAddr("datasource.name", getDataSourceName()));
	ref.add(new StringRefAddr("datasource.url", getUrl()));
	ref.add(new StringRefAddr("datasource.username", getUserName()));
	ref.add(new StringRefAddr("datasource.password", getPassword()));
	return ref;
    }

    // ----------------------------------------------------------------------------------
    // interface org.objectweb.jonas.ldap.DataSourceLDAP
    // ----------------------------------------------------------------------------------

    public void setBinAttribs(String s) {
	pBinAttribs = s;
    }

    /**
     * Attempts to get a context to interface the LDAP server via JNDI.
     * @return  a DirContext to the database.
     * @exception  NamingException - if a database-access error occurs.
     */
    public DirContext getDirContext() throws NamingException {
	Trace.outln(trace, "DataSourceLDAPImpl.getDirContext()");
	if (ldapCtx == null) {
	    return getDirContext(pUserName, pPassword);
	} else {
	    return ldapCtx;
	}
    }



    /**
     * Attempts to get a context to interface the LDAP server via JNDI.
     *
     * @param  user - the database user on whose behalf the Connection 
     *  is being made.
     * @param password - the user's password
     * @return  a DirContext to the database
     * @exception  NamingException - if a jndi error occurs.
     */
    public DirContext getDirContext(String userName, String password) throws NamingException {
	Trace.outln(trace, "DataSourceLDAPImpl.getDirContext(String "+userName+", String "+password+")");

	if (ldapCtx == null) {
	    try {
		// Creates env 
		Hashtable envldap = new Hashtable(11);
		envldap.put(Context.INITIAL_CONTEXT_FACTORY, "com.sun.jndi.ldap.LdapCtxFactory");
		envldap.put(Context.PROVIDER_URL, pUrl);
		if (Trace.isLevelSet(Trace.DB_30)) // very verbose mode
		    envldap.put("com.sun.naming.ldap.trace.ber", System.err);
		if ((userName!=null) & (password!=null) & 
		    (!userName.equals("")) & (!password.equals(""))) {
		    // all IOs in LDAP are done for the account of:
		    envldap.put(Context.SECURITY_PRINCIPAL, userName);
		    envldap.put(Context.SECURITY_CREDENTIALS, password);
		}
		if (pBinAttribs != null) {
		    Trace.outln(trace, "notinitialized, then ... : " + pBinAttribs);
		    envldap.put("java.naming.ldap.attributes.binary", pBinAttribs);
		}

		// Creation initial context ldap
		ldapCtx = new InitialDirContext(envldap);
		Trace.outln(" Have got connection() with URL "+ pUrl);
	    } catch (NamingException e) {
		Trace.errln("Error getDirContext() : " + e);
		throw e;
	    } catch (Exception e) {
		Trace.errln("Error getDirContext() : " + e);
		System.exit(2);
	    }	
	}
	return ldapCtx;
    }
}

