/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.lib;

import java.lang.*;
import java.sql.Types;
import java.util.Hashtable;
import java.util.Vector;
import org.objectweb.jonas.common.Trace;


/**
 * This class implements methods which permit to get informations about Java Types:
 * <ul>
 * <li>
 * the JDBC Types associated
 * <li>
 * the ResultSet.getXXX() methods associated
 * <li>
 * the PreparedStatement.setXXX() methods associated
 * </ul>
 * <br>
 * (The Java Type may be or not a primitive type)
 */
public class JavaType {

    /**
     * Returns the Class object associated with the given type.
     * @param  name the fully-qualified name for a Java type.
     * @return a class object, null in error case
     */
    public static Class getTClass(String name) {
	Class c;
	Hashtable PTypes = JavaPrimitiveTypes.get();
	if ((c = (Class)PTypes.get(name)) == null) {
	    try {
		c = Class.forName(name);
	    } catch (ClassNotFoundException e) {
		c = null;
	    }
	}
	return(c);
    }

    /**
     * Returns true if the given type is the type 'void'
     */
    public static boolean isVoid(Class c) {
	return(c.equals(java.lang.Void.TYPE));
    }

    /**
     * Returns the name of if the given type
     */
    public static String getName(Class c) {
	String name;
        if (c.isArray())
        {
	    name = getName(c.getComponentType())+"[]";
        } else {
	    name = c.getName();
	}
	
        return(name);
    }

    /**
     * Returns true if the given class implements the given interface
     */
    public static boolean implementsInterface(Class c, Class interf){

	if (!interf.isInterface())
	    throw new IllegalArgumentException("JavaType.ImplementsInterface: " +
					       interf + " is not an interface");

	if (c.equals(interf)) {
	    return true;
	}

	boolean impl = false;
	Class [] interfaces = c.getInterfaces();
	for (int i=0; (i<interfaces.length && !impl); i++) {
	    if (interfaces[i].equals(interf)) {
		impl = true;
	    }
	}
	if (!impl) {
	    if (c.getSuperclass() != null) {
		impl = implementsInterface(c.getSuperclass(), interf);
	    }
	    if (!impl) {
		for (int i=0; (i<interfaces.length && !impl); i++) {
		    impl = implementsInterface(interfaces[i], interf);
		}
	    }
	}
	return(impl);
    }

    /**
     * Returns true if the given class is a Serializable Java Type,
     * false otherwise.
     */
    public static boolean isSerializable(Class c) {
	boolean isSerializable = false;
	if (c.isArray()) {
	    isSerializable = isSerializable(c.getComponentType());
	}
	else if (c.isPrimitive()) {
	    isSerializable = true;
	} else {
	    isSerializable = implementsInterface(c, java.io.Serializable.class);
	}
	return(isSerializable);
    }


    /**
     * Returns the name of the getXXX method associated with the given type in java.sql.ResultSet.
     * @param  class the class object for a Java type.
     * @return a string representing the name of the get method, null in error case
     */
    public static String getSQLGetMethod(Class c) {
	return(JavaTypesForSQL.getSQLGetMethod(c));
    }

    /**
     * Returns the name of the setXXX method associated with the given type in java.sql.ResultSet.
     * @param  class the class object for a Java type.
     * @return a string representing the name of the set method, null in error case
     */
    public static String getSQLSetMethod(Class c) {
	return(JavaTypesForSQL.getSQLSetMethod(c));
    }

    /**
     * Returns true if the given method name is getObject() or setObject()
     */
    public static boolean isXxxObjectMethod(String name) {
	return(name.equals("getObject") || name.equals("setObject"));
    }

    /**
     * Returns the SQL Type mapping the given Java Type.
     * @param  class the class object for a Java type.
     * @return the SQL Type mapping the given Java Type, (Types.OTHER in error case)
     */
    public static String getSQLType(Class c) {
	return(JavaTypesForSQL.getSQLType(c));
    }


}


class JavaPrimitiveTypes {

    private static JavaPrimitiveTypes mPrimTypes = new JavaPrimitiveTypes();
    private static Hashtable mPTypes;

    private JavaPrimitiveTypes() {
	mPTypes = new Hashtable();
	mPTypes.put(new String("boolean"), java.lang.Boolean.TYPE);
	mPTypes.put(new String("char"),    java.lang.Character.TYPE);
	mPTypes.put(new String("byte"),    java.lang.Byte.TYPE);
	mPTypes.put(new String("short"),   java.lang.Short.TYPE);
	mPTypes.put(new String("int"),     java.lang.Integer.TYPE);
	mPTypes.put(new String("long"),    java.lang.Long.TYPE);
	mPTypes.put(new String("float"),   java.lang.Float.TYPE);
	mPTypes.put(new String("double"),  java.lang.Double.TYPE);
	mPTypes.put(new String("void"),    java.lang.Void.TYPE);
    }
    
    static Hashtable get() {
	return(mPTypes);
    }

}


class JavaTypesForSQL {

    private static JavaTypesForSQL mTypesForSQL = new JavaTypesForSQL();
    private static Hashtable mGets;
    private static Hashtable mSets;
    private static Hashtable mSQLTypes;

    private JavaTypesForSQL() {
	Class c = null;
	String className = null;

	mGets = new Hashtable();
	mSets = new Hashtable();
	mSQLTypes = new Hashtable();

	mGets.put(java.lang.Boolean.TYPE, "getBoolean");
	mSets.put(java.lang.Boolean.TYPE, "setBoolean");
	mSQLTypes.put(java.lang.Boolean.TYPE, "Types.BIT");
	mGets.put(java.lang.Byte.TYPE, "getByte");
	mSets.put(java.lang.Byte.TYPE, "setByte");
	mSQLTypes.put(java.lang.Byte.TYPE, "Types.TINYINT");
	mGets.put(java.lang.Short.TYPE, "getShort");
	mSets.put(java.lang.Short.TYPE, "setShort");
	mSQLTypes.put(java.lang.Short.TYPE, "Types.SMALLINT");
	mGets.put(java.lang.Integer.TYPE, "getInt");
	mSets.put(java.lang.Integer.TYPE, "setInt");
	mSQLTypes.put(java.lang.Integer.TYPE, "Types.INTEGER");
	mGets.put(java.lang.Long.TYPE, "getLong");
	mSets.put(java.lang.Long.TYPE, "setLong");
	mSQLTypes.put(java.lang.Long.TYPE, "Types.BIGINT");
	mGets.put(java.lang.Float.TYPE, "getFloat");
	mSets.put(java.lang.Float.TYPE, "setFloat");
	mSQLTypes.put(java.lang.Float.TYPE, "Types.FLOAT");
	mGets.put(java.lang.Double.TYPE, "getDouble");
	mSets.put(java.lang.Double.TYPE, "setDouble");
	mSQLTypes.put(java.lang.Double.TYPE, "Types.DOUBLE");
	//mGets.put(java.lang.Character.TYPE, "????");
	//mSets.put(java.lang.Character.TYPE, "????");
	//mSQLTypes.put(java.lang.Character.TYPE, "???");

	try {
	    className = "java.lang.String"; c = Class.forName(className);
	    mGets.put(c, "getString");
	    mSets.put(c, "setString");
	    mSQLTypes.put(c, "Types.VARBINARY");
	    className = "java.sql.Date"; c = Class.forName(className);
	    mGets.put(c, "getDate");
	    mSets.put(c, "setDate");
	    mSQLTypes.put(c, "Types.DATE");
	    className = "java.sql.Time"; c = Class.forName(className);
	    mGets.put(c, "getTime");
	    mSets.put(c, "setTime");
	    mSQLTypes.put(c, "Types.TIME");
	    className = "java.sql.Timestamp"; c = Class.forName(className);
	    mGets.put(c, "getTimestamp");
	    mSets.put(c, "setTimestamp");
	    mSQLTypes.put(c, "Types.TIMESTAMP");
	    className = "java.lang.Short"; c = Class.forName(className);
	    mGets.put(c, "getShort");
	    mSets.put(c, "setShort");
	    mSQLTypes.put(c, "Types.SMALLINT");
	} catch (ClassNotFoundException e) {
	    Trace.errln("ERROR (JavaType): Class '" + className + "' not found");
	}

	Vector tObject = new Vector();
	tObject.addElement(new String("java.lang.Boolean"));
	tObject.addElement(new String("java.lang.Integer"));
	tObject.addElement(new String("java.lang.Long"));
	tObject.addElement(new String("java.lang.Float"));
	tObject.addElement(new String("java.lang.Double"));
	tObject.addElement(new String("java.lang.Short"));
	//tObject.addElement(new String("java.lang.byte[]"));
	tObject.addElement(new String("java.math.BigDecimal"));
	for (int i=0; i<tObject.size(); i++) {
	    className = (String)tObject.elementAt(i);
	    try {
		c = Class.forName(className);
		mGets.put(c, "getObject");
		mSets.put(c, "setObject");
		if (className.equals("java.lang.Boolean")) {
		    mSQLTypes.put(c, "Types.BIT");
		} else if (className.equals("java.lang.Integer")) {
		    mSQLTypes.put(c, "Types.INTEGER");
		} else if (className.equals("java.lang.Long")) {
		    mSQLTypes.put(c, "Types.BIGINT");
		} else if (className.equals("java.lang.Float")) {
		    mSQLTypes.put(c, "Types.REAL");
		} else if (className.equals("java.lang.Double")) {
		    mSQLTypes.put(c, "Types.DOUBLE");
		} else if (className.equals("java.lang.Short")) {
		    mSQLTypes.put(c, "Types.SMALLINT");    
		} else if (className.equals("java.lang.byte[]")) {
		    mSQLTypes.put(c, "Types.VARBINARY");
		} else if (className.equals("java.math.BigDecimal")) {
		    mSQLTypes.put(c, "Types.NUMERIC");
		}
	    } catch (ClassNotFoundException e) {
		Trace.errln("ERROR (JavaType): Class '" + className + "' not found");
	    }
	}
	
	
    }

    /**
     * Returns the name of the getXXX method associated with the given type in java.sql.ResultSet.
     * @param  class the class object for a Java type.
     * @return a string representing the name of the get method, null in error case
     */
    public static String getSQLGetMethod(Class c) {
	String val = null;
	val = (String)mGets.get(c);

	if (val == null) {
	    // See if c is byte[]or serializable[]
	    if (c.isArray()) {
		Class  component = c.getComponentType();
		if (component != null) {
		    if (component.equals(java.lang.Byte.TYPE)) {
			val = new String("getBytes");
		    } else if (JavaType.isSerializable(component)) {
			val = new String("getSerializable");
		    }
		}
	    } else if (JavaType.isSerializable(c)) {
		val = new String("getSerializable");
	    }
	}

	return(val);
    }

    /**
     * Returns the name of the setXXX method associated with the given type in java.sql.ResultSet.
     * @param  class the class object for a Java type.
     * @return a string representing the name of the set method, null in error case
     */
    public static String getSQLSetMethod(Class c) {
	String val = null;

	val = (String)mSets.get(c);


	if (val == null) {
	    if (c.isArray()) {
		// See if c is byte[] or serializable[]
		Class  component = c.getComponentType();
		if (component != null) {
		    if (component.equals(java.lang.Byte.TYPE)) {
			val = new String("setBytes");
		    } else if (JavaType.isSerializable(component)) {
			val = new String("setSerializable");
		    }
		}
	    } else if (JavaType.isSerializable(c)) {
		val = new String("setSerializable");
	    }
	}

	return(val);
    }

    /**
     * Returns the SQL Type mapping the given Java Type.
     * @param  class the class object for a Java type.
     * @return the SQL Type mapping the given Java Type, (Types.OTHER in error case)
     */
    public static String getSQLType(Class c) {
	String val = null;
	val = (String)mSQLTypes.get(c);
	if (val == null) {
	    val = new String("Types.OTHER");
	    if (c.isArray()) {
		// See if c is byte[] or serializable[]
		Class  component = c.getComponentType();
		if (component != null) {
		    if (component.equals(java.lang.Byte.TYPE)) {
			val = new String("Types.VARBINARY");
		    } else if (JavaType.isSerializable(component)) {
			val = new String("Types.VARBINARY");
		    }
		}
	    } else if (JavaType.isSerializable(c)) {
		val = new String("Types.VARBINARY");
	    }
	}

	return(val);
    }

}



