/*
 *
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.naming;

import java.util.Hashtable;
import java.util.Properties;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import org.objectweb.jonas.common.Trace;

/**
 * Naming Manager for an EJB Server.
 * this singleton class must exist in each jonas server.
 */
public class NamingManager implements ContainerNaming {

    private static final int trace = Trace.DB_3;

    private static final String JNDI_PREFIX = "java:comp/env/";

    // Naming Context associated with the thread
    private ThreadLocal threadContext = new ThreadLocal();

    // Contexts
    private InitialContext ictx = null;
    private Context serverContext = null;

    // Env.
    private Hashtable myEnv = null;

    // ------------------------------------------------------------------
    // Singleton management:
    // - the constructor is private.
    // - use static method getInstance to retrieve/create the instance.
    // ------------------------------------------------------------------
    private static NamingManager unique = null;

    /**
     * Create the naming manager.
     * useful to create InitialContext because it may throw an exception.
     */
    private NamingManager() throws NamingException {
	// Create the initial context
	// Don't use jonas.properties anymore for jndi properties
	// jndi.properties will be used instead.
	try {
	    ictx = new InitialContext();
	    myEnv = ictx.getEnvironment();
	} catch (NamingException n) {
	    Trace.errln("NamingManager: "+n.getExplanation());
	    Throwable t = n.getRootCause();
	    if (t != null) {
		if (t.getMessage().startsWith("Connection refused to host:")) {
		    Trace.errln("NamingManager: rmi registry not started ?");
		} else if (t.getMessage().startsWith("error during remote invocation")) {
		    Trace.errln("NamingManager: jrmi registry not started ?");
		} else {
		    Trace.errln("NamingManager: "+t.getMessage());
		}
	    }
	    throw n;
	}

	// Create a CompNamingContext global for this server
	// not really used today, but could be useful later.
	serverContext = new CompNamingContext("server", myEnv);
    }

    public static NamingManager getInstance() {
	if (unique == null) {
	    try {
		unique = new NamingManager();
	    } catch (NamingException e) {
		Trace.errln("Cannot create NamingManager"+e);
	    }
	}
	return unique;
    }


    // ------------------------------------------------------------------
    // ContainerNaming implementation
    // ------------------------------------------------------------------

    /**
     * return the initialContext used in this jonas server
     */
    public InitialContext getInitialContext() {
	return ictx;
    }

    /**
     * Create Context for application and component environments.
     */
    public Context createComponentContext(String namespace) throws NamingException {
	CompNamingContext ctx = new CompNamingContext(namespace);
	return (Context) ctx;
    }

    /**
     * return the Context associated with the current thread.
     */
    public Context getComponentContext() {

	Context ctx = (Context) threadContext.get();
	if (ctx == null) {
	    Trace.outln(trace, "NamingManager: getComponentContext: null context");
	} else {
	    try {
		Trace.outln(trace, "NamingManager: getComponentContext: "+ctx.getNameInNamespace());
	    } catch (NamingException e) {
		Trace.errln("NamingManager: getComponentContext: (bad name)");
	    }
	}
	return ctx;
    }

    /**
     * associate this CompNamingContext with the current thread.
     * this method should be called in preinvoke/postinvoke
     * and when we build the bean environment.
     */
    public Context setComponentContext(Context ctx) {
	if (ctx == null) {
	    Trace.outln(trace, "NamingManager: setComponentContext: reset to null");
	} else {
	    try {
		Trace.outln(trace, "NamingManager: setComponentContext: "+ctx.getNameInNamespace());
	    } catch (NamingException e) {
		Trace.errln("NamingManager: setComponentContext: (bad name)");
	    }
	}
	Context ret = (Context) threadContext.get();
	threadContext.set(ctx);
	return ret;
    }

    /**
     * Return the environment for JNDI
     * This is used only for handles today.
     */
    public Hashtable getEnv() {
	return myEnv;
    }

    // ------------------------------------------------------------------
    // other proprietary methods
    // ------------------------------------------------------------------

    /**
     * gets the server component context
     * This is used only internally in the jonas NamingManager.
     */
    public Context getServerContext() {
	return serverContext;
    }

}
