/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.rmifilters;

import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.IOException;
import java.rmi.MarshalException;
import java.rmi.UnmarshalException;

import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.jtm.Current;
import org.objectweb.jonas.jtm.TransactionImpl;
import org.objectweb.jonas.jtm.PropagationContext;

/**
 * The Skeleton class is the common superclass to all server
 *  skeletons generated by the the Bull rmic compiler used for the
 *  implicit transactional context propagation.
 *
 * inRequest() extracts in the ObjectInputStream the Control
 *  object corresponding to the propagated transaction (it may be null).
 *  and attaches the current thread to the transaction.
 * outReply() detaches the association thread/transaction.
 */
public class Skeleton {

    static final int trace = Trace.DB_6;

    /**
     * inRequest is called before invoking the operation in the implementation
     *  object. 
     * It extracts the transactional context propagated as the last parameter 
     *  of the remote operation then resumes the corresponding transaction 
     *  in the server.
     * It releases the Input Stream associated to the Remote Call object.
     */
    public void inRequest(java.rmi.server.RemoteCall call)
	throws java.lang.ClassNotFoundException, java.rmi.UnmarshalException {

	Trace.outln(trace, "inRequest --->");

        // Extracts PropagationContext from input args
        PropagationContext pc;
        try {
            ObjectInput in = call.getInputStream();
            pc = (PropagationContext) in.readObject();
            call.releaseInputStream();
        } catch (IOException e) {
            throw new UnmarshalException("error unmarshalling arguments ", e);
        }

	// Gets Current object (always existing in EJBServer)
        Current ut = Current.getCurrent();
	if (ut == null) {
	    Trace.errln("No Current in this server!");
	    return;
	}

	// Associates the current thread with this propagation context
        if (pc != null) {
	    Trace.outln(trace, "inRequest: valid transaction context");
        }
	ut.setPropagationContext(pc);
    }


    /**
     * outReply is called before return to the caller.
     * The current transaction is suspended.
     */
    public ObjectOutput outReply(java.rmi.server.RemoteCall call) 
	throws java.rmi.MarshalException, java.io.IOException, java.io.StreamCorruptedException {

	Trace.outln(trace, "outReply <---");

	ObjectOutput out = call.getResultStream(true);

	// Gets Current object (always existing in EJBServer)
        Current ut = Current.getCurrent();
	if (ut == null) {
	    Trace.errln("No Current in this server!");
	    throw new MarshalException("No Current in this server");
	}
	// Get the Propagation Context
	PropagationContext pctx = ut.getPropagationContext();

	// Dissociates the current thread from any propagation context
	if (pctx != null) {
	    Trace.outln(trace, "outReply: passing back a non null Prop.Ctxt to client");
	    ut.setPropagationContext(null);
	}

	// Add the propagation context to result stream
	try {
	    out.writeObject(pctx);
	} catch (IOException e) {
	    Trace.errln("Cannot marshall propagation context:"+e);
	    throw new MarshalException("error marshalling propagation context on reply", e);
	}

	// Used by skeleton to put the returned value, if any.
	return out;
    }
}
